//+--------------------------------------------------------------------+
//| Copyright:  (C) 2014                   |
//|                                                                    |
//| NO LIABILITY FOR CONSEQUENTIAL DAMAGES                             |
//|                                                                    |
//| In no event shall the author be liable for any damages whatsoever  |
//| (including, without limitation, incidental, direct, indirect and   |
//| consequential damages, damages for loss of business profits,       |
//| business interruption, loss of business information, or other      |
//| pecuniary loss) arising out of the use or inability to use this    |
//| product, even if advised of the possibility of such damages.       |
//+--------------------------------------------------------------------+

#property copyright "Copyright (C) 2014 Forex Software Ltd."
#property link      "http://forexsb.com"
#property version   "27.0"
#property strict



// -----------------------    External variables   ----------------------- //

extern double Entry_Amount    = 1; // Entry amount for a new position [lot]
extern double Maximum_Amount  = 20; // Maximum position amount [lot]
extern double Adding_Amount   = 1; // Amount to add on addition [lot]
extern double Reducing_Amount = 1; // Amount to close on reduction [lot]

// If account equity drops below this value, the expert will close out all positions and stop automatic trade.
// The value must be set in account currency. Example:
// Protection_Min_Account = 700 will close positions if the equity drops below 700 USD (EUR if you account is in EUR).
extern int Protection_Min_Account=0; // Stop trading at min account

// The expert checks the open positions at every tick and if found no SL or SL lower (higher for short) than selected,
// It sets SL to the defined value. The value is in points. Example:
// Protection_Max_StopLoss = 200 means 200 pips for 4 digit broker and 20 pips for 5 digit broker.
extern int Protection_Max_StopLoss=0; // Ensure maximum Stop Loss [point]

// How many seconds before the expected bar closing to rise a Bar Closing event.
extern int Bar_Close_Advance=15; // Bar closing advance [sec]

// Expert writes a log file when Write_Log_File = true.
extern bool Write_Log_File=false; // Write a log file

// Custom comment. It can be used for setting a binnary option epxiration perod
extern string Order_Comment=""; // Custom order comment

// ----------------------------    Options   ---------------------------- //

// Data bars for calculating the indicator values with the necessary precission.
int Max_Data_Bars=0;

// Have to be set to true for STP brokers that cannot set SL and TP together with the position (with OrderSend()).
// When Separate_SL_TP = true, the expert first opens the position and after that sets StopLoss and TakeProfit.
bool Separate_SL_TP=false; // Separate SL and TP orders

// The expert loads this XML file form MetaTrader "Files" folder if no XML string is provided.
string Strategy_File_Name="Strategy.xml"; // FSB Strategy Name

// The strategy as an XML string. If XML is provide, the expert loads it instead of a file.
string Strategy_XML="<?xml version=\"1.0\"?><strategy><programName>Forex Strategy Builder Professional</programName><programVersion>v3.2.4</programVersion><strategyName>New Strategy</strategyName><profileName>Default profile</profileName><dataSourceName>FSB Demo data</dataSourceName><instrumentSymbol>EURUSD</instrumentSymbol><instrumentPeriod>1440</instrumentPeriod><sameDirSignalAction>Nothing</sameDirSignalAction><oppDirSignalAction>Nothing</oppDirSignalAction><maxOpenLots>20</maxOpenLots><useAccountPercentEntry>False</useAccountPercentEntry><entryLots>1</entryLots><addingLots>1</addingLots><reducingLots>1</reducingLots><useMartingale>False</useMartingale><martingaleMultiplier>2</martingaleMultiplier><description></description><recommendedBars>1000</recommendedBars><openFilters>1</openFilters><closeFilters>0</closeFilters><firstBar>102</firstBar><minBarsRequired>103</minBarsRequired><permanentStopLoss usePermanentSL=\"True\" permanentSLType=\"Relative\">1000</permanentStopLoss><permanentTakeProfit usePermanentTP=\"True\" permanentTPType=\"Relative\">1000</permanentTakeProfit><breakEven useBreakEven=\"False\">1000</breakEven><slot slotNumber=\"0\" slotType=\"Open\"><indicatorName>Bar Opening</indicatorName><listParam paramNumber=\"0\"><caption>Logic</caption><index>0</index><value>Enter the market at the beginning of the bar</value></listParam><listParam paramNumber=\"1\"><caption>Base price</caption><index>0</index><value>Open</value></listParam></slot><slot slotNumber=\"1\" slotType=\"OpenFilter\" logicalGroup=\"A\"><indicatorName>Moving Average</indicatorName><listParam paramNumber=\"0\"><caption>Logic</caption><index>2</index><value>The bar opens above Moving Average</value></listParam><listParam paramNumber=\"1\"><caption>Smoothing method</caption><index>2</index><value>Exponential</value></listParam><listParam paramNumber=\"2\"><caption>Base price</caption><index>3</index><value>Close</value></listParam><numParam paramNumber=\"0\"><caption>Period</caption><value>100</value></numParam><numParam paramNumber=\"1\"><caption>Shift</caption><value>0</value></numParam><checkParam paramNumber=\"0\"><caption>Use previous bar value</caption><value>True</value></checkParam><signalShift>0</signalShift><signalRepeat>0</signalRepeat><indicatorSymbol></indicatorSymbol><indicatorPeriod>M1</indicatorPeriod></slot><slot slotNumber=\"2\" slotType=\"Close\"><indicatorName>Day Closing</indicatorName><listParam paramNumber=\"0\"><caption>Logic</caption><index>0</index><value>Exit the market at the end of the day</value></listParam><listParam paramNumber=\"1\"><caption>Base price</caption><index>0</index><value>Close</value></listParam></slot></strategy>"; // XML String

// TrailingStop_Moving_Step determines the step of changing the Trailing Stop.
// 0 <= TrailingStop_Moving_Step <= 2000
// If TrailingStop_Moving_Step = 0, the Trailing Stop trails at every new extreme price in the position's direction.
// If TrailingStop_Moving_Step > 0, the Trailing Stop moves at steps equal to the number of pips chosen.
// This prevents sending multiple order modifications.
int TrailingStop_Moving_Step=10;

// FIFO (First In First Out) forces the expert to close positions starting from
// the oldest one. This rule complies with the new NFA regulations.
// If you want to close the positions from the newest one (FILO), change the variable to "false".
// This doesn't change the normal work of Forex Strategy Builder.
bool FIFO_order=true;

// When the log file reaches number of lines, the expert starts a new file.
int MaxLogLinesInFile=2000;

// Used to detect a chart change
string __symbol = "";
int    __period = -1;

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum DataPeriod
  {
   DataPeriod_M1  = 1,
   DataPeriod_M5  = 5,
   DataPeriod_M15 = 15,
   DataPeriod_M30 = 30,
   DataPeriod_H1  = 60,
   DataPeriod_H4  = 240,
   DataPeriod_D1  = 1440,
   DataPeriod_W1  = 10080,
   DataPeriod_MN1 = 43200
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum PosDirection
  {
   PosDirection_None,
   PosDirection_Long,
   PosDirection_Short,
   PosDirection_Closed
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum OrderDirection
  {
   OrderDirection_None,
   OrderDirection_Buy,
   OrderDirection_Sell
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum StrategyPriceType
  {
   StrategyPriceType_Open,
   StrategyPriceType_Close,
   StrategyPriceType_Indicator,
   StrategyPriceType_CloseAndReverse,
   StrategyPriceType_Unknown
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum ExecutionTime
  {
   ExecutionTime_DuringTheBar,
   ExecutionTime_AtBarOpening,
   ExecutionTime_AtBarClosing,
   ExecutionTime_CloseAndReverse
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum TraderOrderType
  {
   TraderOrderType_Buy       = 0,
   TraderOrderType_Sell      = 1,
   TraderOrderType_BuyLimit  = 2,
   TraderOrderType_SellLimit = 3,
   TraderOrderType_BuyStop   = 4,
   TraderOrderType_SellStop  = 5
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum TradeDirection
  {
   TradeDirection_None,
   TradeDirection_Long,
   TradeDirection_Short,
   TradeDirection_Both
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum LongTradeEntryPrice
  {
   LongTradeEntryPrice_Bid,
   LongTradeEntryPrice_Ask,
   LongTradeEntryPrice_Chart
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum OperationType
  {
   OperationType_Buy,
   OperationType_Sell,
   OperationType_Close,
   OperationType_Modify
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum TickType
  {
   TickType_Open       = 0,
   TickType_OpenClose  = 1,
   TickType_Regular    = 2,
   TickType_Close      = 3,
   TickType_AfterClose = 4
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum InstrumentType
  {
   InstrumentType_Forex,
   InstrumentType_CFD,
   InstrumentType_Index
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum SlotTypes
  {
   SlotTypes_NotDefined  = 0,
   SlotTypes_Open        = 1,
   SlotTypes_OpenFilter  = 2,
   SlotTypes_Close       = 4,
   SlotTypes_CloseFilter = 8
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum IndComponentType
  {
   IndComponentType_NotDefined,
   IndComponentType_OpenLongPrice,
   IndComponentType_OpenShortPrice,
   IndComponentType_OpenPrice,
   IndComponentType_CloseLongPrice,
   IndComponentType_CloseShortPrice,
   IndComponentType_ClosePrice,
   IndComponentType_OpenClosePrice,
   IndComponentType_IndicatorValue,
   IndComponentType_AllowOpenLong,
   IndComponentType_AllowOpenShort,
   IndComponentType_ForceCloseLong,
   IndComponentType_ForceCloseShort,
   IndComponentType_ForceClose,
   IndComponentType_Other
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum PositionPriceDependence
  {
   PositionPriceDependence_None,
   PositionPriceDependence_PriceBuyHigher,
   PositionPriceDependence_PriceBuyLower,
   PositionPriceDependence_PriceSellHigher,
   PositionPriceDependence_PriceSellLower,
   PositionPriceDependence_BuyHigherSellLower,
   PositionPriceDependence_BuyLowerSelHigher,
   PositionPriceDependence_PriceBuyCrossesUpBandInwards,
   PositionPriceDependence_PriceBuyCrossesUpBandOutwards,
   PositionPriceDependence_PriceBuyCrossesDownBandInwards,
   PositionPriceDependence_PriceBuyCrossesDownBandOutwards,
   PositionPriceDependence_PriceSellCrossesUpBandInwards,
   PositionPriceDependence_PriceSellCrossesUpBandOutwards,
   PositionPriceDependence_PriceSellCrossesDownBandInwards,
   PositionPriceDependence_PriceSellCrossesDownBandOutwards
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum BasePrice
  {
   BasePrice_Open     = 0,
   BasePrice_High     = 1,
   BasePrice_Low      = 2,
   BasePrice_Close    = 3,
   BasePrice_Median   = 4, // Price[bar] = (Low[bar] + High[bar]) / 2;
   BasePrice_Typical  = 5, // Price[bar] = (Low[bar] + High[bar] + Close[bar]) / 3;
   BasePrice_Weighted = 6  // Price[bar] = (Low[bar] + High[bar] + 2 * Close[bar]) / 4;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum MAMethod
  {
   MAMethod_Simple      = 0,
   MAMethod_Weighted    = 1,
   MAMethod_Exponential = 2,
   MAMethod_Smoothed    = 3
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum SameDirSignalAction
  {
   SameDirSignalAction_Nothing,
   SameDirSignalAction_Add,
   SameDirSignalAction_Winner,
   SameDirSignalAction_Loser,
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum OppositeDirSignalAction
  {
   OppositeDirSignalAction_Nothing,
   OppositeDirSignalAction_Reduce,
   OppositeDirSignalAction_Close,
   OppositeDirSignalAction_Reverse
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum PermanentProtectionType
  {
   PermanentProtectionType_Relative,
   PermanentProtectionType_Absolute
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum TrailingStopMode
  {
   TrailingStopMode_Bar,
   TrailingStopMode_Tick
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum IndicatorLogic
  {
   IndicatorLogic_The_indicator_rises,
   IndicatorLogic_The_indicator_falls,
   IndicatorLogic_The_indicator_is_higher_than_the_level_line,
   IndicatorLogic_The_indicator_is_lower_than_the_level_line,
   IndicatorLogic_The_indicator_crosses_the_level_line_upward,
   IndicatorLogic_The_indicator_crosses_the_level_line_downward,
   IndicatorLogic_The_indicator_changes_its_direction_upward,
   IndicatorLogic_The_indicator_changes_its_direction_downward,
   IndicatorLogic_The_price_buy_is_higher_than_the_ind_value,
   IndicatorLogic_The_price_buy_is_lower_than_the_ind_value,
   IndicatorLogic_The_price_open_is_higher_than_the_ind_value,
   IndicatorLogic_The_price_open_is_lower_than_the_ind_value,
   IndicatorLogic_It_does_not_act_as_a_filter,
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
enum BandIndLogic
  {
   BandIndLogic_The_bar_opens_below_the_Upper_Band,
   BandIndLogic_The_bar_opens_above_the_Upper_Band,
   BandIndLogic_The_bar_opens_below_the_Lower_Band,
   BandIndLogic_The_bar_opens_above_the_Lower_Band,
   BandIndLogic_The_position_opens_below_the_Upper_Band,
   BandIndLogic_The_position_opens_above_the_Upper_Band,
   BandIndLogic_The_position_opens_below_the_Lower_Band,
   BandIndLogic_The_position_opens_above_the_Lower_Band,
   BandIndLogic_The_bar_opens_below_Upper_Band_after_above,
   BandIndLogic_The_bar_opens_above_Upper_Band_after_below,
   BandIndLogic_The_bar_opens_below_Lower_Band_after_above,
   BandIndLogic_The_bar_opens_above_Lower_Band_after_below,
   BandIndLogic_The_bar_closes_below_the_Upper_Band,
   BandIndLogic_The_bar_closes_above_the_Upper_Band,
   BandIndLogic_The_bar_closes_below_the_Lower_Band,
   BandIndLogic_The_bar_closes_above_the_Lower_Band,
   BandIndLogic_It_does_not_act_as_a_filter
  };

//## Import Start

//+------------------------------------------------------------------+
//| Create a text label                                              |
//+------------------------------------------------------------------+
bool LabelCreate(const long              chart_ID=0,               // chart's ID
                 const string            name="Label",             // label name
                 const int               sub_window=0,             // subwindow index
                 const int               x=0,                      // X coordinate
                 const int               y=0,                      // Y coordinate
                 const ENUM_BASE_CORNER  corner=CORNER_LEFT_UPPER, // chart corner for anchoring
                 const string            text="Label",             // text
                 const string            font="Arial",             // font
                 const int               font_size=8,              // font size
                 const color             clr=clrWhite,             // color
                 const double            angle=0.0,                // text slope
                 const ENUM_ANCHOR_POINT anchor=ANCHOR_LEFT_UPPER, // anchor type
                 const bool              back=false,               // in the background
                 const bool              selection=false,          // highlight to move
                 const bool              hidden=true,              // hidden in the object list
                 const string            tooltip="\n",             // sets the tooltip
                 const long              z_order=0)                // priority for mouse click
  {
   ResetLastError();
   if(!ObjectCreate(chart_ID,name,OBJ_LABEL,sub_window,0,0))
     {
      Print(__FUNCTION__,": failed to create text label! Error code = ",GetLastError());
      return(false);
     }
   ObjectSetInteger(chart_ID,name,OBJPROP_XDISTANCE,x);
   ObjectSetInteger(chart_ID,name,OBJPROP_YDISTANCE,y);
   ObjectSetInteger(chart_ID,name,OBJPROP_CORNER,corner);
   ObjectSetString(chart_ID,name,OBJPROP_TEXT,text);
   ObjectSetString(chart_ID,name,OBJPROP_FONT,font);
   ObjectSetInteger(chart_ID,name,OBJPROP_FONTSIZE,font_size);
   ObjectSetDouble(chart_ID,name,OBJPROP_ANGLE,angle);
   ObjectSetInteger(chart_ID,name,OBJPROP_ANCHOR,anchor);
   ObjectSetInteger(chart_ID,name,OBJPROP_COLOR,clr);
   ObjectSetInteger(chart_ID,name,OBJPROP_BACK,back);
   ObjectSetInteger(chart_ID,name,OBJPROP_SELECTABLE,selection);
   ObjectSetInteger(chart_ID,name,OBJPROP_SELECTED,selection);
   ObjectSetInteger(chart_ID,name,OBJPROP_HIDDEN,hidden);
   ObjectSetString(chart_ID,name,OBJPROP_TOOLTIP,tooltip);
   ObjectSetInteger(chart_ID,name,OBJPROP_ZORDER,z_order);
   return(true);
  }
//+------------------------------------------------------------------+
//| Change the label text                                            |
//+------------------------------------------------------------------+
bool LabelTextChange(const long chart_ID,const string name,const string text)
  {
   ResetLastError();
   if(!ObjectSetString(chart_ID,name,OBJPROP_TEXT,text))
     {
      Print(__FUNCTION__,": failed to change the text! Error code = ",GetLastError());
      return(false);
     }
   return(true);
  }
//+------------------------------------------------------------------+
//| Delete a text label                                              |
//+------------------------------------------------------------------+
bool LabelDelete(const long chart_ID=0,const string name="Label")
  {
   if(!ObjectDelete(chart_ID,name))
     {
      Print(__FUNCTION__,": failed to delete a text label! Error code = ",GetLastError());
      return(false);
     }
   return(true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
color GetChartForeColor(const long chartId=0)
  {
   long foreColor=clrWhite;
   ChartGetInteger(chartId,CHART_COLOR_FOREGROUND,0,foreColor);
   return ((color)foreColor);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string LoadStringFromFile(string filename)
  {
   string text;
   int    intSize;

   int handle= FileOpen(filename,FILE_TXT|FILE_READ|FILE_ANSI);
   if(handle == INVALID_HANDLE)
      return "";

   while(!FileIsEnding(handle))
     {
      intSize=FileReadInteger(handle,INT_VALUE);
      text+=FileReadString(handle,intSize);
     }

   FileClose(handle);
   return text;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void SaveStringToFile(string filename,string text)
  {
   int handle= FileOpen(filename,FILE_TXT|FILE_WRITE|FILE_ANSI);
   if(handle == INVALID_HANDLE)
      return;

   FileWriteString(handle,text);
   FileClose(handle);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ArrayContainsString(const string &array[],string text)
  {
   for(int i=0; i<ArraySize(array); i++)
      if(array[i]==text)
         return true;
   return false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ArrayAppendString(string &array[],string text)
  {
   int size=ArraySize(array);
   ArrayResize(array,size+1);
   array[size]=text;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string DataPeriodToString(DataPeriod period)
  {
   switch(period)
     {
      case DataPeriod_M1:
         return ("M1");
      case DataPeriod_M5:
         return ("M5");
      case DataPeriod_M15:
         return ("M15");
      case DataPeriod_M30:
         return ("M30");
      case DataPeriod_H1:
         return ("H1");
      case DataPeriod_H4:
         return ("H4");
      case DataPeriod_D1:
         return ("D1");
      case DataPeriod_W1:
         return ("W1");
      case DataPeriod_MN1:
         return ("MN1");
     }

   return ("D1");
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
DataPeriod StringToDataPeriod(string period)
  {
   if(period == "M1")
      return (DataPeriod_M1);
   if(period == "M5")
      return (DataPeriod_M5);
   if(period == "M15")
      return (DataPeriod_M15);
   if(period == "M30")
      return (DataPeriod_M30);
   if(period == "H1")
      return (DataPeriod_H1);
   if(period == "H4")
      return (DataPeriod_H4);
   if(period == "D1")
      return (DataPeriod_D1);
   if(period == "W1")
      return (DataPeriod_W1);
   if(period == "MN1")
      return (DataPeriod_MN1);
   return (DataPeriod_D1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
DataPeriod EnumTimeFramesToPeriod(int period)
  {
   switch(period)
     {
      case PERIOD_M1:
         return DataPeriod_M1;
      case PERIOD_M5:
         return DataPeriod_M5;
      case PERIOD_M15:
         return DataPeriod_M15;
      case PERIOD_M30:
         return DataPeriod_M30;
      case PERIOD_H1:
         return DataPeriod_H1;
      case PERIOD_H4:
         return DataPeriod_H4;
      case PERIOD_D1:
         return DataPeriod_D1;
      case PERIOD_W1:
         return DataPeriod_W1;
      case PERIOD_MN1:
         return DataPeriod_MN1;
     }
   return (DataPeriod_D1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void SetMAMethodsText(string &list[])
  {
   ArrayResize(list,4);
   list[0] = "Simple";
   list[1] = "Weighted";
   list[2] = "Exponential";
   list[3] = "Smoothed";
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void SetBasePricesText(string &list[])
  {
   ArrayResize(list,8);
   list[0] = "Open";
   list[1] = "High";
   list[2] = "Low";
   list[3] = "Close";
   list[4] = "Median";
   list[5] = "Typical";
   list[6] = "Low";
   list[7] = "Weighted";
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string SameDirSignalActionToString(SameDirSignalAction action)
  {
   switch(action)
     {
      case SameDirSignalAction_Add:
         return "Add";
      case SameDirSignalAction_Winner:
         return "Winner";
      case SameDirSignalAction_Loser:
         return "Loser";
      case SameDirSignalAction_Nothing:
         return "Nothing";
     }
   return "";
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string OppositeDirSignalActionToString(OppositeDirSignalAction action)
  {
   switch(action)
     {
      case OppositeDirSignalAction_Close:
         return "Close";
      case OppositeDirSignalAction_Nothing:
         return "Nothing";
      case OppositeDirSignalAction_Reduce:
         return "Reduce";
      case OppositeDirSignalAction_Reverse:
         return "Reverse";
     }
   return "";
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string SlotTypeToString(SlotTypes slotType)
  {
   string stringCaptionText="Not Defined";
   switch(slotType)
     {
      case SlotTypes_Open:
         stringCaptionText="Opening Point of the Position";
         break;
      case SlotTypes_OpenFilter:
         stringCaptionText="Opening Logic Condition";
         break;
      case SlotTypes_Close:
         stringCaptionText="Closing Point of the Position";
         break;
      case SlotTypes_CloseFilter:
         stringCaptionText="Closing Logic Condition";
         break;
     }

   return stringCaptionText;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
SlotTypes SlotTypeFromShortString(string shortString)
  {
   if(shortString == "Open")
      return (SlotTypes_Open);
   if(shortString == "OpenFilter")
      return (SlotTypes_OpenFilter);
   if(shortString == "Close")
      return (SlotTypes_Close);
   if(shortString == "CloseFilter")
      return (SlotTypes_CloseFilter);

   return SlotTypes_NotDefined;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool StringBoolToBool(string flag)
  {
   return (flag == "True" || flag == "true");
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string StringRemoveWhite(string instring)
  {
   if(instring == "" || instring == NULL)
      return "";
   string out=instring;
   string white[4]= {" ","\r","\n","\t"};
   for(int i=0; i<ArraySize(white); i++)
      StringReplace(out,white[i],"");
   return out;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class ListString
  {
   int               m_count;
   string            m_data[];
public:
   int               Count() {return m_count;}
   bool              Contains(string element);
   void              Add(string element);
   string            Get(int index) {return m_data[index];};
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ListString::Add(string element)
  {
   ArrayResize(m_data,m_count+1);
   m_data[m_count]=element;
   m_count++;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ListString::Contains(string element)
  {
   for(int i=0; i<m_count; i++)
      if(m_data[i]==element)
         return true;
   return false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class DictStringBool
  {
   int               m_count;
   string            m_key[];
   bool              m_val[];
public:
   int               Count() {return m_count;}
   bool              ContainsKey(string key);
   void              Add(string key,bool value);
   void              Set(string key,bool value);
   string            Key(int index) {return m_key[index];}
   bool              Value(string key);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool DictStringBool::ContainsKey(string key)
  {
   if(m_count==0)
      return false;
   for(int i=0; i<m_count; i++)
      if(m_key[i]==key)
         return true;
   return false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void DictStringBool::Add(string key,bool value)
  {
   ArrayResize(m_key,m_count+1);
   ArrayResize(m_val,m_count+1);
   m_key[m_count] = key;
   m_val[m_count] = value;
   m_count++;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void DictStringBool::Set(string key,bool value)
  {
   for(int i=0; i<m_count; i++)
      if(m_key[i]==key)
        {m_val[i]=value; break;}
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool DictStringBool::Value(string key)
  {
   for(int i=0; i<m_count; i++)
      if(m_key[i]==key)
         return m_val[i];

   Print("ERROR DictStringBool::Value: Geven key does not exist.");
   return false;
  }
//+------------------------------------------------------------------+

//## Import Start

#define OP_BUY   ORDER_TYPE_BUY
#define OP_SELL  ORDER_TYPE_SELL

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int AccountNumber()
  {
   return ((int)AccountInfoInteger(ACCOUNT_LOGIN));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
ENUM_TIMEFRAMES DataPeriodToTimeFrame(DataPeriod period)
  {
   switch(period)
     {
      case DataPeriod_M1:
         return PERIOD_M1;
      case DataPeriod_M5:
         return PERIOD_M5;
      case DataPeriod_M15:
         return PERIOD_M15;
      case DataPeriod_M30:
         return PERIOD_M30;
      case DataPeriod_H1:
         return PERIOD_H1;
      case DataPeriod_H4:
         return PERIOD_H4;
      case DataPeriod_D1:
         return PERIOD_D1;
      case DataPeriod_W1:
         return PERIOD_W1;
      case DataPeriod_MN1:
         return PERIOD_MN1;
     }
   return (PERIOD_D1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
datetime Time(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   datetime datetime_array[];
   ArrayResize(datetime_array,1);
   int result=CopyTime(symbol,period,bar,1,datetime_array);
   return (result==1)?datetime_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Open(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   double double_array[];
   ArrayResize(double_array,1);
   int result=CopyOpen(symbol,period,bar,1,double_array);
   return (result==1)?double_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double High(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   double double_array[];
   ArrayResize(double_array,1);
   int result=CopyHigh(symbol,period,bar,1,double_array);
   return (result==1)?double_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Low(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   double double_array[];
   ArrayResize(double_array,1);
   int result=CopyLow(symbol,period,bar,1,double_array);
   return (result==1)?double_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Close(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   double double_array[];
   ArrayResize(double_array,1);
   int result=CopyClose(symbol,period,bar,1,double_array);
   return (result==1)?double_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
long Volume(string symbol,ENUM_TIMEFRAMES period,int bar)
  {
   long long_array[];
   ArrayResize(long_array,1);
   int result=CopyTickVolume(symbol,period,bar,1,long_array);
   return (result==1)?long_array[0]:0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeDay(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.day);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeDayOfWeek(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.day_of_week);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeDayOfYear(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.day_of_year);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeHour(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.hour);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeMinute(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.min);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeMonth(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.mon);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeSeconds(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.sec);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int TimeYear(datetime date)
  {
   MqlDateTime tm;
   TimeToStruct(date,tm);
   return(tm.year);
  }
//+------------------------------------------------------------------+
//| ������� ����������� ���������� �������� �������� �� ��� ����     |
//+------------------------------------------------------------------+
string ResultRetcodeDescription(int retcode)
  {
   string str;
   switch(retcode)
     {
      case TRADE_RETCODE_REQUOTE:
         str="Requote";
         break;
      case TRADE_RETCODE_REJECT:
         str="Request rejected";
         break;
      case TRADE_RETCODE_CANCEL:
         str="Request canceled by trader";
         break;
      case TRADE_RETCODE_PLACED:
         str="Order placed";
         break;
      case TRADE_RETCODE_DONE:
         str="Request completed";
         break;
      case TRADE_RETCODE_DONE_PARTIAL:
         str="Only part of the request was completed";
         break;
      case TRADE_RETCODE_ERROR:
         str="Request processing error";
         break;
      case TRADE_RETCODE_TIMEOUT:
         str="Request canceled by timeout";
         break;
      case TRADE_RETCODE_INVALID:
         str="Invalid request";
         break;
      case TRADE_RETCODE_INVALID_VOLUME:
         str="Invalid volume in the request";
         break;
      case TRADE_RETCODE_INVALID_PRICE:
         str="Invalid price in the request";
         break;
      case TRADE_RETCODE_INVALID_STOPS:
         str="Invalid stops in the request";
         break;
      case TRADE_RETCODE_TRADE_DISABLED:
         str="Trade is disabled";
         break;
      case TRADE_RETCODE_MARKET_CLOSED:
         str="Market is closed";
         break;
      case TRADE_RETCODE_NO_MONEY:
         str="There is not enough money to complete the request";
         break;
      case TRADE_RETCODE_PRICE_CHANGED:
         str="Prices changed";
         break;
      case TRADE_RETCODE_PRICE_OFF:
         str="There are no quotes to process the request";
         break;
      case TRADE_RETCODE_INVALID_EXPIRATION:
         str="Invalid order expiration date in the request";
         break;
      case TRADE_RETCODE_ORDER_CHANGED:
         str="Order state changed";
         break;
      case TRADE_RETCODE_TOO_MANY_REQUESTS:
         str="Too frequent requests";
         break;
      case TRADE_RETCODE_NO_CHANGES:
         str="No changes in request";
         break;
      case TRADE_RETCODE_SERVER_DISABLES_AT:
         str="Autotrading disabled by server";
         break;
      case TRADE_RETCODE_CLIENT_DISABLES_AT:
         str="Autotrading disabled by client terminal";
         break;
      case TRADE_RETCODE_LOCKED:
         str="Request locked for processing";
         break;
      case TRADE_RETCODE_FROZEN:
         str="Order or position frozen";
         break;
      case TRADE_RETCODE_INVALID_FILL:
         str="Invalid order filling type";
         break;
      case TRADE_RETCODE_CONNECTION:
         str="No connection with the trade server";
         break;
      case TRADE_RETCODE_ONLY_REAL:
         str="Operation is allowed only for live accounts";
         break;
      case TRADE_RETCODE_LIMIT_ORDERS:
         str="The number of pending orders has reached the limit";
         break;
      case TRADE_RETCODE_LIMIT_VOLUME:
         str="The volume of orders and positions for the symbol has reached the limit";
         break;
      case TRADE_RETCODE_INVALID_ORDER:
         str="Incorrect or prohibited order type";
         break;
      case TRADE_RETCODE_POSITION_CLOSED:
         str="Position specified has already been closed";
         break;
      default:
         str="Unknown result";
     }
   return(str);
  }
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class DataSet
  {
public:
   // Constructor
                     DataSet(string chart);
                    ~DataSet();

   // Properties
   string            Chart;
   string            Symbol;
   DataPeriod        Period;

   int               LotSize;
   double            Spread;
   int               Digits;
   double            Point;
   double            Pip;
   bool              IsFiveDigits;
   double            StopLevel;
   double            TickValue;
   double            MinLot;
   double            MaxLot;
   double            LotStep;
   double            MarginRequired;

   int               Bars;

   datetime          ServerTime;
   double            Bid;
   double            Ask;

   datetime          Time[];
   double            Open[];
   double            High[];
   double            Low[];
   double            Close[];
   long              Volume[];

   // Methods
   void              SetPrecision();
  };

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
DataSet::DataSet(string chart)
  {
   Chart = chart;
   string parts[];
   StringSplit(chart, ',', parts);
   Symbol = parts[0];
   Period = StringToDataPeriod(parts[1]);
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
DataSet::~DataSet()
  {
  }

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void DataSet::SetPrecision()
  {
   IsFiveDigits = (Digits == 3 || Digits == 5);
   Point = 1/MathPow(10, Digits);
   Pip = IsFiveDigits ? 10*Point : Point;
  }

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class DataMarket
  {
public:
   bool              IsNewBid;

   double            OldBid;
   double            OldAsk;
   double            OldClose;
   double            Bid;
   double            Ask;
   double            Close;
   long              Volume;

   datetime          TickLocalTime;
   datetime          TickServerTime;

   double            AccountBalance;
   double            AccountEquity;
   double            AccountFreeMargin;

   int               PositionTicket;
   double            PositionLots;
   double            PositionOpenPrice;
   datetime          PositionOpenTime;
   double            PositionStopLoss;
   double            PositionTakeProfit;
   double            PositionProfit;
   string            PositionComment;
   PosDirection      PositionDirection;

   int               ConsecutiveLosses;
   int               WrongStopLoss;
   int               WrongTakeProf;
   int               WrongStopsRetry;
   bool              IsFailedCloseOrder;
   int               CloseOrderTickCounter;
   bool              IsSentCloseOrder;
  };
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class IndicatorComp
  {
public:
   // Constructors
                     IndicatorComp()
     {
      CompName          = "Not defined";
      DataType          = IndComponentType_NotDefined;
      FirstBar          = 0;
      UsePreviousBar    = 0;
      ShowInDynInfo     = true;
      PosPriceDependence=PositionPriceDependence_None;
     }

   // Properties
   string                  CompName;
   int                     FirstBar;
   int                     UsePreviousBar;
   IndComponentType        DataType;
   PositionPriceDependence PosPriceDependence;
   bool                    ShowInDynInfo;
   double                  Value[];
  };
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CListParam
  {
public:
   // Constructors
                     CListParam()
     {
      Caption = "";
      Text    = "";
      Index   = -1;
      Enabled = false;
     }

   // Properties
   string            Caption;
   string            Text;
   int               Index;
   bool              Enabled;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CNumericParam
  {
public:
   // Constructor
                     CNumericParam()
     {
      Caption = "";
      Value   = 0;
      Enabled = false;
     }

   // Properties
   string            Caption;
   double            Value;
   bool              Enabled;
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CCheckParam
  {
public:
   // Constructor
                     CCheckParam()
     {
      Caption = "";
      Checked = false;
      Enabled = false;
     }

   // Properties
   string            Caption;
   bool              Checked;
   bool              Enabled;
  };
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class Indicator
  {
protected:
   double            Sigma();
   double            Epsilon();
   void              NormalizeComponentValue(const double &componentValue[],const datetime &strategyTime[],double &output[],int ltfShift);
   int               NormalizeComponentFirstBar(int componentFirstBar,datetime &strategyTime[]);
   bool              IsSignalComponent(IndComponentType componentType);
   void              Price(BasePrice priceType,double &price[]);
   void              MovingAverage(int period,int shift,MAMethod maMethod,const double &source[],double &movingAverage[]);

   // Logic rules and signlas
   void              OscillatorLogic(int firstBar,int prvs,const double &adIndValue[],double levelLong,double levelShort,IndicatorComp &indCompLong,IndicatorComp &indCompShort,IndicatorLogic indLogic);
   void              NoDirectionOscillatorLogic(int firstBar,int prvs,const double &adIndValue[],double dLevel,IndicatorComp &indComp,IndicatorLogic indLogic);
   void              BandIndicatorLogic(int firstBar,int prvs,const double &adUpperBand[],const double &adLowerBand[],IndicatorComp &indCompLong,IndicatorComp &indCompShort,BandIndLogic indLogic);
   void              IndicatorRisesLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorFallsLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorChangesItsDirectionUpward(int firstBar,int prvs,double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorChangesItsDirectionDownward(int firstBar,int prvs,double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorIsHigherThanAnotherIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorIsLowerThanAnotherIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorCrossesAnotherIndicatorUpwardLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              IndicatorCrossesAnotherIndicatorDownwardLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarOpensAboveIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarOpensBelowIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarOpensAboveIndicatorAfterOpeningBelowLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarOpensBelowIndicatorAfterOpeningAboveLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarClosesAboveIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);
   void              BarClosesBelowIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort);

public:
   // Constructors
                     Indicator();
                    ~Indicator();

   // Properties
   string            IndicatorName;
   string            WarningMessage;
   bool              IsDiscreteValues;
   bool              UsePreviousBarValue;
   bool              IsSeparateChart;
   bool              IsBacktester;
   bool              IsDeafultGroupAll;
   bool              IsAllowLTF;

   SlotTypes         SlotType;
   ExecutionTime     ExecTime;

   CListParam        *ListParam[5];
   CNumericParam     *NumParam[6];
   CCheckParam       *CheckParam[2];

   IndicatorComp     *Component[10];
   DataSet           *Data;

   // Methods
   virtual void      Calculate(DataSet &dataSet);

   void              NormalizeComponents(DataSet &strategyDataSet);
   void              ShiftSignal(int shift);
   void              RepeatSignal(int repeat);

   int               Components();
   string            IndicatorParamToString();
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Indicator::Indicator()
  {
   IndicatorName="";

   IsBacktester      = false;
   IsDiscreteValues  = false;
   IsSeparateChart   = false;
   IsDeafultGroupAll = false;
   IsAllowLTF        = true;

   SlotType = SlotTypes_NotDefined;
   ExecTime = ExecutionTime_DuringTheBar;

   for(int i=0; i<5; i++)
      ListParam[i]=new CListParam();

   for(int i=0; i<6; i++)
      NumParam[i]=new CNumericParam();

   for(int i=0; i<2; i++)
      CheckParam[i]=new CCheckParam();

   for(int i=0; i<10; i++)
      Component[i]=new IndicatorComp();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Indicator::~Indicator()
  {
   for(int i=0; i<5; i++)
      delete ListParam[i];

   for(int i=0; i<6; i++)
      delete NumParam[i];

   for(int i=0; i<2; i++)
      delete CheckParam[i];

   for(int i=0; i<10; i++)
      delete Component[i];
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::Calculate(DataSet &dataSet)
  {
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::NormalizeComponents(DataSet &strategyDataSet)
  {
   for(int i=0; i<Components(); i++)
     {
      double value[];
      int ltfShift = Data.Period>strategyDataSet.Period ? 1 : 0;
      NormalizeComponentValue(Component[i].Value,strategyDataSet.Time,value,ltfShift);
      ArrayCopy(Component[i].Value,value);
      Component[i].FirstBar=NormalizeComponentFirstBar(Component[i].FirstBar,strategyDataSet.Time);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::ShiftSignal(int shift)
  {
   for(int i=0; i<Components(); i++)
     {
      if(!IsSignalComponent(Component[i].DataType))
         continue;
      int bars=ArraySize(Component[i].Value);
      double value[];
      ArrayResize(value,bars);
      ArrayInitialize(value,0);
      ArrayCopy(value,Component[i].Value,shift,0,WHOLE_ARRAY);
      for(int bar=0; bar<bars; bar++)
         Component[i].Value[bar]=value[bar];
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::RepeatSignal(int repeat)
  {
   for(int i=0; i<Components(); i++)
     {
      if(!IsSignalComponent(Component[i].DataType))
         continue;
      int bars=ArraySize(Component[i].Value);
      for(int bar=0; bar<bars; bar++)
        {
         if(Component[i].Value[bar]<0.5)
            continue;
         for(int r=1; r<=repeat; r++)
            if(++bar<bars)
               Component[i].Value[bar]=1;
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::NormalizeComponentValue(const double &componentValue[],const datetime &strategyTime[],double &output[],int ltfShift)
  {
   int strategyBars=ArraySize(strategyTime);
   ArrayResize(output,strategyBars);
   ArrayInitialize(output,0);

   int reachedBar = 0;
   for(int ltfBar=ltfShift; ltfBar<Data.Bars; ltfBar++)
     {
      datetime ltfOpenTime=Data.Time[ltfBar];
      datetime ltfCloseTime=ltfOpenTime+((int) Data.Period)*60;

      for(int bar=reachedBar; bar<strategyBars; bar++)
        {
         reachedBar=bar;
         datetime time=strategyTime[bar];
         if(time>=ltfOpenTime && time<ltfCloseTime)
            output[bar]=componentValue[ltfBar-ltfShift];
         else
            if(time>=ltfCloseTime)
               break;
           {return;}
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Indicator::NormalizeComponentFirstBar(int componentFirstBar,datetime &strategyTime[])
  {
   datetime firstBarTime=Data.Time[componentFirstBar];
   for(int bar=0; bar<ArraySize(strategyTime); bar++)
      if(strategyTime[bar]>=firstBarTime)
         return bar;
   return componentFirstBar;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Indicator::IsSignalComponent(IndComponentType componentType)
  {
   return
      componentType == IndComponentType_AllowOpenLong   ||
      componentType == IndComponentType_AllowOpenShort  ||
      componentType == IndComponentType_CloseLongPrice  ||
      componentType == IndComponentType_ClosePrice      ||
      componentType == IndComponentType_CloseShortPrice ||
      componentType == IndComponentType_ForceClose      ||
      componentType == IndComponentType_ForceCloseLong  ||
      componentType == IndComponentType_ForceCloseShort ||
      componentType == IndComponentType_OpenClosePrice  ||
      componentType == IndComponentType_OpenLongPrice   ||
      componentType == IndComponentType_OpenPrice       ||
      componentType == IndComponentType_OpenShortPrice;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Indicator::Components()
  {
   for(int i=0; i<10; i++)
      if(Component[i].DataType==IndComponentType_NotDefined)
         return i;
   return 10;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Indicator::IndicatorParamToString()
  {
   string text;

   for(int i=0; i<5; i++)
      if(ListParam[i].Enabled)
         text+=StringFormat("%s: %s\n",ListParam[i].Caption,ListParam[i].Text);

   for(int i=0; i<6; i++)
      if(NumParam[i].Enabled)
         text+=StringFormat("%s: %g\n",NumParam[i].Caption,NumParam[i].Value);

   for(int i=0; i<2; i++)
      if(CheckParam[i].Enabled)
         text+=StringFormat("%s: %s\n",CheckParam[i].Caption,(CheckParam[i].Checked ? "Yes" : "No"));

   return text;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::Price(BasePrice priceType,double &price[])
  {
   ArrayResize(price,Data.Bars);
   ArrayInitialize(price,0);

   switch(priceType)
     {
      case BasePrice_Open:
         ArrayCopy(price,Data.Open);
         break;
      case BasePrice_High:
         ArrayCopy(price,Data.High);
         break;
      case BasePrice_Low:
         ArrayCopy(price,Data.Low);
         break;
      case BasePrice_Close:
         ArrayCopy(price,Data.Close);
         break;
      case BasePrice_Median:
         for(int bar=0; bar<Data.Bars; bar++)
            price[bar]=(Data.Low[bar]+Data.High[bar])/2;
         break;
      case BasePrice_Typical:
         for(int bar=0; bar<Data.Bars; bar++)
            price[bar]=(Data.Low[bar]+Data.High[bar]+Data.Close[bar])/3;
         break;
      case BasePrice_Weighted:
         for(int bar=0; bar<Data.Bars; bar++)
            price[bar]=(Data.Low[bar]+Data.High[bar]+2*Data.Close[bar])/4;
         break;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::MovingAverage(int period,int shift,MAMethod maMethod,const double &source[],double &movingAverage[])
  {
   int bars=ArraySize(source);
   ArrayResize(movingAverage,bars);
   ArrayInitialize(movingAverage,0);

   if(period<=1 && shift==0)
     {
      // There is no smoothing
      ArrayCopy(movingAverage,source);
      return;
     }

   if(period>bars || period+shift<=0 || period+shift>bars)
     {
      // Error in the parameters
      string message=IndicatorName+" "+Data.Symbol+" "+DataPeriodToString(Data.Period)+
                     "Wrong MovingAverage parameters(Period: "+IntegerToString(period)+
                     ", Shift: "+IntegerToString(shift)+
                     ", Source bars: "+IntegerToString(bars)+")";
      Print(message);
      ArrayCopy(movingAverage,source);
      return;
     }

   for(int bar=0; bar<period+shift-1; bar++)
      movingAverage[bar]=0;

   double sum=0;
   for(int bar=0; bar<period; bar++)
      sum+=source[bar];

   movingAverage[period+shift-1]=sum/period;
   int lastBar=MathMin(bars,bars-shift);

   switch(maMethod)
     {
      case MAMethod_Simple:
         for(int bar=period; bar<lastBar; bar++)
            movingAverage[bar+shift]=movingAverage[bar+shift-1]+source[bar]/period-source[bar-period]/period;
         break;
      case MAMethod_Exponential:
        {
         double pr=2.0/(period+1);
         for(int bar=period; bar<lastBar; bar++)
            movingAverage[bar+shift]=source[bar]*pr+movingAverage[bar+shift-1]*(1-pr);
        }
      break;
      case MAMethod_Weighted:
        {
         double weight=period*(period+1)/2.0;
         for(int bar=period; bar<lastBar; bar++)
           {
            sum=0;
            for(int i=0; i<period; i++)
               sum+=source[bar-i]*(period-i);
            movingAverage[bar+shift]=sum/weight;
           }
        }
      break;
      case MAMethod_Smoothed:
         for(int bar=period; bar<lastBar; bar++)
            movingAverage[bar+shift]=(movingAverage[bar+shift-1]*(period-1)+source[bar])/period;
         break;
     }

   for(int bar=bars+shift; bar<bars; bar++)
      movingAverage[bar]=0;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Indicator::Sigma()
  {
   return (IsSeparateChart ? 0.000005 : Data.Point*0.5);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double Indicator::Epsilon()
  {
   return (0.0000001);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::OscillatorLogic(int firstBar,int prvs,const double &adIndValue[],double levelLong,double levelShort,IndicatorComp &indCompLong,IndicatorComp &indCompShort,IndicatorLogic indLogic)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   switch(indLogic)
     {
      case IndicatorLogic_The_indicator_rises:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int currentBar=bar-prvs;
            int baseBar=currentBar-1;
            bool isHigher=adIndValue[currentBar]>adIndValue[baseBar];

            if(!IsDiscreteValues) // Aroon oscillator uses IsDiscreteValues = true
              {
               bool isNoChange=true;
               while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange && baseBar>firstBar)
                 {
                  isNoChange=(isHigher==(adIndValue[baseBar+1]>adIndValue[baseBar]));
                  baseBar--;
                 }
              }

            indCompLong.Value[bar]=adIndValue[baseBar]<adIndValue[currentBar]-sigma ? 1 : 0;
            indCompShort.Value[bar]=adIndValue[baseBar]>adIndValue[currentBar]+sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_falls:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int currentBar=bar-prvs;
            int baseBar=currentBar-1;
            bool isHigher=adIndValue[currentBar]>adIndValue[baseBar];

            if(!IsDiscreteValues) // Aroon oscillator uses IsDiscreteValues = true
              {
               bool isNoChange=true;
               while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange && baseBar>firstBar)
                 {
                  isNoChange=(isHigher==(adIndValue[baseBar+1]>adIndValue[baseBar]));
                  baseBar--;
                 }
              }

            indCompLong.Value[bar]=adIndValue[baseBar]>adIndValue[currentBar]+sigma  ? 1 : 0;
            indCompShort.Value[bar]=adIndValue[baseBar]<adIndValue[currentBar]-sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_is_higher_than_the_level_line:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar]=adIndValue[bar-prvs]>levelLong+sigma   ? 1 : 0;
            indCompShort.Value[bar]=adIndValue[bar-prvs]<levelShort-sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_is_lower_than_the_level_line:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar]  = adIndValue[bar - prvs] < levelLong - sigma  ? 1 : 0;
            indCompShort.Value[bar] = adIndValue[bar - prvs] > levelShort + sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_crosses_the_level_line_upward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-prvs-1;
            while(MathAbs(adIndValue[baseBar]-levelLong)<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=(adIndValue[baseBar]<levelLong-sigma && adIndValue[bar-prvs]>levelLong+sigma) ? 1 : 0;
            indCompShort.Value[bar]=(adIndValue[baseBar]>levelShort+sigma && adIndValue[bar-prvs]<levelShort-sigma) ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_crosses_the_level_line_downward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-prvs-1;
            while(MathAbs(adIndValue[baseBar]-levelLong)<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=(adIndValue[baseBar]>levelLong+sigma && adIndValue[bar-prvs]<levelLong-sigma) ? 1 : 0;
            indCompShort.Value[bar]=(adIndValue[baseBar]<levelShort-sigma && adIndValue[bar-prvs]>levelShort+sigma) ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_changes_its_direction_upward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int bar0 = bar - prvs;
            int bar1 = bar0 - 1;
            while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
               bar1--;

            int iBar2=bar1-1>firstBar ? bar1-1 : firstBar;
            while(MathAbs(adIndValue[bar1]-adIndValue[iBar2])<sigma && iBar2>firstBar)
               iBar2--;

            indCompLong.Value[bar]=(adIndValue[iBar2]>adIndValue[bar1] && adIndValue[bar1]<adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
            indCompShort.Value[bar]=(adIndValue[iBar2]<adIndValue[bar1] && adIndValue[bar1]>adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_changes_its_direction_downward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int bar0 = bar - prvs;
            int bar1 = bar0 - 1;
            while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
               bar1--;

            int iBar2=bar1-1>firstBar ? bar1-1 : firstBar;
            while(MathAbs(adIndValue[bar1]-adIndValue[iBar2])<sigma && iBar2>firstBar)
               iBar2--;

            indCompLong.Value[bar]=(adIndValue[iBar2]<adIndValue[bar1] && adIndValue[bar1]>adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
            indCompShort.Value[bar]=(adIndValue[iBar2]>adIndValue[bar1] && adIndValue[bar1]<adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
           }
         break;

      default:
         return;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::NoDirectionOscillatorLogic(int firstBar,int prvs,const double &adIndValue[],double dLevel,IndicatorComp &indComp,IndicatorLogic indLogic)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
      indComp.Value[bar]=0;

   switch(indLogic)
     {
      case IndicatorLogic_The_indicator_rises:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int currentBar=bar-prvs;
            int baseBar=currentBar-1;
            bool isHigher=adIndValue[currentBar]>adIndValue[baseBar];
            bool isNoChange=true;

            while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange && baseBar>firstBar)
              {
               isNoChange=(isHigher==(adIndValue[baseBar+1]>adIndValue[baseBar]));
               baseBar--;
              }

            indComp.Value[bar]=adIndValue[baseBar]<adIndValue[currentBar]-sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_falls:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int currentBar=bar-prvs;
            int baseBar=currentBar-1;
            bool isHigher=adIndValue[currentBar]>adIndValue[baseBar];
            bool isNoChange=true;

            while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange && baseBar>firstBar)
              {
               isNoChange=(isHigher==(adIndValue[baseBar+1]>adIndValue[baseBar]));
               baseBar--;
              }

            indComp.Value[bar]=adIndValue[baseBar]>adIndValue[currentBar]+sigma ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_is_higher_than_the_level_line:
         for(int bar=firstBar; bar<Data.Bars; bar++)
            indComp.Value[bar]=adIndValue[bar-prvs]>dLevel+sigma ? 1 : 0;
         break;

      case IndicatorLogic_The_indicator_is_lower_than_the_level_line:
         for(int bar=firstBar; bar<Data.Bars; bar++)
            indComp.Value[bar]=adIndValue[bar-prvs]<dLevel-sigma ? 1 : 0;
         break;

      case IndicatorLogic_The_indicator_crosses_the_level_line_upward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-prvs-1;
            while(MathAbs(adIndValue[baseBar]-dLevel)<sigma && baseBar>firstBar)
               baseBar--;

            indComp.Value[bar]=(adIndValue[baseBar]<dLevel-sigma && adIndValue[bar-prvs]>dLevel+sigma) ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_crosses_the_level_line_downward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-prvs-1;
            while(MathAbs(adIndValue[baseBar]-dLevel)<sigma && baseBar>firstBar)
               baseBar--;

            indComp.Value[bar]=(adIndValue[baseBar]>dLevel+sigma && adIndValue[bar-prvs]<dLevel-sigma)? 1: 0;
           }
         break;

      case IndicatorLogic_The_indicator_changes_its_direction_upward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int bar0 = bar - prvs;
            int bar1 = bar0 - 1;
            while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
               bar1--;

            int bar2=bar1-1>firstBar ? bar1-1 : firstBar;
            while(MathAbs(adIndValue[bar1]-adIndValue[bar2])<sigma && bar2>firstBar)
               bar2--;

            indComp.Value[bar]=(adIndValue[bar2]>adIndValue[bar1] && adIndValue[bar1]<adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
           }
         break;

      case IndicatorLogic_The_indicator_changes_its_direction_downward:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int bar0 = bar - prvs;
            int bar1 = bar0 - 1;
            while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
               bar1--;

            int bar2=bar1-1>firstBar ? bar1-1 : firstBar;
            while(MathAbs(adIndValue[bar1]-adIndValue[bar2])<sigma && bar2>firstBar)
               bar2--;

            indComp.Value[bar]=(adIndValue[bar2]<adIndValue[bar1] && adIndValue[bar1]>adIndValue[bar0] && bar1==bar0-1) ? 1 : 0;
           }
         break;

      default:
         return;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BandIndicatorLogic(int firstBar,int prvs,const double &adUpperBand[],const double &adLowerBand[],IndicatorComp &indCompLong,IndicatorComp &indCompShort,BandIndLogic indLogic)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   switch(indLogic)
     {
      case BandIndLogic_The_bar_opens_below_the_Upper_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Open[bar]<adUpperBand[bar-prvs]-sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Open[bar]>adLowerBand[bar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_above_the_Upper_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Open[bar]>adUpperBand[bar-prvs]+sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Open[bar]<adLowerBand[bar-prvs]-sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_below_the_Lower_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Open[bar]<adLowerBand[bar-prvs]-sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Open[bar]>adUpperBand[bar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_above_the_Lower_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Open[bar]>adLowerBand[bar-prvs]+sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Open[bar]<adUpperBand[bar-prvs]-sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_below_Upper_Band_after_above:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adUpperBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=Data.Open[bar]<adUpperBand[bar-prvs]-sigma && Data.Open[baseBar]>adUpperBand[baseBar-prvs]+sigma ? 1 : 0;

            baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adLowerBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompShort.Value[bar]=Data.Open[bar]>adLowerBand[bar-prvs]+sigma && Data.Open[baseBar]<adLowerBand[baseBar-prvs]-sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_above_Upper_Band_after_below:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adUpperBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=Data.Open[bar]>adUpperBand[bar-prvs]+sigma && Data.Open[baseBar]<adUpperBand[baseBar-prvs]-sigma ? 1 : 0;

            baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adLowerBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompShort.Value[bar]=Data.Open[bar]<adLowerBand[bar-prvs]-sigma && Data.Open[baseBar]>adLowerBand[baseBar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_below_Lower_Band_after_above:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adLowerBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=Data.Open[bar]<adLowerBand[bar-prvs]-sigma && Data.Open[baseBar]>adLowerBand[baseBar-prvs]+sigma ? 1 : 0;

            baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adUpperBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompShort.Value[bar]=Data.Open[bar]>adUpperBand[bar-prvs]+sigma && Data.Open[baseBar]<adUpperBand[baseBar-prvs]-sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_opens_above_Lower_Band_after_below:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            int baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adLowerBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompLong.Value[bar]=Data.Open[bar]>adLowerBand[bar-prvs]+sigma && Data.Open[baseBar]<adLowerBand[baseBar-prvs]-sigma ? 1 : 0;

            baseBar=bar-1;
            while(MathAbs(Data.Open[baseBar]-adUpperBand[baseBar-prvs])<sigma && baseBar>firstBar)
               baseBar--;

            indCompShort.Value[bar]=Data.Open[bar]<adUpperBand[bar-prvs]-sigma && Data.Open[baseBar]>adUpperBand[baseBar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_closes_below_the_Upper_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Close[bar]<adUpperBand[bar-prvs]-sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Close[bar]>adLowerBand[bar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_closes_above_the_Upper_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Close[bar]>adUpperBand[bar-prvs]+sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Close[bar]<adLowerBand[bar-prvs]-sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_closes_below_the_Lower_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Close[bar]<adLowerBand[bar-prvs]-sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Close[bar]>adUpperBand[bar-prvs]+sigma ? 1 : 0;
           }
         break;

      case BandIndLogic_The_bar_closes_above_the_Lower_Band:
         for(int bar=firstBar; bar<Data.Bars; bar++)
           {
            indCompLong.Value[bar] =Data.Close[bar]>adLowerBand[bar-prvs]+sigma ? 1 : 0;
            indCompShort.Value[bar]=Data.Close[bar]<adUpperBand[bar-prvs]-sigma ? 1 : 0;
           }
         break;

      default:
         return;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorRisesLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      int baseBar=currentBar-1;
      bool isNoChange=true;
      bool isHigher=adIndValue[currentBar]>adIndValue[baseBar];

      while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange &&
            baseBar>firstBar)
        {
         isNoChange=(isHigher==(adIndValue[baseBar+1]>adIndValue[baseBar]));
         baseBar--;
        }

      indCompLong.Value[bar] =adIndValue[currentBar]>adIndValue[baseBar]+sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]<adIndValue[baseBar]-sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorFallsLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      int baseBar=currentBar-1;
      bool isNoChange=true;
      bool isLower=adIndValue[currentBar]<adIndValue[baseBar];

      while(MathAbs(adIndValue[currentBar]-adIndValue[baseBar])<sigma && isNoChange &&
            baseBar>firstBar)
        {
         isNoChange=(isLower==(adIndValue[baseBar+1]<adIndValue[baseBar]));
         baseBar--;
        }

      indCompLong.Value[bar] =adIndValue[currentBar]<adIndValue[baseBar]-sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]>adIndValue[baseBar]+sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorIsHigherThanAnotherIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      indCompLong.Value[bar] =adIndValue[currentBar]>adAnotherIndValue[currentBar]+sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]<adAnotherIndValue[currentBar]-sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorIsLowerThanAnotherIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      indCompLong.Value[bar] =adIndValue[currentBar]<adAnotherIndValue[currentBar]-sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]>adAnotherIndValue[currentBar]+sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorChangesItsDirectionUpward(int firstBar,int prvs,double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma= Sigma();
   for(int bar = firstBar; bar<Data.Bars; bar++)
     {
      int bar0 = bar - prvs;
      int bar1 = bar0 - 1;
      while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
         bar1--;

      int bar2=bar1-1>firstBar ? bar1-1 : firstBar;
      while(MathAbs(adIndValue[bar1]-adIndValue[bar2])<sigma && bar2>firstBar)
         bar2--;

      indCompLong.Value[bar]  = (adIndValue[bar2] > adIndValue[bar1] && adIndValue[bar1] < adIndValue[bar0] && bar1 == bar0 - 1) ? 1 : 0;
      indCompShort.Value[bar] = (adIndValue[bar2] < adIndValue[bar1] && adIndValue[bar1] > adIndValue[bar0] && bar1 == bar0 - 1) ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorChangesItsDirectionDownward(int firstBar,int prvs,double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma= Sigma();
   for(int bar = firstBar; bar<Data.Bars; bar++)
     {
      int bar0 = bar - prvs;
      int bar1 = bar0 - 1;
      while(MathAbs(adIndValue[bar0]-adIndValue[bar1])<sigma && bar1>firstBar)
         bar1--;

      int bar2=bar1-1>firstBar ? bar1-1 : firstBar;
      while(MathAbs(adIndValue[bar1]-adIndValue[bar2])<sigma && bar2>firstBar)
         bar2--;

      indCompLong.Value[bar]  = (adIndValue[bar2] < adIndValue[bar1] && adIndValue[bar1] > adIndValue[bar0] && bar1 == bar0 - 1) ? 1 : 0;
      indCompShort.Value[bar] = (adIndValue[bar2] > adIndValue[bar1] && adIndValue[bar1] < adIndValue[bar0] && bar1 == bar0 - 1) ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorCrossesAnotherIndicatorUpwardLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      int baseBar=currentBar-1;
      while(MathAbs(adIndValue[baseBar]-adAnotherIndValue[baseBar])<sigma && baseBar>firstBar)
         baseBar--;

      indCompLong.Value[bar] =adIndValue[currentBar]>adAnotherIndValue[currentBar]+sigma && adIndValue[baseBar]<adAnotherIndValue[baseBar]-sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]<adAnotherIndValue[currentBar]-sigma && adIndValue[baseBar]>adAnotherIndValue[baseBar]+sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::IndicatorCrossesAnotherIndicatorDownwardLogic(int firstBar,int prvs,const double &adIndValue[],double &adAnotherIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int currentBar=bar-prvs;
      int baseBar=currentBar-1;
      while(MathAbs(adIndValue[baseBar]-adAnotherIndValue[baseBar])<sigma && baseBar>firstBar)
        {
         baseBar--;
        }

      indCompLong.Value[bar] =adIndValue[currentBar]<adAnotherIndValue[currentBar]-sigma && adIndValue[baseBar]>adAnotherIndValue[baseBar]+sigma ? 1 : 0;
      indCompShort.Value[bar]=adIndValue[currentBar]>adAnotherIndValue[currentBar]+sigma && adIndValue[baseBar]<adAnotherIndValue[baseBar]-sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarOpensAboveIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      indCompLong.Value[bar] =Data.Open[bar]>adIndValue[bar-prvs]+sigma ? 1 : 0;
      indCompShort.Value[bar]=Data.Open[bar]<adIndValue[bar-prvs]-sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarOpensBelowIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      indCompLong.Value[bar] =Data.Open[bar]<adIndValue[bar-prvs]-sigma ? 1 : 0;
      indCompShort.Value[bar]=Data.Open[bar]>adIndValue[bar-prvs]+sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarOpensAboveIndicatorAfterOpeningBelowLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int baseBar=bar-1;
      while(MathAbs(Data.Open[baseBar]-adIndValue[baseBar-prvs])<sigma && baseBar>firstBar)
         baseBar--;

      indCompLong.Value[bar] =Data.Open[bar]>adIndValue[bar-prvs]+sigma && Data.Open[baseBar]<adIndValue[baseBar-prvs]-sigma ? 1 : 0;
      indCompShort.Value[bar]=Data.Open[bar]<adIndValue[bar-prvs]-sigma && Data.Open[baseBar]>adIndValue[baseBar-prvs]+sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarOpensBelowIndicatorAfterOpeningAboveLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      int baseBar=bar-1;
      while(MathAbs(Data.Open[baseBar]-adIndValue[baseBar-prvs])<sigma && baseBar>firstBar)
         baseBar--;

      indCompLong.Value[bar] =Data.Open[bar]<adIndValue[bar-prvs]-sigma && Data.Open[baseBar]>adIndValue[baseBar-prvs]+sigma ? 1 : 0;
      indCompShort.Value[bar]=Data.Open[bar]>adIndValue[bar-prvs]+sigma && Data.Open[baseBar]<adIndValue[baseBar-prvs]-sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarClosesAboveIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      indCompLong.Value[bar]  = Data.Close[bar] > adIndValue[bar - prvs] + sigma ? 1 : 0;
      indCompShort.Value[bar] = Data.Close[bar] < adIndValue[bar - prvs] - sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Indicator::BarClosesBelowIndicatorLogic(int firstBar,int prvs,const double &adIndValue[],IndicatorComp &indCompLong,IndicatorComp &indCompShort)
  {
   double sigma=Sigma();
   firstBar=MathMax(firstBar,2);

   for(int bar=0; bar<firstBar; bar++)
     {
      indCompLong.Value[bar]=0;
      indCompShort.Value[bar]=0;
     }

   for(int bar=firstBar; bar<Data.Bars; bar++)
     {
      indCompLong.Value[bar]  = Data.Close[bar] < adIndValue[bar - prvs] - sigma ? 1 : 0;
      indCompShort.Value[bar] = Data.Close[bar] > adIndValue[bar - prvs] + sigma ? 1 : 0;
     }
  }
//+------------------------------------------------------------------+


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class BarOpening : public Indicator
  {
public:
                     BarOpening(SlotTypes slotType)
     {
      SlotType=slotType;

      IndicatorName="Bar Opening";

      WarningMessage    = "";
      IsAllowLTF        = true;
      ExecTime          = ExecutionTime_AtBarOpening;
      IsSeparateChart   = false;
      IsDiscreteValues  = false;
      IsDeafultGroupAll = false;
     }

   virtual void      Calculate(DataSet &dataSet);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void BarOpening::Calculate(DataSet &dataSet)
  {
   Data=GetPointer(dataSet);

// Saving the components
   ArrayResize(Component[0].Value,Data.Bars);
   Component[0].CompName = "Opening price of the bar";
   Component[0].DataType = IndComponentType_OpenPrice;
   Component[0].FirstBar = 2;
   ArrayCopy(Component[0].Value,Data.Open);
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class DayClosing : public Indicator
  {
public:
                     DayClosing(SlotTypes slotType)
     {
      SlotType=slotType;

      IndicatorName="Day Closing";

      WarningMessage    = "";
      IsAllowLTF        = true;
      ExecTime          = ExecutionTime_AtBarClosing;
      IsSeparateChart   = false;
      IsDiscreteValues  = false;
      IsDeafultGroupAll = false;
     }

   virtual void      Calculate(DataSet &dataSet);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void DayClosing::Calculate(DataSet &dataSet)
  {
   Data=GetPointer(dataSet);

// Calculation
   double adClosePrice[];
   ArrayResize(adClosePrice,Data.Bars);
   ArrayInitialize(adClosePrice,0);

   for(int bar=1; bar<Data.Bars; bar++)
     {
      MqlDateTime time0;
      TimeToStruct(Data.Time[bar - 0], time0);
      MqlDateTime time1;
      TimeToStruct(Data.Time[bar - 1], time1);
      if(time1.day!=time0.day)
         adClosePrice[bar-1]=Data.Close[bar-1];
     }

   datetime dayOpeningTime=(Data.ServerTime/86400)*86400;
   datetime closeTime=dayOpeningTime+23*3600+59*60;

   if(Data.ServerTime>closeTime)
      adClosePrice[Data.Bars-1]=Data.Close[Data.Bars-1];

// Saving the components
   ArrayResize(Component[0].Value,Data.Bars);
   Component[0].CompName = "Closing price of the day";
   Component[0].DataType = IndComponentType_ClosePrice;
   Component[0].FirstBar = 2;
   ArrayCopy(Component[0].Value,adClosePrice);
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class MovingAvrg : public Indicator
  {
public:
                     MovingAvrg(SlotTypes slotType)
     {
      SlotType=slotType;

      IndicatorName="Moving Average";

      WarningMessage    = "";
      IsAllowLTF        = true;
      ExecTime          = ExecutionTime_DuringTheBar;
      IsSeparateChart   = false;
      IsDiscreteValues  = false;
      IsDeafultGroupAll = false;
     }

   virtual void      Calculate(DataSet &dataSet);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void MovingAvrg::Calculate(DataSet &dataSet)
  {
   Data=GetPointer(dataSet);

// Reading the parameters
   MAMethod  maMethod =(MAMethod) ListParam[1].Index;
   BasePrice basePrice=(BasePrice) ListParam[2].Index;
   int period=(int) NumParam[0].Value;
   int shift =(int) NumParam[1].Value;
   int previous=CheckParam[0].Checked ? 1 : 0;

// TimeExecution
   if(period==1 && shift==0)
     {
      if(basePrice == BasePrice_Open)
         ExecTime = ExecutionTime_AtBarOpening;
      else
         if(basePrice==BasePrice_Close)
            ExecTime = ExecutionTime_AtBarClosing;
     }
   else
      ExecTime = ExecutionTime_DuringTheBar;

// Calculation
   double basePrc[];
   Price(basePrice,basePrc);
   double movingAverage[];
   MovingAverage(period,shift,maMethod,basePrc,movingAverage);
   int firstBar=period+shift+1+previous;

// Saving the components
   if(SlotType==SlotTypes_Open || SlotType==SlotTypes_Close)
     {
      ArrayResize(Component[1].Value, Data.Bars);
      ArrayInitialize(Component[1].Value,0);
      for(int bar=firstBar; bar<Data.Bars; bar++)
        {
         // Covers the cases when the price can pass through the MA without a signal
         double value   = movingAverage[bar-previous];   // Current value
         double value1  = movingAverage[bar-previous-1]; // Previous value
         double tempVal = value;
         if((value1 > Data.High[bar - 1] && value < Data.Open[bar]) || // The Data.Open price jumps above the indicator
            (value1 < Data.Low[bar - 1]  && value > Data.Open[bar]) || // The Data.Open price jumps below the indicator
            (Data.Close[bar-1]<value     && value < Data.Open[bar]) || // The Data.Open price is in a positive gap
            (Data.Close[bar-1]>value     && value > Data.Open[bar])) // The Data.Open price is in a negative gap
            tempVal=Data.Open[bar];
         Component[1].Value[bar]=tempVal; // Entry or exit value
        }
     }
   else
     {
      ArrayResize(Component[1].Value, Data.Bars);
      ArrayResize(Component[2].Value, Data.Bars);
     }

   ArrayResize(Component[0].Value,Data.Bars);
   Component[0].CompName = "MA Value";
   Component[0].DataType = IndComponentType_IndicatorValue;
   Component[0].FirstBar = firstBar;
   ArrayCopy(Component[0].Value,movingAverage);

   if(SlotType==SlotTypes_Open)
     {
      Component[1].CompName = "Position opening price";
      Component[1].DataType = IndComponentType_OpenPrice;
     }
   if(SlotType==SlotTypes_OpenFilter)
     {
      Component[1].DataType = IndComponentType_AllowOpenLong;
      Component[1].CompName = "Is long entry allowed";
      Component[2].DataType = IndComponentType_AllowOpenShort;
      Component[2].CompName = "Is short entry allowed";
     }
   if(SlotType==SlotTypes_Close)
     {
      Component[1].CompName = "Position closing price";
      Component[1].DataType = IndComponentType_ClosePrice;
     }
   if(SlotType==SlotTypes_CloseFilter)
     {
      Component[1].DataType = IndComponentType_ForceCloseLong;
      Component[1].CompName = "Close out long position";
      Component[2].DataType = IndComponentType_ForceCloseShort;
      Component[2].CompName = "Close out short position";
     }

   if(SlotType==SlotTypes_OpenFilter || SlotType==SlotTypes_CloseFilter)
     {
      if(ListParam[0].Text=="Moving Average rises")
         IndicatorRisesLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
      else
         if(ListParam[0].Text=="Moving Average falls")
            IndicatorFallsLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
         else
            if(ListParam[0].Text=="The bar opens above Moving Average")
               BarOpensAboveIndicatorLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
            else
               if(ListParam[0].Text=="The bar opens below Moving Average")
                  BarOpensBelowIndicatorLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
               else
                  if(ListParam[0].Text=="The bar opens above Moving Average after opening below it")
                     BarOpensAboveIndicatorAfterOpeningBelowLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
                  else
                     if(ListParam[0].Text=="The bar opens below Moving Average after opening above it")
                        BarOpensBelowIndicatorAfterOpeningAboveLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
                     else
                        if(ListParam[0].Text=="The position opens above Moving Average")
                          {
                           Component[0].PosPriceDependence=PositionPriceDependence_BuyHigherSellLower;
                           Component[0].UsePreviousBar=previous;
                           Component[1].DataType=IndComponentType_Other;
                           Component[1].ShowInDynInfo=false;
                           Component[2].DataType=IndComponentType_Other;
                           Component[2].ShowInDynInfo=false;
                          }
                        else
                           if(ListParam[0].Text=="The position opens below Moving Average")
                             {
                              Component[0].PosPriceDependence=PositionPriceDependence_BuyLowerSelHigher;
                              Component[0].UsePreviousBar=previous;
                              Component[1].DataType=IndComponentType_Other;
                              Component[1].ShowInDynInfo=false;
                              Component[2].DataType=IndComponentType_Other;
                              Component[2].ShowInDynInfo=false;
                             }
                           else
                              if(ListParam[0].Text=="The bar closes below Moving Average")
                                 BarClosesBelowIndicatorLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
                              else
                                 if(ListParam[0].Text=="The bar closes above Moving Average")
                                    BarClosesAboveIndicatorLogic(firstBar,previous,movingAverage,Component[1],Component[2]);
     }
  }
//+------------------------------------------------------------------+



//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class IndicatorManager
  {
public:
   Indicator         *CreateIndicator(string indicatorName,SlotTypes slotType);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Indicator *IndicatorManager::CreateIndicator(string indicatorName,SlotTypes slotType)
  {
   if(indicatorName == "Bar Opening")
      return new BarOpening(slotType);
   if(indicatorName == "Day Closing")
      return new DayClosing(slotType);
   if(indicatorName == "Moving Average")
      return new MovingAvrg(slotType);

   return NULL;
  }
//+------------------------------------------------------------------+


//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class IndicatorSlot
  {
public:
   // Constructors
                     IndicatorSlot();
                    ~IndicatorSlot();

   // Properties
   int               SlotNumber;
   SlotTypes         SlotType;
   string            IndicatorName;
   string            LogicalGroup;
   int               SignalShift;
   int               SignalRepeat;
   string            IndicatorSymbol;
   DataPeriod        IndicatorPeriod;

   Indicator         *IndicatorPointer;

   // Methods
   bool              GetUsePreviousBarValue();
   string            LogicalGroupToString();
   string            AdvancedParamsToString();
   string            GetIndicatorSymbol(string baseSymbol);
   DataPeriod        GetIndicatorPeriod(DataPeriod basePeriod);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
IndicatorSlot::IndicatorSlot()
  {
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
IndicatorSlot::~IndicatorSlot()
  {
   if(CheckPointer(IndicatorPointer)==POINTER_DYNAMIC)
      delete IndicatorPointer;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool IndicatorSlot::GetUsePreviousBarValue()
  {
   for(int i=0; i<ArraySize(IndicatorPointer.CheckParam); i++)
      if(IndicatorPointer.CheckParam[i].Caption=="Use previous bar value")
         return IndicatorPointer.CheckParam[i].Checked;
   return false;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string IndicatorSlot::LogicalGroupToString()
  {
   return ("Logical group: " + LogicalGroup);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string IndicatorSlot::AdvancedParamsToString()
  {
   string text = "Signal shift: " + IntegerToString(SignalShift) + "\n";
   if(SlotType == SlotTypes_OpenFilter || SlotType == SlotTypes_CloseFilter)
      text+="Signal repeat: "+IntegerToString(SignalRepeat)+"\n";

   string symbol = (IndicatorSymbol == "")            ? "Default" : IndicatorSymbol;
   string period = (IndicatorPeriod == DataPeriod_M1) ? "Default" : DataPeriodToString(IndicatorPeriod);
   text += "Symbol: " + symbol + "\n";
   text += "Period: " + period + "\n";

   return text;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string IndicatorSlot::GetIndicatorSymbol(string baseSymbol)
  {
   return (IndicatorSymbol == "") ? baseSymbol : IndicatorSymbol;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
DataPeriod IndicatorSlot::GetIndicatorPeriod(DataPeriod basePeriod)
  {
   return (IndicatorPeriod < basePeriod) ? basePeriod : IndicatorPeriod;
  }
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class Strategy
  {
private:
   // Fields
   string            m_symbol;
   DataPeriod        m_period;
   bool              m_isTester;
   int               m_openSlots;
   int               m_closeSlots;

   // Methods
   bool              IsLongerTimeFrame(int slotNumber);
   string            GetSlotChart(int slotNumber);
public:
   // Constructors
                     Strategy(int openSlots,int closeSlots);
                    ~Strategy();

   // Properties
   string            StrategyName;
   string            Description;
   double            AddingLots;
   double            ReducingLots;
   double            EntryLots;
   double            MaxOpenLots;
   bool              UseAccountPercentEntry;
   bool              UsePermanentSL;
   int               PermanentSL;
   bool              UsePermanentTP;
   int               PermanentTP;
   bool              UseBreakEven;
   int               BreakEven;
   bool              UseMartingale;
   double            MartingaleMultiplier;
   int               FirstBar;
   int               MinBarsRequired;
   int               RecommendedBars;

   PermanentProtectionType PermanentTPType;
   PermanentProtectionType PermanentSLType;
   OppositeDirSignalAction OppSignalAction;
   SameDirSignalAction     SameSignalAction;

   IndicatorSlot     *Slot[];

   // Methods
   void              SetSymbol(string symbol)   {m_symbol   = symbol;}
   void              SetPeriod(int period)      {m_period   = EnumTimeFramesToPeriod(period);}
   void              SetIsTester(bool isTester) {m_isTester = isTester;}
   string            GetSymbol()                {return m_symbol;}
   DataPeriod        GetPeriod()                {return m_period;}
   bool              IsTester()                 {return m_isTester;}
   int               OpenSlots()                {return m_openSlots;};
   int               CloseSlots()               {return m_closeSlots;};
   int               Slots()                    {return m_openSlots + m_closeSlots + 2;}
   int               CloseSlotNumber()          {return m_openSlots + 1;}
   SlotTypes         GetSlotType(int slotNumber);
   void              GetRequiredCharts(string &charts[]);
   bool              IsUsingLogicalGroups();
   bool              IsLogicalGroupSpecial(int slotNumber);
   string            GetDefaultGroup(int slotNumber);
   void              CalculateStrategy(DataSet *&dataSet[]);
   string            DynamicInfoText();
   void              DynamicInfoInitArrays(string &params[],string &values[]);
   void              DynamicInfoSetValues(string &values[]);
   string            ToString();
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Strategy::Strategy(int openSlots,int closeSlots)
  {
   m_openSlots  = openSlots;
   m_closeSlots = closeSlots;

   ArrayResize(Slot,Slots());

   for(int i=0; i<Slots(); i++)
     {
      Slot[i]=new IndicatorSlot();
      Slot[i].SlotNumber=i;
      Slot[i].SlotType=GetSlotType(i);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Strategy::~Strategy()
  {
   for(int slot=0; slot<ArraySize(Slot); slot++)
      delete Slot[slot];
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
SlotTypes Strategy::GetSlotType(int slotNumber)
  {
   if(slotNumber == 0)
      return (SlotTypes_Open);
   else
      if(slotNumber <  CloseSlotNumber())
         return (SlotTypes_OpenFilter);
      else
         if(slotNumber == CloseSlotNumber())
            return (SlotTypes_Close);
         else
            return (SlotTypes_CloseFilter);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Strategy::IsLongerTimeFrame(int slotNumber)
  {
   return !(Slot[slotNumber].IndicatorSymbol=="" && Slot[slotNumber].IndicatorPeriod==DataPeriod_M1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Strategy::GetSlotChart(int slotNumber)
  {
   string     symbol = Slot[slotNumber].GetIndicatorSymbol(GetSymbol());
   DataPeriod period = Slot[slotNumber].GetIndicatorPeriod(GetPeriod());
   string chart=symbol+","+DataPeriodToString(period);
   return (chart);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Strategy::GetRequiredCharts(string &charts[])
  {
   ArrayResize(charts,1);
   charts[0]=GetSymbol()+","+DataPeriodToString(GetPeriod());

   for(int i=0; i<Slots(); i++)
     {
      if(!Slot[i].IndicatorPointer.IsAllowLTF)
         continue;
      if(!IsLongerTimeFrame(i))
         continue;
      string chart=GetSlotChart(i);
      if(!ArrayContainsString(charts,chart))
         ArrayAppendString(charts,chart);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Strategy::CalculateStrategy(DataSet *&dataSet[])
  {
   for(int i=0; i<Slots(); i++)
     {
      string chart=GetSlotChart(i);
      for(int j=0; j<ArraySize(dataSet); j++)
        {
         if(dataSet[j].Chart!=chart)
            continue;
         Slot[i].IndicatorPointer.Calculate(dataSet[j]);
         if(IsLongerTimeFrame(i))
            Slot[i].IndicatorPointer.NormalizeComponents(dataSet[0]);
         if(Slot[i].SignalShift>0)
            Slot[i].IndicatorPointer.ShiftSignal(Slot[i].SignalShift);
         if(Slot[i].SignalRepeat>0)
            Slot[i].IndicatorPointer.RepeatSignal(Slot[i].SignalRepeat);
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Strategy::IsUsingLogicalGroups()
  {
   bool isUsingGroups=false;
   for(int slot=0; slot<ArraySize(Slot); slot++)
     {
      SlotTypes slotType=Slot[slot].SlotType;
      if(slotType==SlotTypes_OpenFilter)
        {
         string defaultGroup = GetDefaultGroup(slot);
         string logicalGroup = Slot[slot].LogicalGroup;
         if(defaultGroup!=logicalGroup && logicalGroup!="All")
           {
            isUsingGroups=true;
            break;
           }
        }
      else
         if(slotType==SlotTypes_CloseFilter)
           {
            string defaultGroup = GetDefaultGroup(slot);
            string logicalGroup = Slot[slot].LogicalGroup;
            if(defaultGroup!=logicalGroup)
              {
               isUsingGroups=true;
               break;
              }
           }
     }
   return (isUsingGroups);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Strategy::IsLogicalGroupSpecial(int slotNumber)
  {
   SlotTypes slotType=Slot[slotNumber].SlotType;
   string group=Slot[slotNumber].LogicalGroup;
   if(slotType==SlotTypes_Open || slotType==SlotTypes_Close)
      return (false);
   if(slotType==SlotTypes_OpenFilter && group!=GetDefaultGroup(slotNumber) && group!="[All]")
      return (true);
   if(slotType==SlotTypes_CloseFilter && group!=GetDefaultGroup(slotNumber))
      return (true);
   if(slotType==SlotTypes_CloseFilter)
     {
      int count = 0;
      for(int i = OpenSlots() + 2; i < Slots(); i++)
         if(Slot[i].LogicalGroup==group)
            count++;
      if(count>1)
         return (true);
     }
   return (false);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Strategy::GetDefaultGroup(int slotNumber)
  {
   string group="";
   SlotTypes slotType=GetSlotType(slotNumber);
   if(slotType==SlotTypes_OpenFilter)
      group=Slot[slotNumber].IndicatorPointer.IsDeafultGroupAll ? "All" : "A";
   else
      if(slotType==SlotTypes_CloseFilter)
        {
         int index=slotNumber-CloseSlotNumber()-1;
         group=IntegerToString('a'+index);
        }
   return (group);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Strategy::DynamicInfoInitArrays(string &params[],string &values[])
  {
   ArrayResize(params,200);
   ArrayResize(values,200);
   for(int i=0; i<200; i++)
     {
      params[i] = "";
      values[i] = "";
     }

   int index=-2;
   for(int slot=0; slot<Slots(); slot++)
     {
      index++;
      index++;
      params[index]=Slot[slot].IndicatorName;
      for(int i=0; i<Slot[slot].IndicatorPointer.Components(); i++)
        {
         IndComponentType type=Slot[slot].IndicatorPointer.Component[i].DataType;
         if(type==IndComponentType_NotDefined)
            continue;
         if(Slot[slot].IndicatorPointer.Component[i].ShowInDynInfo)
           {
            index++;
            params[index]=Slot[slot].IndicatorPointer.Component[i].CompName;
           }
        }
     }
   ArrayResize(params,index+1);
   ArrayResize(values,index+1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Strategy::DynamicInfoSetValues(string &values[])
  {
   int index=-1;
   for(int slot=0; slot<Slots(); slot++)
     {
      index++;
      index++;
      for(int i=0; i<Slot[slot].IndicatorPointer.Components(); i++)
        {
         IndicatorComp *component=Slot[slot].IndicatorPointer.Component[i];
         IndComponentType type=component.DataType;
         if(type==IndComponentType_NotDefined)
            continue;
         int bars=ArraySize(component.Value);
         if(bars<3)
            continue;

         string name=component.CompName;
         double value0=component.Value[bars-1];
         double value1=component.Value[bars-2];
         double dl0=MathAbs(value0);
         double dl1=MathAbs(value0);
         string sFr0=dl0<10?"%10.5f":dl0<100?"%10.5f":dl0<1000?"%10.3f":dl0<10000?"%10.3f":dl0<100000?"%10.2f":"%10.1f";
         string sFr1=dl1<10?"%10.5f":dl1<100?"%10.5f":dl1<1000?"%10.3f":dl1<10000?"%10.3f":dl1<100000?"%10.2f":"%10.1f";
         string format=sFr1+"    "+sFr0;
         if(component.ShowInDynInfo)
           {
            if(type==IndComponentType_AllowOpenLong || type==IndComponentType_AllowOpenShort ||
               type==IndComponentType_ForceClose    || type==IndComponentType_ForceCloseLong ||
               type==IndComponentType_ForceCloseShort)
               values[index]=StringFormat("%13s    %13s",(value1<1 ? "No" : "Yes"),(value0<1 ? "No" : "Yes"));
            else
               values[index]=StringFormat(format,value1,value0);
            index++;
           }
         component=NULL;
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Strategy::DynamicInfoText()
  {
   string info;
   for(int slot=0; slot<Slots(); slot++)
     {
      info+="\n\n"+Slot[slot].IndicatorName;
      for(int i=0; i<Slot[slot].IndicatorPointer.Components(); i++)
        {
         IndicatorComp *component=Slot[slot].IndicatorPointer.Component[i];
         IndComponentType type=component.DataType;
         if(type==IndComponentType_NotDefined)
            continue;
         int bars=ArraySize(component.Value);
         if(bars<4)
            continue;

         string name=component.CompName;
         double value0=component.Value[bars-1];
         double value1=component.Value[bars-2];
         double value2=component.Value[bars-3];
         double dl=MathAbs(value0);
         string sFr=dl<10?"%10.6f":dl<100?"%10.5f":dl<1000?"%10.4":dl<10000?"%10.3f":dl<100000?"%10.2f":"%10.1f";
         string format="\n%-40s "+sFr+"    "+sFr+"    "+sFr;
         if(component.ShowInDynInfo)
           {
            if(type==IndComponentType_AllowOpenLong || type==IndComponentType_AllowOpenShort ||
               type==IndComponentType_ForceClose    || type==IndComponentType_ForceCloseLong ||
               type==IndComponentType_ForceCloseShort)
               info+=StringFormat("\n%-42s %-10s    %-10s    %-10s",name,(value2<1?"No":"Yes"),(value1<1?"No":"Yes"),(value0<1?"No":"Yes"));
            else
               info+=StringFormat(format,name,value2,value1,value0);
           }
         component=NULL;
        }
     }
   return (info);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Strategy::ToString()
  {
   string stopLoss   = UsePermanentSL ? IntegerToString(PermanentSL)            : "None";
   string takeProfit = UsePermanentTP ? IntegerToString(PermanentTP)            : "None";
   string breakEven  = UseBreakEven   ? IntegerToString(BreakEven)              : "None";
   string martingale = UseMartingale  ? DoubleToString(MartingaleMultiplier, 2) : "None";

   string text="Name: "           +StrategyName                                    +"\n"+
               "Symbol: "         +GetSymbol()                                     +"\n"+
               "Period: "         +DataPeriodToString(GetPeriod())                 +"\n\n"+
               "Trade unit: "     +(UseAccountPercentEntry ? "Percent" : "Lot")    +"\n"+
               "Entry amount: "   +DoubleToString(EntryLots,2)                     +"\n"+
               "Max open lots: "  +DoubleToString(MaxOpenLots,2)                   +"\n\n"+
               "Same signal: "    +SameDirSignalActionToString(SameSignalAction)   +"\n"+
               "Adding amount: "  +DoubleToString(AddingLots,2)                    +"\n"+
               "Opposite signal: "+OppositeDirSignalActionToString(OppSignalAction)+"\n"+
               "Reducing amount: "+DoubleToString(ReducingLots,2)                  +"\n\n"+
               "Stop Loss: "      +stopLoss   +"\n"+
               "Take Profit: "    +takeProfit +"\n"+
               "Break Even: "     +breakEven  +"\n\n"+
               "Martingale: "     +martingale +"\n\n"+
               "Description: "    +Description+"\n\n";

   for(int slot=0; slot<ArraySize(Slot); slot++)
     {
      text+=SlotTypeToString(Slot[slot].SlotType)+"\n"+
            Slot[slot].IndicatorName+"\n" +
            Slot[slot].IndicatorPointer.IndicatorParamToString()+"\n";

      if(Slot[slot].SlotType==SlotTypes_OpenFilter || Slot[slot].SlotType==SlotTypes_CloseFilter)
         text+=Slot[slot].LogicalGroupToString();
      if(Slot[slot].IndicatorPointer.IsAllowLTF)
         text+=Slot[slot].AdvancedParamsToString()+"\n";
     }
   return (text);
  }
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//| Includes                                                         |
//+------------------------------------------------------------------+
#include <Object.mqh>
#include <Arrays\ArrayObj.mqh>

//+------------------------------------------------------------------+
//| macros                                                           |
//+------------------------------------------------------------------+
#define EASYXML_START_OPEN          "<"
#define EASYXML_START_CLOSE         ">"
#define EASYXML_SELFCLOSE           "/>"
#define EASYXML_CLOSE_OPEN          "</"
#define EASYXML_CLOSE_CLOSE         ">"
#define EASYXML_WHITESPACE          " \r\n\t"
#define EASYXML_LATIN               "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz"
#define EASYXML_CDATA_OPEN          "<![CDATA["
#define EASYXML_CDATA_CLOSE         "]]>"
#define EASYXML_COMMENT_OPEN        "<!--"
#define EASYXML_COMMENT_CLOSE       "-->"
#define EASYXML_PROLOG_OPEN         "<?xml"
#define EASYXML_PROLOG_CLOSE        "?>"
#define EASYXML_DOCTYPE_OPEN        "<!DOCTYPE"
#define EASYXML_DOCTYPE_CLOSE        ">"
#define EASYXML_ATTRIBUTE_SEPARATOR "="
#define EASYXML_ATTRIBUTE_COLON     "\"'"
#define EASYXML_XMLFILE_ENDING      ".xml"
#define HTTP_QUERY_CONTENT_LENGTH   5

#define EASYXML_ERR_CONNECTION_ATTEMPT      1
#define EASYXML_ERR_CONNECTION_OPEN         2
#define EASYXML_ERR_CONNECTION_URL          3
#define EASYXML_ERR_CONNECTION_FILEOPEN     4
#define EASYXML_ERR_CONNECTION_EMPTYSTREAM  5
#define EASYXML_INVALID_PROLOG              6
#define EASYXML_INVALID_COMMENT             7
#define EASYXML_INVALID_OPENTAG_START       8
#define EASYXML_INVALID_OPENTAG_CLOSE       9
#define EASYXML_INVALID_CLOSETAG            10
#define EASYXML_NO_CLOSETAG                 11
#define EASYXML_INVALID_CDATA               12
#define EASYXML_INVALID_ATTRIBUTE           13
#define EASYXML_INVALID_FILENAME            14
#define EASYXML_INVALID_DOCTYPE             15


//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CEasyXmlNode : public CObject
  {
private:
   CArrayObj         *ChildNodes;
   CArrayObj         *AttributeNodes;
   CEasyXmlNode      *ParentNode;
   string            sName;
   string            sValue;

public:
   void              SetName(string pName);
   void              SetValue(string pValue);
   void              SetAttribute(string pName,string pValue);

   string            GetName(void);
   string            GetValue(void);
   string            GetAttribute(string pName);

   CEasyXmlNode      *CreateChild(CEasyXmlNode *pChildNode);
   CEasyXmlNode      *CreateSibling(CEasyXmlNode *pSiblingNode);

   CEasyXmlNode      *Parent(void);
   void              Parent(CEasyXmlNode *pParentNode);

   CArrayObj         *Children(void);
   CArrayObj         *Attributes(void);
   CEasyXmlNode      *LastChild(void);
   CEasyXmlNode      *FirstChild(void);

                     CEasyXmlNode();
                    ~CEasyXmlNode();
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CEasyXmlNode::CEasyXmlNode()
  {
   ChildNodes=new CArrayObj;
   AttributeNodes=new CArrayObj;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
CEasyXmlNode::~CEasyXmlNode()
  {
   if(CheckPointer(ChildNodes) == POINTER_DYNAMIC)
     {
      for(int i = 0; i < ChildNodes.Total(); i++)
        {
         ChildNodes.Delete(i);
        }
      delete ChildNodes;
     }
   if(CheckPointer(AttributeNodes) == POINTER_DYNAMIC)
     {
      for(int i = 0; i < AttributeNodes.Total(); i++)
        {
         AttributeNodes.Delete(i);
        }
      delete AttributeNodes;
     }
  }
//+------------------------------------------------------------------+
//| set tag name                                                     |
//+------------------------------------------------------------------+
void CEasyXmlNode::SetName(string pName)
  {
   sName=pName;
  }
//+------------------------------------------------------------------+
//| set tag content value                                            |
//+------------------------------------------------------------------+
void CEasyXmlNode::SetValue(string pValue)
  {
   sValue=pValue;
  }
//+------------------------------------------------------------------+
//| set attribute                                                    |
//+------------------------------------------------------------------+
void CEasyXmlNode::SetAttribute(string pName,string pValue)
  {
   CEasyXmlAttribute *Attribute=new CEasyXmlAttribute;

   Attribute.SetName(pName);
   Attribute.SetValue(pValue);

   AttributeNodes.Add(Attribute);
  }
//+------------------------------------------------------------------+
//| get attribute                                                    |
//+------------------------------------------------------------------+
string CEasyXmlNode::GetAttribute(string pName)
  {
   CEasyXmlAttribute *Attribute;

   for(int i=0; i<AttributeNodes.Total(); i++)
     {
      Attribute=AttributeNodes.At(i);

      if(Attribute.GetName()==pName)
        {
         return Attribute.GetValue();
        }
     }

   return("");
  }
//+------------------------------------------------------------------+
//| get tag name                                                     |
//+------------------------------------------------------------------+
string CEasyXmlNode::GetName(void)
  {
   return sName;
  }
//+------------------------------------------------------------------+
//| set tag content value                                            |
//+------------------------------------------------------------------+
string CEasyXmlNode::GetValue(void)
  {
   return sValue;
  }
//+------------------------------------------------------------------+
//| get child nodes                                                  |
//+------------------------------------------------------------------+
CArrayObj *CEasyXmlNode::Children(void)
  {
   return ChildNodes;
  }
//+------------------------------------------------------------------+
//| get Attributes                                                   |
//+------------------------------------------------------------------+
CArrayObj *CEasyXmlNode::Attributes(void)
  {
   return AttributeNodes;
  }
//+------------------------------------------------------------------+
//| create child node                                                |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXmlNode::CreateChild(CEasyXmlNode *pChildNode)
  {
   pChildNode.Parent(GetPointer(this));
   ChildNodes.Add(pChildNode);

   return(ChildNodes.At(ChildNodes.Total()-1));
  }
//+------------------------------------------------------------------+
//| create sibling node                                              |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXmlNode::CreateSibling(CEasyXmlNode *pSiblingNode)
  {
   pSiblingNode.Prev(GetPointer(this));
   GetPointer(this).Next(pSiblingNode);

   ParentNode.CreateChild(pSiblingNode);
   return ParentNode.LastChild();
  }
//+------------------------------------------------------------------+
//| get last child node in array                                     |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXmlNode::FirstChild(void)
  {
   return ChildNodes.At(0);
  }
//+------------------------------------------------------------------+
//| get last child node in array                                     |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXmlNode::LastChild(void)
  {
   return ChildNodes.At(ChildNodes.Total()-1);
  }
//+------------------------------------------------------------------+
//| get parent node                                                  |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXmlNode::Parent(void)
  {
   return ParentNode;
  }
//+------------------------------------------------------------------+
//| set parent node                                                  |
//+------------------------------------------------------------------+
void CEasyXmlNode::Parent(CEasyXmlNode *pParentNode)
  {
   ParentNode=pParentNode;
  }
//+------------------------------------------------------------------+


//+------------------------------------------------------------------+
//| class declaration                                                |
//+------------------------------------------------------------------+
class CEasyXml
  {
private:
   // Properties
   CEasyXmlNode      *DocumentRoot;
   bool              blDebug;
   bool              blSaveToCache;
   string            sText;
   string            sFilename;
   int               Err;

   // Methods :: parsing
   bool              ParseRecursive(CEasyXmlNode *pActualNode,string &pText,int &pPos,int pLevel=0);
   bool              ParseAttributes(CEasyXmlNode *pActualNode,string pAttributes,string pDebugSpace);
   void              SkipWhitespace(string &pText,int &pPos);
   bool              SkipWhitespaceAndComments(string &pText,int &pPos,string pDebugSpace);
   bool              HasSiblings(string &pText,int &pPos);
   bool              EndOfXml(string &pText,int &pPos);
   bool              SkipProlog(string &pText,int &pPos);

   // Methods :: helpers
   bool              WriteStreamToCacheFile(string pStream);
   bool              Error(int pPos=-1,bool pClear=true);

public:
   // Methods :: parse by source
   bool              LoadXmlFromFile(string pFilename);
   bool              LoadXmlFromString(string pText);

   // Methods :: setters
   void              SetDebugging(bool pDebug);
   bool              SetUrlCacheFile(string pFilename);

   // Methods :: getters
   CEasyXmlNode      *GetDocumentRoot(void);
   string            GetText(void);
   void              Clear(void);

                     CEasyXml();
                    ~CEasyXml();
  };
//+------------------------------------------------------------------+
//| Constructor                                                      |
//+------------------------------------------------------------------+
CEasyXml::CEasyXml()
  {
   blDebug       = false;
   blSaveToCache = false;
   DocumentRoot  = new CEasyXmlNode;
  }
//+------------------------------------------------------------------+
//| Desctructor                                                      |
//+------------------------------------------------------------------+
CEasyXml::~CEasyXml()
  {
   if(CheckPointer(DocumentRoot)==POINTER_DYNAMIC)
      delete DocumentRoot;
  }
//+------------------------------------------------------------------+
//| enable debug output                                              |
//+------------------------------------------------------------------+
void CEasyXml::SetDebugging(bool pDebug)
  {
   blDebug=pDebug;
  }
//+------------------------------------------------------------------+
//| enable url caching                                               |
//+------------------------------------------------------------------+
bool CEasyXml::SetUrlCacheFile(string pFilename)
  {
   if(StringSubstr(pFilename,StringLen(pFilename)-StringLen(EASYXML_XMLFILE_ENDING),StringLen(EASYXML_XMLFILE_ENDING))==EASYXML_XMLFILE_ENDING)
     {
      sFilename     = pFilename;
      blSaveToCache = true;
     }
   else
     {
      Err=EASYXML_INVALID_FILENAME;
      return(Error(-1,false));
     }

   return(true);
  }
//+------------------------------------------------------------------+
//| write stream to cache file                                       |
//+------------------------------------------------------------------+
bool CEasyXml::WriteStreamToCacheFile(string pStream)
  {
   int hFile=FileOpen(sFilename,FILE_BIN|FILE_WRITE);

   if(hFile==INVALID_HANDLE)
     {
      Err = EASYXML_ERR_CONNECTION_FILEOPEN;
      return(false);
     }

   FileWriteString(hFile,pStream);
   FileClose(hFile);

   return(true);
  }
//+------------------------------------------------------------------+
//| Load XML by given file                                           |
//+------------------------------------------------------------------+
bool CEasyXml::LoadXmlFromFile(string pFilename)
  {
   string sStream;
   int    iStringSize;

   int hFile=FileOpen(pFilename,FILE_TXT|FILE_READ|FILE_ANSI);
   if(hFile==INVALID_HANDLE)
     {
      Err = EASYXML_ERR_CONNECTION_FILEOPEN;
      return(Error());
     }

   while(!FileIsEnding(hFile))
     {
      iStringSize = FileReadInteger(hFile, INT_VALUE);
      sStream    += FileReadString(hFile, iStringSize);
     }

   FileClose(hFile);

   return(LoadXmlFromString(sStream));
  }
//+------------------------------------------------------------------+
//| Load XML by given string                                         |
//+------------------------------------------------------------------+
bool CEasyXml::LoadXmlFromString(string pText)
  {
   bool blSuccess = false;
   int  iPos      = 0;
   sText          = pText;

   StringTrimLeft(pText);
   StringTrimRight(pText);

// Skip xml prolog
   blSuccess=SkipProlog(pText,iPos);
   if(!blSuccess)
      return(Error(iPos));

// Parse
   blSuccess=ParseRecursive(DocumentRoot,pText,iPos);
   if(!blSuccess)
      return(Error(iPos));

   return(true);
  }
//+------------------------------------------------------------------+
//| XML recursive parser logic                                       |
//+------------------------------------------------------------------+
bool CEasyXml::ParseRecursive(CEasyXmlNode *pActualNode,string &pText,int &pPos,int pLevel=0)
  {
   bool blSuccess;
   bool blSibling=false;

//---
// At least parse one child element.
// Continue for as long as there are any siblings.
//---

   do
     {

      // Debugging output vars
      string sDebugSpace;
      string sDebugOutput;

      // Indent debug output for better readability
      StringInit(sDebugSpace,pLevel*4,StringGetCharacter(" ",0));

      string sTagName;
      string sCloseTagName;
      string sTagContent;
      string sAttributes;

      // Create sibling node
      if(blSibling)
        {
         pActualNode=pActualNode.CreateSibling(new CEasyXmlNode);
        }

      // Skip comments
      if(!SkipWhitespaceAndComments(pText,pPos,sDebugSpace))
         return(false);

      //---
      // Get start tag. If it contains attributes, parse them seperately.�
      // If tag is self-closing, continue with siblings loop
      //---

      // Open tag
      if(StringFind(EASYXML_START_OPEN,StringSubstr(pText,pPos,1))==0)
        {
         pPos++;
        }
      else
        {
         Err=EASYXML_INVALID_OPENTAG_START;
         return(false);
        }

      // Tag name
      while(StringFind(EASYXML_WHITESPACE,StringSubstr(pText,pPos,1))==-1 &&
            StringCompare(EASYXML_START_CLOSE,StringSubstr(pText,pPos,1))!=0 &&
            !EndOfXml(pText,pPos))
        {
         sTagName+=StringSubstr(pText,pPos,1);
         pPos++;
        }

      pActualNode.SetName(sTagName);

      // Debugging
      if(blDebug)
         Print(sDebugSpace,"<"+sTagName+"> D:",IntegerToString(pLevel)," | P:",pPos);

      SkipWhitespace(pText,pPos);

      // Attributes
      if(StringFind(EASYXML_LATIN,StringSubstr(pText,pPos,1))!=-1)
        {
         while(StringCompare(EASYXML_START_CLOSE,StringSubstr(pText, pPos, 1)) != 0 &&
               StringCompare(EASYXML_SELFCLOSE, StringSubstr(pText, pPos, 2)) != 0 &&
               !EndOfXml(pText,pPos))
           {
            sAttributes+=StringSubstr(pText,pPos,1);
            pPos++;
           }
         blSuccess=ParseAttributes(pActualNode,sAttributes,sDebugSpace);
         if(!blSuccess)
            return(false);
        }

      // Self closing tag
      if(StringCompare(EASYXML_SELFCLOSE,StringSubstr(pText,pPos,2))==0)
        {
         pPos+=2;
         SkipWhitespace(pText,pPos);

         // Detect if next sibling exists
         blSibling=HasSiblings(pText,pPos);

         continue;
        }

      // Start tag close
      if(StringCompare(EASYXML_START_CLOSE,StringSubstr(pText,pPos,1))==0)
        {
         pPos++;
        }
      else
        {
         Err=EASYXML_INVALID_OPENTAG_CLOSE;
         return(false);
        }

      // Skip comments
      if(!SkipWhitespaceAndComments(pText,pPos,sDebugSpace))
         return(false);

      //---
      // Parse next lower level tag and/or read text content
      //---

      // Next level tag
      if(StringCompare(EASYXML_START_OPEN,StringSubstr(pText,pPos,1))==0 &&
         StringCompare(EASYXML_CLOSE_OPEN,StringSubstr(pText,pPos,2))!=0 &&
         StringCompare(EASYXML_CDATA_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_CDATA_OPEN)))!=0)
        {
         // Delve deeper
         pActualNode=pActualNode.CreateChild(new CEasyXmlNode);

         blSuccess=ParseRecursive(pActualNode,pText,pPos,pLevel+1);
         if(!blSuccess)
            return(false);

         pActualNode=pActualNode.Parent();
        }

      // Read text content, even if it follows a closing tag
      if(StringCompare(EASYXML_CDATA_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_CDATA_OPEN)))!=0)
        {
         // Tags in between text won't get parsed as XML nodes
         while(StringCompare(EASYXML_CLOSE_OPEN+sTagName,StringSubstr(pText,pPos,StringLen(EASYXML_CLOSE_OPEN+sTagName)))!=0 &&
               !EndOfXml(pText,pPos))
           {
            sTagContent+=StringSubstr(pText,pPos,1);
            pPos++;
           }
         pActualNode.SetValue(sTagContent);
        }
      // Else read CDATA content, if there is any
      else
        {
         int iClose=StringFind(pText,EASYXML_CDATA_CLOSE,pPos+StringLen(EASYXML_CDATA_OPEN));

         if(iClose>0)
           {
            sTagContent = StringSubstr(pText, pPos + StringLen(EASYXML_CDATA_OPEN), (iClose - pPos - StringLen(EASYXML_CDATA_OPEN)));
            pPos        = iClose + StringLen(EASYXML_CDATA_CLOSE);
           }
         else
           {
            Err=EASYXML_INVALID_CDATA;
            return(false);
           }
         pActualNode.SetValue(sTagContent);
        }

      // Debugging
      if(blDebug && StringLen(sTagContent)!=0)
        {
         sDebugOutput=sTagContent;
         StringTrimLeft(sDebugOutput);
         sDebugOutput=(StringLen(sDebugOutput)>=50) ? StringSubstr(sDebugOutput,0,50)+"..." : sDebugOutput;
         Print(sDebugSpace,"  ### Content ###    "+sDebugOutput);
        }

      SkipWhitespace(pText,pPos);

      //---
      // Get end tag and compare it to start tag. return to upper level if valid
      //---

      if(StringFind(EASYXML_CLOSE_OPEN,StringSubstr(pText,pPos,2))==0)
        {
         pPos+=2;
        }
      else
        {
         Err=EASYXML_NO_CLOSETAG;
         return(false);
        }

      //read end tag name
      while(StringFind(EASYXML_CLOSE_CLOSE,StringSubstr(pText,pPos,1))==-1 && !EndOfXml(pText,pPos))
        {
         sCloseTagName+=StringSubstr(pText,pPos,1);
         pPos++;
        }

      if(blDebug)
         Print(sDebugSpace,"</",sCloseTagName,"> D:",IntegerToString(pLevel)," | P:",pPos);

      //compare start and end tag names
      if(StringCompare(sCloseTagName,sTagName,false)==0)
        {
         pPos++;
        }
      else
        {
         Err=EASYXML_INVALID_CLOSETAG;
         return(false);
        }

      // Skip comments
      if(!SkipWhitespaceAndComments(pText,pPos,sDebugSpace))
         return(false);

      // Detect if next sibling exists
      blSibling=HasSiblings(pText,pPos);

     }
   while(blSibling==true);

//return to upper level
   return(true);
  }
//+------------------------------------------------------------------+
//| pares attributes                                                 |
//+------------------------------------------------------------------+
bool CEasyXml::ParseAttributes(CEasyXmlNode *pActualNode,string pAttributes,string pDebugSpace)
  {
   int iAttrPos        = 0;
   int iValidAttrStart = 0;
   int iValidAttrEnd   = 0;

   string sDebugOutput;

   while(!EndOfXml(pAttributes,iAttrPos))
     {
      string sAttributeName;
      string sAttributeValue;
      string sAttributeValueColon;

      // Some wellformed validity test
      if(StringFind(EASYXML_LATIN,StringSubstr(pAttributes,iAttrPos,1))!=-1)
        {
         iValidAttrStart++;
        }

      // Read Attributename
      while(StringCompare(EASYXML_ATTRIBUTE_SEPARATOR,StringSubstr(pAttributes,iAttrPos,1))!=0)
        {
         sAttributeName+=StringSubstr(pAttributes,iAttrPos,1);
         iAttrPos++;
        }

      // Skip attribute separator
      if(StringCompare(EASYXML_ATTRIBUTE_SEPARATOR,StringSubstr(pAttributes,iAttrPos,1))==0)
        {
         iAttrPos++;
        }

      // Read attribute value. Store Open Colon and use for further comparison
      if(StringFind(EASYXML_ATTRIBUTE_COLON,StringSubstr(pAttributes,iAttrPos,1))!=-1)
        {
         sAttributeValueColon=StringSubstr(pAttributes,iAttrPos,1);
         iAttrPos++;
        }

      while(StringFind(sAttributeValueColon,StringSubstr(pAttributes,iAttrPos,1))==-1)
        {
         sAttributeValue+=StringSubstr(pAttributes,iAttrPos,1);
         iAttrPos++;
        }

      if(StringFind(sAttributeValueColon,StringSubstr(pAttributes,iAttrPos,1))!=-1)
        {
         iAttrPos++;
         iValidAttrEnd++;
        }

      // If attribute is wellformed, set attribute to node
      if(iValidAttrStart==iValidAttrEnd)
        {
         pActualNode.SetAttribute(sAttributeName,sAttributeValue);
        }
      else
        {
         Err=EASYXML_INVALID_ATTRIBUTE;
         return(false);
        }

      // Debugging
      if(blDebug)
        {
         sDebugOutput += (StringLen(sDebugOutput) != 0) ? " | " : "";
         sDebugOutput += IntegerToString(iValidAttrEnd) + ": " + sAttributeName + " -> " + sAttributeValue;
        }

      SkipWhitespace(pAttributes,iAttrPos);
     }

// Debugging
   if(blDebug)
      Print(pDebugSpace,"  ### Attributes ###    ",sDebugOutput);

   return(true);
  }
//+------------------------------------------------------------------+
//| skip whitespace                                                  |
//+------------------------------------------------------------------+
void CEasyXml::SkipWhitespace(string &pText,int &pPos)
  {
   while(StringFind(EASYXML_WHITESPACE,StringSubstr(pText,pPos,1))!=-1)
     {
      pPos++;
     }
  }
//+------------------------------------------------------------------+
//| check if node has siblings                                       |
//+------------------------------------------------------------------+
bool CEasyXml::HasSiblings(string &pText,int &pPos)
  {
   if(StringFind(EASYXML_START_OPEN,StringSubstr(pText,pPos,1))==0 &&
      StringFind(EASYXML_CLOSE_OPEN,StringSubstr(pText,pPos,2))==-1)
     {
      return true;
     }
   else
     {
      return false;
     }
  }
//+------------------------------------------------------------------+
//| check for end of xml                                             |
//+------------------------------------------------------------------+
bool CEasyXml::EndOfXml(string &pText,int &pPos)
  {
   return !(pPos<StringLen(pText));
  }
//+------------------------------------------------------------------+
//| skip xml prolog                                                  |
//+------------------------------------------------------------------+
bool CEasyXml::SkipProlog(string &pText,int &pPos)
  {
// Skip xml declaration
   if(StringCompare(EASYXML_PROLOG_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_PROLOG_OPEN)))==0)
     {
      int iClose=StringFind(pText,EASYXML_PROLOG_CLOSE,pPos+StringLen(EASYXML_PROLOG_OPEN));

      if(blDebug)
         Print("### Prolog ###    ",StringSubstr(pText,pPos,(iClose-pPos)+StringLen(EASYXML_PROLOG_CLOSE)));

      if(iClose>0)
        {
         pPos=iClose+StringLen(EASYXML_PROLOG_CLOSE);
        }
      else
        {
         Err=EASYXML_INVALID_PROLOG;
         return(false);
        }
     }

// Skip comments
   if(!SkipWhitespaceAndComments(pText,pPos,""))
      return(false);

// Skip doctype
   if(StringCompare(EASYXML_DOCTYPE_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_DOCTYPE_OPEN)))==0)
     {
      int iClose=StringFind(pText,EASYXML_DOCTYPE_CLOSE,pPos+StringLen(EASYXML_DOCTYPE_OPEN));

      if(blDebug)
         Print("### DOCTYPE ###    ",StringSubstr(pText,pPos,(iClose-pPos)+StringLen(EASYXML_DOCTYPE_CLOSE)));

      if(iClose>0)
        {
         pPos=iClose+StringLen(EASYXML_DOCTYPE_CLOSE);
        }
      else
        {
         Err=EASYXML_INVALID_DOCTYPE;
         return(false);
        }
     }

// Skip comments
   if(!SkipWhitespaceAndComments(pText,pPos,""))
      return(false);

   return(true);
  }
//+------------------------------------------------------------------+
//| skip xml comments                                                |
//+------------------------------------------------------------------+
bool CEasyXml::SkipWhitespaceAndComments(string &pText,int &pPos,string pDebugSpace)
  {
   bool blNextComment=false;

// Do while there are consecutive comments
   do
     {
      SkipWhitespace(pText,pPos);

      if(StringCompare(EASYXML_COMMENT_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_COMMENT_OPEN)))==0)
        {
         int iClose=StringFind(pText,EASYXML_COMMENT_CLOSE,pPos+StringLen(EASYXML_COMMENT_OPEN));

         if(blDebug)
            Print(pDebugSpace,"  ### Comment ###    ",StringSubstr(pText,pPos,(iClose-pPos)+StringLen(EASYXML_COMMENT_CLOSE)));

         if(iClose>0)
           {
            pPos=iClose+StringLen(EASYXML_COMMENT_CLOSE);
           }
         else
           {
            Err=EASYXML_INVALID_COMMENT;
            return(false);
           }
        }

      SkipWhitespace(pText,pPos);

      if(StringCompare(EASYXML_COMMENT_OPEN,StringSubstr(pText,pPos,StringLen(EASYXML_COMMENT_OPEN)))==0)
        {
         blNextComment=true;
        }
      else
        {
         blNextComment=false;
        }

     }
   while(blNextComment==true);

   return(true);
  }
//+------------------------------------------------------------------+
//| error handling                                                   |
//+------------------------------------------------------------------+
bool CEasyXml::Error(int pPos=-1,bool pClear=true)
  {
   if(pClear)
      Clear();

   Print(EasyXmlError(Err,pPos));
   return(false);
  }
//+------------------------------------------------------------------+
//| get paresd document root                                         |
//+------------------------------------------------------------------+
CEasyXmlNode *CEasyXml::GetDocumentRoot(void)
  {
   return DocumentRoot;
  }
//+------------------------------------------------------------------+
//| return unparsed text                                             |
//+------------------------------------------------------------------+
string CEasyXml::GetText(void)
  {
   return sText;
  }
//+------------------------------------------------------------------+
//| Clear the doc tree                                               |
//+------------------------------------------------------------------+
void CEasyXml::Clear(void)
  {
   StringInit(sText);
   if(CheckPointer(DocumentRoot)==POINTER_DYNAMIC)
      delete DocumentRoot;

   DocumentRoot=new CEasyXmlNode;
  }
//+------------------------------------------------------------------+



//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class CEasyXmlAttribute : public CObject
  {
private:
   string            sName;
   string            sValue;

public:
   void              SetName(string pName);
   void              SetValue(string pValue);

   string            GetName(void);
   string            GetValue(void);

                     CEasyXmlAttribute();
                    ~CEasyXmlAttribute();
  };
//+------------------------------------------------------------------+
//| constructor                                                      |
//+------------------------------------------------------------------+
CEasyXmlAttribute::CEasyXmlAttribute()
  {
  }
//+------------------------------------------------------------------+
//| destructor                                                       |
//+------------------------------------------------------------------+
CEasyXmlAttribute::~CEasyXmlAttribute()
  {
  }
//+------------------------------------------------------------------+
//| set attribute name                                               |
//+------------------------------------------------------------------+
void CEasyXmlAttribute::SetName(string pName)
  {
   sName=pName;
  }
//+------------------------------------------------------------------+
//| set attribute value                                              |
//+------------------------------------------------------------------+
void CEasyXmlAttribute::SetValue(string pValue)
  {
   sValue=pValue;
  }
//+------------------------------------------------------------------+
//| get attribute name                                               |
//+------------------------------------------------------------------+
string CEasyXmlAttribute::GetName(void)
  {
   return sName;
  }
//+------------------------------------------------------------------+
//| get attribute value                                              |
//+------------------------------------------------------------------+
string CEasyXmlAttribute::GetValue(void)
  {
   return sValue;
  }
//+------------------------------------------------------------------+


//+------------------------------------------------------------------+
//| returns trade server return code description                     |
//+------------------------------------------------------------------+
string EasyXmlError(int return_code,int pPos=-1)
  {
   string sErrorDescription;

   switch(return_code)
     {
      case EASYXML_ERR_CONNECTION_ATTEMPT:
         sErrorDescription = "Error in call of InternetAttemptConnect()";
         break;
      case EASYXML_ERR_CONNECTION_OPEN:
         sErrorDescription = "Error in call of InternetOpenW()";
         break;
      case EASYXML_ERR_CONNECTION_URL:
         sErrorDescription = "Error in call of InternetOpenUrlW()";
         break;
      case EASYXML_ERR_CONNECTION_FILEOPEN:
         sErrorDescription = "Error in call of FileOpen()";
         break;
      case EASYXML_ERR_CONNECTION_EMPTYSTREAM:
         sErrorDescription = "No return data from URL";
         break;
      case EASYXML_INVALID_PROLOG:
         sErrorDescription = "Invalid Prologue";
         break;
      case EASYXML_INVALID_COMMENT:
         sErrorDescription = "Invalid Comment";
         break;
      case EASYXML_INVALID_OPENTAG_START:
         sErrorDescription = "Invalid Character found. Should be the beginning of an open tag";
         break;
      case EASYXML_INVALID_OPENTAG_CLOSE:
         sErrorDescription = "Invalid Character found. Should be the close of an open tag";
         break;
      case EASYXML_INVALID_CLOSETAG:
         sErrorDescription = "Invalid Close Tag. Tags must match!";
         break;
      case EASYXML_NO_CLOSETAG:
         sErrorDescription = "No Close Tag found";
         break;
      case EASYXML_INVALID_CDATA:
         sErrorDescription = "Invalid CDATA";
         break;
      case EASYXML_INVALID_ATTRIBUTE:
         sErrorDescription = "Invalid Attribute";
         break;
      case EASYXML_INVALID_FILENAME:
         sErrorDescription = "Invalid Cach File Name";
         break;
      case EASYXML_INVALID_DOCTYPE:
         sErrorDescription = "Invalid Doctype";
         break;
     }

   if(pPos!=-1)
     {
      sErrorDescription+=" P:"+IntegerToString(pPos);
     }

   return("*** "+sErrorDescription+" ***");
  }
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class StrategyManager
  {
   SameDirSignalAction ParseSameDirSignalAction(string value);
   OppositeDirSignalAction ParseOppositeDirSignalAction(string value);
public:
   Strategy          *ParseXmlStrategy(string xml);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
SameDirSignalAction StrategyManager::ParseSameDirSignalAction(string value)
  {
   if(value == "Add")
      return SameDirSignalAction_Add;
   if(value == "Loser")
      return SameDirSignalAction_Loser;
   if(value == "Winner")
      return SameDirSignalAction_Winner;
   return SameDirSignalAction_Nothing;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
OppositeDirSignalAction  StrategyManager::ParseOppositeDirSignalAction(string value)
  {
   if(value == "Close")
      return OppositeDirSignalAction_Close;
   if(value == "Reduce")
      return OppositeDirSignalAction_Reduce;
   if(value == "Reverse")
      return OppositeDirSignalAction_Reverse;
   return OppositeDirSignalAction_Nothing;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
Strategy *StrategyManager::ParseXmlStrategy(string xml)
  {
   IndicatorManager indicatorManager;

   CEasyXml doc;
   doc.SetDebugging(false);
   bool isLoaded=doc.LoadXmlFromString(xml);

   if(!isLoaded)
     {
      string message="Cannot parse the strategy XML file";
      Comment(message);
      Print(message);
      return NULL;
     }

   CArrayObj *rootChildren=doc.GetDocumentRoot().Children();

   int openFilters  = 0;
   int closeFilters = 0;

   for(int i=0; i<rootChildren.Total(); i++)
     {
      CEasyXmlNode *node=rootChildren.At(i);
      string name  = node.GetName();
      string value = node.GetValue();

      if(name=="openFilters")
         openFilters=(int) StringToInteger(value);
      else
         if(name=="closeFilters")
            closeFilters=(int) StringToInteger(value);
     }

   Strategy *strategy=new Strategy(openFilters,closeFilters);

   for(int i=0; i<rootChildren.Total(); i++)
     {
      CEasyXmlNode *node=rootChildren.At(i);
      string name  = node.GetName();
      string value = node.GetValue();

      // Skip unused tags;
      string unusedTags[]=
        {
         "dataSourceName","profileName","AccountBalance","ProfitPerDay","WinLossRatio","AccountStatsParam",
         "instrumentSymbol","instrumentPeriod","AccountStatsValue","MarketStatsParam","MarketStatsValue","InputParametersParam",
         "InputParametersValue","BalanceLine","EquityLine","slot"
        };
      for(int j=0; j<ArraySize(unusedTags); j++)
         if(name==unusedTags[j])
            continue;

      if(name=="strategyName")
         strategy.StrategyName=value;
      else
         if(name == "sameDirSignalAction")
            strategy.SameSignalAction       = ParseSameDirSignalAction(value);
         else
            if(name == "oppDirSignalAction")
               strategy.OppSignalAction        = ParseOppositeDirSignalAction(value);
            else
               if(name == "maxOpenLots")
                  strategy.MaxOpenLots            = StringToDouble(value);
               else
                  if(name == "useAccountPercentEntry")
                     strategy.UseAccountPercentEntry = StringBoolToBool(value);
                  else
                     if(name == "entryLots")
                        strategy.EntryLots              = StringToDouble(value);
                     else
                        if(name == "addingLots")
                           strategy.AddingLots             = StringToDouble(value);
                        else
                           if(name == "reducingLots")
                              strategy.ReducingLots           = StringToDouble(value);
                           else
                              if(name == "useMartingale")
                                 strategy.UseMartingale          = StringBoolToBool(value);
                              else
                                 if(name == "martingaleMultiplier")
                                    strategy.MartingaleMultiplier   = StringToDouble(value);
                                 else
                                    if(name == "description")
                                       strategy.Description            = value;
                                    else
                                       if(name == "recommendedBars")
                                          strategy.RecommendedBars        = (int) StringToInteger(value);
                                       else
                                          if(name == "firstBar")
                                             strategy.FirstBar               = (int) StringToInteger(value);
                                          else
                                             if(name == "minBarsRequired")
                                                strategy.MinBarsRequired        = (int) StringToInteger(value);
                                             else
                                                if(name == "permanentStopLoss")
                                                  {
                                                   strategy.PermanentSL=(int) StringToInteger(value);
                                                   string usePermanentSL=node.GetAttribute("usePermanentSL");
                                                   strategy.UsePermanentSL=StringBoolToBool(usePermanentSL);
                                                   strategy.PermanentSLType=(PermanentProtectionType)StringToInteger(node.GetAttribute("permanentSLType"));
                                                  }
                                                else
                                                   if(name=="permanentTakeProfit")
                                                     {
                                                      strategy.PermanentTP=(int) StringToInteger(value);
                                                      string usePermanentTP=node.GetAttribute("usePermanentTP");
                                                      strategy.UsePermanentTP=StringBoolToBool(usePermanentTP);
                                                      strategy.PermanentTPType=(PermanentProtectionType)StringToInteger(node.GetAttribute("permanentTPType"));
                                                     }
                                                   else
                                                      if(name=="breakEven")
                                                        {
                                                         strategy.BreakEven=(int) StringToInteger(value);
                                                         string useBreakEven=node.GetAttribute("useBreakEven");
                                                         strategy.UseBreakEven=StringBoolToBool(useBreakEven);
                                                        }
     }

   for(int i=0; i<rootChildren.Total(); i++)
     {
      CEasyXmlNode *node=rootChildren.At(i);

      if(node.GetName()!="slot")
         continue;

      string slotNumberText=node.GetAttribute("slotNumber");
      int slot=(int) StringToInteger(slotNumberText);

      string slotTypeText= node.GetAttribute("slotType");
      SlotTypes slotType = SlotTypeFromShortString(slotTypeText);

      string logicalGroup = node.GetAttribute("logicalGroup");

      string indicatorName;
      int signalShift  = 0;
      int signalRepeat = 0;
      DataPeriod indicatorPeriod=DataPeriod_M1;
      string indicatorSymbol="";

      for(int r=0; r<node.Children().Total(); r++)
        {
         CEasyXmlNode *child=node.Children().At(r);
         string name  = child.GetName();
         string value = child.GetValue();

         if(name=="indicatorName")
            indicatorName=value;
         else
            if(name == "signalShift")
               signalShift     = (int) StringToInteger(value);
            else
               if(name == "signalRepeat")
                  signalRepeat    = (int) StringToInteger(value);
               else
                  if(name == "indicatorPeriod")
                     indicatorPeriod = StringToDataPeriod(value);
                  else
                     if(name == "indicatorSymbol")
                        indicatorSymbol = StringRemoveWhite(value);
        }

      strategy.Slot[slot].IndicatorName    = indicatorName;
      strategy.Slot[slot].IndicatorPointer = indicatorManager.CreateIndicator(indicatorName, slotType);
      strategy.Slot[slot].SlotType         = slotType;
      strategy.Slot[slot].SignalShift      = signalShift;
      strategy.Slot[slot].SignalRepeat     = signalRepeat;
      strategy.Slot[slot].IndicatorPeriod  = indicatorPeriod;
      strategy.Slot[slot].IndicatorSymbol  = indicatorSymbol;
      strategy.Slot[slot].LogicalGroup     = logicalGroup;

      for(int j=1; j<node.Children().Total(); j++)
        {
         CEasyXmlNode* subNodes = node.Children().At(j);
         string paramNumberText = subNodes.GetAttribute("paramNumber");
         int paramNumber=(int) StringToInteger(paramNumberText);

         if(subNodes.GetName()=="listParam")
           {
            strategy.Slot[slot].IndicatorPointer.ListParam[paramNumber].Enabled=true;

            CEasyXmlNode *param0=subNodes.Children().At(0);
            string text0=param0.GetValue();
            strategy.Slot[slot].IndicatorPointer.ListParam[paramNumber].Caption=text0;

            CEasyXmlNode *param1=subNodes.Children().At(1);
            string text1=param1.GetValue();
            strategy.Slot[slot].IndicatorPointer.ListParam[paramNumber].Index=(int) StringToInteger(text1);

            CEasyXmlNode *param2=subNodes.Children().At(2);
            string text2=param2.GetValue();
            strategy.Slot[slot].IndicatorPointer.ListParam[paramNumber].Text=text2;
           }
         else
            if(subNodes.GetName()=="numParam")
              {
               strategy.Slot[slot].IndicatorPointer.NumParam[paramNumber].Enabled=true;

               CEasyXmlNode *param0=subNodes.Children().At(0);
               string text0=param0.GetValue();
               strategy.Slot[slot].IndicatorPointer.NumParam[paramNumber].Caption=text0;

               CEasyXmlNode *param1=subNodes.Children().At(1);
               string text1=param1.GetValue();
               strategy.Slot[slot].IndicatorPointer.NumParam[paramNumber].Value=StringToDouble(text1);
              }
            else
               if(subNodes.GetName()=="checkParam")
                 {
                  strategy.Slot[slot].IndicatorPointer.CheckParam[paramNumber].Enabled=true;

                  CEasyXmlNode *param0=subNodes.Children().At(0);
                  string text0=param0.GetValue();
                  strategy.Slot[slot].IndicatorPointer.CheckParam[paramNumber].Caption=text0;

                  CEasyXmlNode *param1=subNodes.Children().At(1);
                  string text1=param1.GetValue();
                  strategy.Slot[slot].IndicatorPointer.CheckParam[paramNumber].Checked=StringBoolToBool(text1);
                 }
        }
     }

   return strategy;
  }
//+------------------------------------------------------------------+

//## Import Start

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class Logger
  {
   int               m_LogLines;
   int               m_FileHandle;

public:
   string            GetLogFileName(string symbol,int dataPeriod,int expertMagic);
   int               CreateLogFile(string fileName);
   void              WriteLogLine(string text);
   void              WriteNewLogLine(string text);
   void              WriteLogRequest(string text,string request);
   bool              IsLogLinesLimitReached(int maxLines);
   void              FlushLogFile(void);
   void              CloseLogFile(void);
   int               CloseExpert(void);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string Logger::GetLogFileName(string symbol,int dataPeriod,int expertMagic)
  {
   string time=TimeToString(TimeCurrent(),TIME_DATE|TIME_SECONDS);
   StringReplace(time,":","");
   StringReplace(time," ","_");
   string rnd=IntegerToString(MathRand());
   string fileName=symbol+"_"+IntegerToString(dataPeriod)+"_"+IntegerToString(expertMagic)+"_"+time+"_"+rnd+".log";
   return (fileName);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int Logger::CreateLogFile(string fileName)
  {
   m_LogLines = 0;
   int handle = FileOpen(fileName, FILE_CSV|FILE_WRITE, ",");
   if(handle>0)
      m_FileHandle=handle;
   else
      Print("CreateFile: Error while creating log file!");
   return (handle);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Logger::WriteLogLine(string text)
  {
   if(m_FileHandle <= 0)
      return;
   FileWrite(m_FileHandle,TimeToString(TimeCurrent(),TIME_DATE|TIME_SECONDS),text);
   m_LogLines++;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Logger::WriteNewLogLine(string text)
  {
   if(m_FileHandle <= 0)
      return;
   FileWrite(m_FileHandle,"");
   FileWrite(m_FileHandle,TimeToString(TimeCurrent(),TIME_DATE|TIME_SECONDS),text);
   m_LogLines+=2;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Logger::WriteLogRequest(string text,string request)
  {
   if(m_FileHandle <= 0)
      return;
   FileWrite(m_FileHandle,"\n"+text);
   FileWrite(m_FileHandle,TimeToString(TimeCurrent(),TIME_DATE|TIME_SECONDS),request);
   m_LogLines+=3;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Logger::FlushLogFile()
  {
   if(m_FileHandle <= 0)
      return;
   FileFlush(m_FileHandle);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void Logger::CloseLogFile()
  {
   if(m_FileHandle <= 0)
      return;
   WriteNewLogLine(StringFormat("%s Closed.",MQLInfoString(MQL_PROGRAM_NAME)));
   FileClose(m_FileHandle);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool Logger::IsLogLinesLimitReached(int maxLines)
  {
   return (m_LogLines>maxLines);
  }
//+------------------------------------------------------------------+

//## Import Start

#define TRADE_RETRY_COUNT 4
#define TRADE_RETRY_WAIT  100
#define OP_FLAT          -1

const double Epsilon=0.000001;
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
class ActionTrade5
  {
   // Fields
   Strategy          *m_Strategy;
   DataSet           *m_DataSet[];
   DataMarket        *m_DataMarket;
   Logger            *m_Logger;

   // Aggregate position
   int               m_PositionTicket;
   int               m_PositionType;
   datetime          m_PositionTime;
   double            m_PositionLots;
   double            m_PositionOpenPrice;
   double            m_PositionStopLoss;
   double            m_PositionTakeProfit;
   double            m_PositionProfit;
   double            m_PositionCommission;
   string            m_PositionComment;
   int               m_ConsecutiveLosses;

   // Properties
   int               m_LastError;
   double            m_PipsValue;
   int               m_PipsPoint;
   int               m_StopLevel;
   datetime          m_BarTime;
   datetime          m_TickTime;
   datetime          m_BarHighTime;
   datetime          m_BarLowTime;
   double            m_CurrentBarHigh;
   double            m_CurrentBarLow;
   int               m_TrailingStop;
   TrailingStopMode  m_TrailingMode;
   int               m_BreakEven;
   int               m_NBarExit;
   int               m_Digits;

   StrategyPriceType m_OpenStrPriceType;
   StrategyPriceType m_CloseStrPriceType;

   ExecutionTime     m_OpenTimeExec;
   ExecutionTime     m_CloseTimeExec;
   bool              m_UseLogicalGroups;
   DictStringBool    *m_GroupsAllowLong;
   DictStringBool    *m_GroupsAllowShort;
   ListString        *m_OpeningLogicGroups;
   ListString        *m_ClosingLogicGroups;
   bool              m_IsEnteredLong;
   bool              m_IsEnteredShort;
   datetime          m_TimeLastEntryBar;
   datetime          m_BarOpenTimeForLastCloseTick;
   datetime          m_BarOpenTimeForLastCloseEvent;

   string            m_DynamicInfoParams[];
   string            m_DynamicInfoValues[];
   string            m_DynamicInfoAccount[3];

   // Methods
   bool              CheckEnvironment(int minimumBars);
   bool              CheckChartBarsCount(int minimumBars);
   int               FindBarsCountNeeded(void);
   int               SetAggregatePosition(void);
   string            AggregatePositionToString(void);
   void              AggregatePositionToNormalString(string &posinfo[]);
   void              UpdateDataSet(DataSet *dataSet,int maxBars);
   void              UpdateDataMarket(DataMarket *dataMarket);
   bool              IsTradeContextFree(void);

   // Trading methods
   bool              ManageOrderSend(int type,double lots,double stoploss,double takeprofit);
   bool              CloseCurrentPosition();
   bool              ModifyPosition(double stoploss,double takeprofit);
   double            GetTakeProfitPrice(int type,double takeprofit);
   double            GetStopLossPrice(int type,double stoploss);
   double            CorrectTakeProfitPrice(int type,double takeprofit);
   double            CorrectStopLossPrice(int type,double stoploss);
   double            NormalizeEntrySize(double size);
   double            NormalizeEntryPrice(double price);
   void              SetMaxStopLoss(void);
   void              SetBreakEvenStop(void);
   void              SetTrailingStop(bool isNewBar);
   void              SetTrailingStopBarMode(void);
   void              SetTrailingStopTickMode(void);
   void              DetectSLTPActivation(void);

   // Strategy Trader methods
   TickType          GetTickType(DataSet *dataSet,bool isNewBar);
   void              CalculateTrade(TickType ticktype);
   PosDirection      GetNewPositionDirection(OrderDirection ordDir,double ordLots,PosDirection posDir,double posLots);
   TradeDirection    AnalyzeEntryPrice(void);
   TradeDirection    AnalyzeEntryDirection(void);
   void              AnalyzeEntryLogicConditions(int bar,string group,double buyPrice,double sellPrice,bool &canOpenLong,bool &canOpenShort);
   double            AnalyzeEntrySize(OrderDirection ordDir,PosDirection &newPosDir);
   TradeDirection    AnalyzeExitPrice(void);
   double            TradingSize(double size);
   double            AccountPercentStopPoints(double percent,double lots);
   TradeDirection    AnalyzeExitDirection(void);
   TradeDirection    ReduceDirectionStatus(TradeDirection baseDirection,TradeDirection direction);
   TradeDirection    IncreaseDirectionStatus(TradeDirection baseDirection,TradeDirection direction);
   TradeDirection    GetClosingDirection(TradeDirection baseDirection,IndComponentType compDataType);
   double            GetStopLossPoints(double lots);
   double            GetTakeProfitPoints(void);
   void              DoEntryTrade(TradeDirection tradeDir);
   bool              DoExitTrade(void);
   bool              IsWrongStopsExecution(void);
   void              ResendWrongStops(void);
   void              InitTrade(void);

public:
   // Constructors
                     ActionTrade5(void);
                    ~ActionTrade5(void);

   // Properties
   double            Entry_Amount;
   double            Maximum_Amount;
   double            Adding_Amount;
   double            Reducing_Amount;

   string            Strategy_File_Name;
   string            Strategy_XML;
   int               Max_Data_Bars;
   int               Protection_Min_Account;
   int               Protection_Max_StopLoss;
   bool              Separate_SL_TP;
   bool              Write_Log_File;
   bool              FIFO_order;
   int               TrailingStop_Moving_Step;
   int               MaxLogLinesInFile;
   int               Bar_Close_Advance;

   // Methods
   int               OnInit(void);
   void              OnTick(void);
   void              OnTrade(void);
   void              OnDeinit(const int reason);
  };
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::ActionTrade5(void)
  {
   m_Logger = new Logger();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::~ActionTrade5(void)
  {
   if(CheckPointer(m_Logger)==POINTER_DYNAMIC)
      delete m_Logger;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int ActionTrade5::OnInit()
  {
   m_DataMarket         = new DataMarket();
   m_GroupsAllowLong    = new DictStringBool();
   m_GroupsAllowShort   = new DictStringBool();
   m_OpeningLogicGroups = new ListString();
   m_ClosingLogicGroups = new ListString();

   m_BarHighTime    = 0;
   m_BarLowTime     = 0;
   m_CurrentBarHigh = 0;
   m_CurrentBarLow  = 1000000;
   m_Digits         = Digits();

   DataPeriod dataPeriod=(DataPeriod)Period();

   string message=StringFormat("%s loaded.",MQLInfoString(MQL_PROGRAM_NAME));
   Comment(message);
   Print(message);

   if(Write_Log_File)
     {
      m_Logger.CreateLogFile(m_Logger.GetLogFileName(_Symbol,_Period,5));
      m_Logger.WriteLogLine(message);
      m_Logger.WriteLogLine("Entry Amount: "   +DoubleToString(Entry_Amount,2)  +", "+
                            "Maximum Amount: " +DoubleToString(Maximum_Amount,2)+", "+
                            "Adding Amount: "  +DoubleToString(Adding_Amount,2) +", "+
                            "Reducing Amount: "+DoubleToString(Reducing_Amount,2));
      m_Logger.WriteLogLine("Protection Min Account: " +IntegerToString(Protection_Min_Account)+", "+
                            "Protection Max StopLoss: "+IntegerToString(Protection_Max_StopLoss));
      m_Logger.WriteLogLine("Bar Close Advance: "+IntegerToString(Bar_Close_Advance));
      m_Logger.FlushLogFile();
     }

   if(_Digits==2 || _Digits==3)
      m_PipsValue=0.01;
   else
      if(_Digits==4 || _Digits==5)
         m_PipsValue=0.0001;
      else
         m_PipsValue=_Digits;

   if(_Digits==3 || _Digits==5)
      m_PipsPoint=10;
   else
      m_PipsPoint=1;

   m_StopLevel=(int)SymbolInfoInteger(_Symbol,SYMBOL_TRADE_STOPS_LEVEL)+m_PipsPoint;
   if(m_StopLevel<3*m_PipsPoint)
      m_StopLevel=3*m_PipsPoint;

   if(Protection_Max_StopLoss>0 && Protection_Max_StopLoss<m_StopLevel)
      Protection_Max_StopLoss=m_StopLevel;

   if(TrailingStop_Moving_Step<m_PipsPoint)
      TrailingStop_Moving_Step=m_PipsPoint;

   string xml=(Strategy_XML=="##STRATEGY##")
              ? LoadStringFromFile(Strategy_File_Name)
              : Strategy_XML;

   StrategyManager *strategyManager=new StrategyManager();

// Strategy initialization
   m_Strategy=strategyManager.ParseXmlStrategy(xml);
   m_Strategy.SetSymbol(_Symbol);
   m_Strategy.SetPeriod(dataPeriod);
   m_Strategy.EntryLots    = Entry_Amount;
   m_Strategy.MaxOpenLots  = Maximum_Amount;
   m_Strategy.AddingLots   = Adding_Amount;
   m_Strategy.ReducingLots = Reducing_Amount;
   m_Strategy.SetIsTester(MQLInfoInteger(MQL_TESTER));

   delete strategyManager;

// Checks the requirements.
   bool isEnvironmentGood=CheckEnvironment(m_Strategy.MinBarsRequired);
   if(!isEnvironmentGood)
     {
      // There is a non fulfilled condition, therefore we must exit.
      Sleep(20*1000);
      ExpertRemove();
      return (INIT_FAILED);
     }

// Market initialization
   string charts[];
   m_Strategy.GetRequiredCharts(charts);

   string chartsNote = "Loading data: ";
   for(int i=0; i<ArraySize(charts); i++)
      chartsNote += charts[i] + ", ";
   chartsNote += "Minumum bars: " + IntegerToString(m_Strategy.MinBarsRequired) + "...";
   Comment(chartsNote);
   Print(chartsNote);

// Initial data loading
   ArrayResize(m_DataSet,ArraySize(charts));
   for(int i=0; i<ArraySize(charts); i++)
      m_DataSet[i]=new DataSet(charts[i]);

   SetAggregatePosition();

// Checks the necessary bars.
   Max_Data_Bars=FindBarsCountNeeded();

// Initial strategy calculation
   for(int i=0; i<ArraySize(m_DataSet); i++)
      UpdateDataSet(m_DataSet[i],Max_Data_Bars);
   m_Strategy.CalculateStrategy(m_DataSet);

   InitTrade();

   m_Strategy.DynamicInfoInitArrays(m_DynamicInfoParams,m_DynamicInfoValues);
   int paramsX=0;
   int valuesX=140;
   int locationY=40;
   color foreColor=GetChartForeColor(0);
   int count= ArraySize(m_DynamicInfoParams);
   for(int i=0; i<count; i++)
     {
      string namep = "Lbl_prm_"+IntegerToString(i);
      string namev = "Lbl_val_"+IntegerToString(i);
      string param = m_DynamicInfoParams[i] == "" ? "." : m_DynamicInfoParams[i];
      LabelCreate(0,namep,0,paramsX,locationY,CORNER_LEFT_UPPER,param,"Ariel",8,foreColor);
      LabelCreate(0,namev,0,valuesX,locationY,CORNER_LEFT_UPPER,".","Ariel",8,foreColor);
      locationY+=12;
     }

   LabelCreate(0,"Lbl_pos_0",0,350,0,CORNER_LEFT_UPPER,".","Ariel",10,foreColor);
   LabelCreate(0,"Lbl_pos_1",0,350,15,CORNER_LEFT_UPPER,".","Ariel",10,foreColor);
   LabelCreate(0,"Lbl_pos_2",0,350,29,CORNER_LEFT_UPPER,".","Ariel",10,foreColor);

   Comment("");

   return(INIT_SUCCEEDED);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::OnTick()
  {
   for(int i=0; i<ArraySize(m_DataSet); i++)
      UpdateDataSet(m_DataSet[i],Max_Data_Bars);
   UpdateDataMarket(m_DataMarket);

   datetime barTime=Time(_Symbol,_Period,0);
   bool isNewBar=(m_BarTime<barTime&&m_DataMarket.Volume<5);
   m_BarTime=barTime;
   m_TickTime=m_DataMarket.TickServerTime;
   m_LastError=0;

// Checks if minimum account was reached.
   if(Protection_Min_Account>0 && AccountInfoDouble(ACCOUNT_EQUITY)<Protection_Min_Account)
     {
      if(m_PositionLots>Epsilon)
         CloseCurrentPosition();
      return;
     }

// Checks and sets Max SL protection.
   if(Protection_Max_StopLoss>0)
      SetMaxStopLoss();

// Checks if position was closed.
   DetectSLTPActivation();

   if(m_BreakEven>0)
      SetBreakEvenStop();

   if(m_TrailingStop>0)
      SetTrailingStop(isNewBar);

   SetAggregatePosition();

   if(isNewBar && Write_Log_File)
      m_Logger.WriteNewLogLine(AggregatePositionToString());

   if(m_DataSet[0].Bars >= m_Strategy.MinBarsRequired)
     {
      m_Strategy.CalculateStrategy(m_DataSet);
      TickType tickType=GetTickType(m_DataSet[0],isNewBar);
      CalculateTrade(tickType);
     }

// Sends OrderModify on SL/TP errors
   if(IsWrongStopsExecution())
      ResendWrongStops();

   AggregatePositionToNormalString(m_DynamicInfoAccount);
   for(int i=0; i<3; i++)
     {
      string lblName="Lbl_pos_"+IntegerToString(i);
      string val=m_DynamicInfoAccount[i]=="" ? "." : m_DynamicInfoAccount[i];
      LabelTextChange(0,lblName,val);
     }

   m_Strategy.DynamicInfoSetValues(m_DynamicInfoValues);
   int count= ArraySize(m_DynamicInfoValues);
   for(int i=0; i<count; i++)
     {
      string namev="Lbl_val_"+IntegerToString(i);
      string val=m_DynamicInfoValues[i]=="" ? "." : m_DynamicInfoValues[i];
      LabelTextChange(0,namev,val);
     }
   ChartRedraw();

   if(Write_Log_File)
     {
      if(m_Logger.IsLogLinesLimitReached(MaxLogLinesInFile))
        {
         m_Logger.CloseLogFile();
         m_Logger.CreateLogFile(m_Logger.GetLogFileName(_Symbol,_Period,5));
        }
      m_Logger.FlushLogFile();
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::OnTrade(void)
  {
   datetime now = TimeCurrent();
   HistorySelect(now-30*24*60*60, now);
   int total = (int)HistoryDealsTotal();
   ulong ticket=0;
   int maxLosses=0;
   for(int i=total; i>=0; i--)
     {
      if((ticket=HistoryDealGetTicket(i))>0)
        {
         double price =HistoryDealGetDouble(ticket, DEAL_PRICE);
         long   time  =HistoryDealGetInteger(ticket,DEAL_TIME);
         string symbol=HistoryDealGetString(ticket, DEAL_SYMBOL);
         double profit=HistoryDealGetDouble(ticket, DEAL_PROFIT);
         if(price>Epsilon && time>0 && symbol==_Symbol)
           {
            if(profit<-Epsilon)
               maxLosses++;
            if(profit>Epsilon)
              {
               m_ConsecutiveLosses = maxLosses;
               break;
              }
           }
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::OnDeinit(const int reason)
  {
   if(Write_Log_File)
      m_Logger.CloseLogFile();

   if(CheckPointer(m_Strategy)==POINTER_DYNAMIC)
      delete m_Strategy;

   for(int i=0; i<ArraySize(m_DataSet); i++)
      if(CheckPointer(m_DataSet[i])==POINTER_DYNAMIC)
         delete m_DataSet[i];
   ArrayFree(m_DataSet);

   if(CheckPointer(m_DataMarket)==POINTER_DYNAMIC)
      delete m_DataMarket;

   delete m_GroupsAllowLong;
   delete m_GroupsAllowShort;
   delete m_OpeningLogicGroups;
   delete m_ClosingLogicGroups;

   int count= ArraySize(m_DynamicInfoParams);
   for(int i=0; i<count; i++)
     {
      string namev = "Lbl_val_"+IntegerToString(i);
      string namep = "Lbl_prm_"+IntegerToString(i);
      LabelDelete(0,namev);
      LabelDelete(0,namep);
     }

   for(int i=0; i<3; i++)
     {
      string lblName="Lbl_pos_"+IntegerToString(i);
      LabelDelete(0,lblName);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::CheckEnvironment(int minimumBars)
  {
// Checks the count of bars available.
   if(!CheckChartBarsCount(minimumBars))
      return (false);

   if(MQLInfoInteger(MQL_TESTER))
     {
      SetAggregatePosition();
      return (true);
     }

// Checks if you are logged in.
   if(AccountNumber()==0)
     {
      Comment("\n You are not logged in. Please login first.");
      for(int attempt=0; attempt<200; attempt++)
        {
         if(AccountNumber()==0)
            Sleep(300);
         else
            break;
        }
      if(AccountNumber()==0)
         return (false);
     }

// Checks the open positions.
   if(SetAggregatePosition()==-1)
      return (false);  // Some error with the current positions.

// Everything looks OK.
   return (true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::CheckChartBarsCount(int barsNecessary)
  {
   int bars=Bars(_Symbol,_Period);
   bool isEnoughBars=bars>=barsNecessary;
   if(isEnoughBars)
      return(true);

   string message="\n Cannot load enough bars! The expert needs minimum "+
                  IntegerToString(barsNecessary)+" bars."+
                  "\n Currently "+IntegerToString(bars)+" bars are loaded.";

   if(MQLInfoInteger(MQL_TESTER))
      message+="\n Please use a custom date period and set the dates properly.";

   Comment(message);
   Print(message);
   return (false);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int ActionTrade5::FindBarsCountNeeded()
  {
   int barStep=50;
   int minBars=50;
   int maxBars=3000;

// Initial state
   int initialBars=MathMax(m_Strategy.MinBarsRequired,minBars);
   initialBars=MathMax(m_Strategy.FirstBar,initialBars);
   for(int i=0; i<ArraySize(m_DataSet); i++)
      UpdateDataSet(m_DataSet[i],initialBars);
   UpdateDataMarket(m_DataMarket);
   double initialBid=m_DataMarket.Bid;
   m_Strategy.CalculateStrategy(m_DataSet);
   string dynamicInfo=m_Strategy.DynamicInfoText();
   int necessaryBars=initialBars;
   int roundedInitialBars=(int)(barStep*MathCeil(((double)initialBars)/barStep));
   int firstTestBars=roundedInitialBars>=initialBars+barStep/2
                     ? roundedInitialBars
                     : roundedInitialBars+barStep;

   for(int bars=firstTestBars; bars<=maxBars; bars+=barStep)
     {
      for(int i=0; i<ArraySize(m_DataSet); i++)
         UpdateDataSet(m_DataSet[i],bars);
      UpdateDataMarket(m_DataMarket);
      m_Strategy.CalculateStrategy(m_DataSet);
      string currentInfo=m_Strategy.DynamicInfoText();

      if(dynamicInfo==currentInfo)
         break;

      dynamicInfo=currentInfo;
      necessaryBars=bars;

      if(MathAbs(initialBid-m_DataMarket.Bid)>Epsilon)
        {
         // Reset the test if new tick has arrived.
         for(int i=0; i<ArraySize(m_DataSet); i++)
            UpdateDataSet(m_DataSet[i],initialBars);
         UpdateDataMarket(m_DataMarket);
         initialBid=m_DataMarket.Bid;
         m_Strategy.CalculateStrategy(m_DataSet);
         dynamicInfo=m_Strategy.DynamicInfoText();
         bars=firstTestBars-barStep;
        }
     }

   string barsMessage = "The expert uses "+IntegerToString(necessaryBars)+" bars.";
   if(Write_Log_File)
     {
      m_Logger.WriteLogLine(barsMessage);
      string timeLastBar=TimeToString(m_DataMarket.TickServerTime,TIME_DATE|TIME_MINUTES);
      m_Logger.WriteLogLine("Indicator values: "+m_DataSet[0].Chart+", Time last bar: "+timeLastBar);
      m_Logger.WriteLogLine(dynamicInfo);
     }
   Print(barsMessage);

   return (necessaryBars);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::UpdateDataSet(DataSet *dataSet,int maxBars)
  {
   string  symbol = dataSet.Symbol;
   int     bars   = MathMin(Bars(symbol, DataPeriodToTimeFrame(dataSet.Period)), maxBars);
   MqlTick tick;
   SymbolInfoTick(symbol,tick);

   dataSet.LotSize        = (int) SymbolInfoDouble(symbol,SYMBOL_TRADE_CONTRACT_SIZE);
   dataSet.Digits         = (int) SymbolInfoInteger(symbol,SYMBOL_DIGITS);
   dataSet.StopLevel      = (int) SymbolInfoInteger(symbol,SYMBOL_TRADE_STOPS_LEVEL);
   dataSet.Point          = SymbolInfoDouble(symbol,SYMBOL_POINT);
   dataSet.TickValue      = SymbolInfoDouble(symbol,SYMBOL_TRADE_TICK_VALUE);
   dataSet.MinLot         = SymbolInfoDouble(symbol,SYMBOL_VOLUME_MIN);
   dataSet.MaxLot         = SymbolInfoDouble(symbol,SYMBOL_VOLUME_MAX);
   dataSet.LotStep        = SymbolInfoDouble(symbol,SYMBOL_VOLUME_STEP);
   dataSet.MarginRequired = SymbolInfoDouble(symbol,SYMBOL_MARGIN_INITIAL);
   dataSet.Bars           = bars;
   dataSet.ServerTime     = tick.time;
   dataSet.Bid            = tick.bid;
   dataSet.Ask            = tick.ask;
   dataSet.Spread         = (tick.ask-tick.bid)/dataSet.Point;

   if(dataSet.MarginRequired < Epsilon)
      bool ok = OrderCalcMargin(ORDER_TYPE_BUY, symbol, 1, tick.ask, dataSet.MarginRequired);
   if(dataSet.MarginRequired < Epsilon)
      dataSet.MarginRequired = tick.bid*dataSet.LotSize/100;

   dataSet.SetPrecision();

   MqlRates rates[];
   ArraySetAsSeries(rates,false);
   int copied=CopyRates(symbol,DataPeriodToTimeFrame(dataSet.Period),0,bars,rates);

   ArrayResize(dataSet.Time,   bars);
   ArrayResize(dataSet.Open,   bars);
   ArrayResize(dataSet.High,   bars);
   ArrayResize(dataSet.Low,    bars);
   ArrayResize(dataSet.Close,  bars);
   ArrayResize(dataSet.Volume, bars);

   for(int i=0; i<bars; i++)
     {
      dataSet.Time[i]   = rates[i].time;
      dataSet.Open[i]   = rates[i].open;
      dataSet.High[i]   = rates[i].high;
      dataSet.Low[i]    = rates[i].low;
      dataSet.Close[i]  = rates[i].close;
      dataSet.Volume[i] = (int) rates[i].tick_volume;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::UpdateDataMarket(DataMarket *dataMarket)
  {
   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   dataMarket.TickLocalTime       = TimeLocal();
   dataMarket.TickServerTime      = tick.time;

   dataMarket.PositionTicket      = m_PositionTicket;
   dataMarket.PositionLots        = m_PositionLots;
   dataMarket.PositionOpenPrice   = m_PositionOpenPrice;
   dataMarket.PositionOpenTime    = m_PositionTime;
   dataMarket.PositionStopLoss    = m_PositionStopLoss;
   dataMarket.PositionTakeProfit  = m_PositionTakeProfit;
   dataMarket.PositionProfit      = m_PositionProfit;
   dataMarket.PositionComment     = m_PositionComment;
   dataMarket.PositionDirection   = m_PositionType==OP_FLAT
                                    ? PosDirection_None
                                    : m_PositionType==OP_BUY
                                    ? PosDirection_Long
                                    : PosDirection_Short;

   dataMarket.AccountBalance      = AccountInfoDouble(ACCOUNT_BALANCE);
   dataMarket.AccountEquity       = AccountInfoDouble(ACCOUNT_EQUITY);
   dataMarket.AccountFreeMargin   = AccountInfoDouble(ACCOUNT_FREEMARGIN);
   dataMarket.ConsecutiveLosses   = m_ConsecutiveLosses;
   dataMarket.IsNewBid            = MathAbs(tick.bid-dataMarket.Bid)>Epsilon;
   dataMarket.OldAsk              = dataMarket.Ask;
   dataMarket.OldBid              = dataMarket.Bid;
   dataMarket.OldClose            = dataMarket.Close;
   dataMarket.Ask                 = tick.ask;
   dataMarket.Bid                 = tick.bid;
   dataMarket.Close               = Close(_Symbol,_Period,0);
   dataMarket.Volume              = Volume(_Symbol,_Period,0);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int ActionTrade5::SetAggregatePosition()
  {
   m_PositionTicket     = 0;
   m_PositionType       = OP_FLAT;
   m_PositionTime       = D'2050.01.01 00:00';
   m_PositionLots       = 0;
   m_PositionOpenPrice  = 0;
   m_PositionStopLoss   = 0;
   m_PositionTakeProfit = 0;
   m_PositionProfit     = 0;
   m_PositionCommission = 0;
   m_PositionComment    = "";

   if(!PositionSelect(_Symbol))
      return (0);

   m_PositionTicket      = (int) PositionGetInteger(POSITION_IDENTIFIER);
   m_PositionType        = (int) PositionGetInteger(POSITION_TYPE);
   m_PositionTime        = (datetime) MathMax(PositionGetInteger(POSITION_TIME),
                           PositionGetInteger(POSITION_TIME_UPDATE));
   m_PositionOpenPrice   = NormalizeDouble(PositionGetDouble(POSITION_PRICE_OPEN),_Digits);
   m_PositionLots        = NormalizeDouble(PositionGetDouble(POSITION_VOLUME),2);
   m_PositionProfit      = NormalizeDouble(PositionGetDouble(POSITION_PROFIT)+
                                           PositionGetDouble(POSITION_COMMISSION),2);
   m_PositionCommission  = NormalizeDouble(PositionGetDouble(POSITION_COMMISSION),2);
   m_PositionStopLoss    = NormalizeDouble(PositionGetDouble(POSITION_SL),_Digits);
   m_PositionTakeProfit  = NormalizeDouble(PositionGetDouble(POSITION_TP),_Digits);
   m_PositionComment     = PositionGetString(POSITION_COMMENT);

   return (1);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
string ActionTrade5::AggregatePositionToString()
  {
   if(m_PositionType==OP_FLAT)
      return ("Position: Flat");

   string text="Position: "    +
               "Ticket="       + IntegerToString(m_PositionTicket)+
               ", Time="       + TimeToString(m_PositionTime,TIME_SECONDS)+
               ", Type="       + (m_PositionType==OP_BUY ? "Long" : "Short")+
               ", Lots="       + DoubleToString(m_PositionLots,2)+
               ", Price="      + DoubleToString(m_PositionOpenPrice,_Digits)+
               ", StopLoss="   + DoubleToString(m_PositionStopLoss,_Digits)+
               ", TakeProfit=" + DoubleToString(m_PositionTakeProfit,_Digits)+
               ", Commission=" + DoubleToString(m_PositionCommission,2)+
               ", Profit="     + DoubleToString(m_PositionProfit,2);

   if(m_PositionComment!="")
      text=text+", \""+m_PositionComment+"\"";

   return (text);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::AggregatePositionToNormalString(string &posinfo[])
  {
   posinfo[0]=StringFormat("%s Balance: %.2f, Equity: %.2f",
                           TimeToString(m_TickTime,TIME_SECONDS),
                           AccountInfoDouble(ACCOUNT_BALANCE),
                           AccountInfoDouble(ACCOUNT_EQUITY));
   if(m_PositionType==OP_FLAT)
     {
      posinfo[1] = "Position: Flat";
      posinfo[2] = "";
     }
   else
     {
      string type=m_PositionType==OP_BUY ? "Long" : "Short";
      posinfo[1]=StringFormat("Position: %s %.2f at %s, Profit %.2f",
                              type,m_PositionLots,
                              DoubleToString(m_PositionOpenPrice,_Digits),
                              m_PositionProfit);
      posinfo[2]=StringFormat("Stop Loss: %s, Take Profit: %s",
                              DoubleToString(m_PositionStopLoss,_Digits),
                              DoubleToString(m_PositionTakeProfit,_Digits));
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::ManageOrderSend(int type,double lots,double stoploss,double takeprofit)
  {
   for(int attempt=0; attempt<TRADE_RETRY_COUNT; attempt++)
     {
      if(IsTradeContextFree())
        {
         ResetLastError();

         MqlTick tick;
         SymbolInfoTick(_Symbol,tick);

         double orderLots       = NormalizeEntrySize(lots);
         double stopLossPrice   = GetStopLossPrice(type,stoploss);
         double takeProfitPrice = GetTakeProfitPrice(type,takeprofit);

         MqlTradeRequest     request;
         MqlTradeResult      result;
         MqlTradeCheckResult check;
         ZeroMemory(request);
         ZeroMemory(result);
         ZeroMemory(check);

         request.action       = TRADE_ACTION_DEAL;
         request.symbol       = _Symbol;
         request.volume       = orderLots;
         request.type         = (type==OP_BUY)?ORDER_TYPE_BUY:ORDER_TYPE_SELL;
         request.type_filling = ORDER_FILLING_FOK;
         request.deviation    = 10;
         request.sl           = stopLossPrice;
         request.tp           = takeProfitPrice;
         request.comment      = Order_Comment;

         bool isOrderCheck=OrderCheck(request,check);
         string retcode=ResultRetcodeDescription(check.retcode);

         if(!isOrderCheck)
            Print("Error: ",__FUNCTION__,"(): OrderCheck(): ",retcode);

         bool isOrderSend=false;
         if(isOrderCheck)
           {
            isOrderSend=OrderSend(request,result);
            retcode=ResultRetcodeDescription(result.retcode);

            if(!isOrderSend)
               Print("Error: ",__FUNCTION__,"(): OrderSend(): ",retcode);
           }

         bool isExecuted=isOrderCheck && isOrderSend && result.retcode==TRADE_RETCODE_DONE;

         SetAggregatePosition();

         m_LastError=GetLastError();
         if(Write_Log_File)
            m_Logger.WriteLogLine(__FUNCTION__+
                                  ": "+_Symbol+
                                  ", Type="       +(type==OP_BUY ? "Long" : "Short")+
                                  ", Lots="       +DoubleToString(orderLots,2)+
                                  ", Price="      +DoubleToString(result.price,_Digits)+
                                  ", StopLoss="   +DoubleToString(stopLossPrice,_Digits)+
                                  ", TakeProfit=" +DoubleToString(takeProfitPrice,_Digits)+
                                  ", RetCode="    +retcode+
                                  ", LastError="  +IntegerToString(m_LastError));

         if(isExecuted)
            return (true);
        }

      Sleep(TRADE_RETRY_WAIT);
     }
   return (false);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::CloseCurrentPosition()
  {
   if(m_PositionType==OP_FLAT)
      return (true);
   int    type=m_PositionType==OP_BUY?OP_SELL:OP_BUY;
   double lots=m_PositionLots;
   bool   orderResponse=ManageOrderSend(type,lots,0,0);
   return (orderResponse);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::ModifyPosition(double stoploss,double takeprofit)
  {
   for(int attempt=0; attempt<TRADE_RETRY_COUNT; attempt++)
     {
      if(IsTradeContextFree())
        {
         if(m_PositionType==OP_FLAT)
            return(false);
         double stopLossPrice   = CorrectStopLossPrice(m_PositionType, stoploss);
         double takeProfitPrice = CorrectTakeProfitPrice(m_PositionType, takeprofit);

         MqlTradeRequest     request;
         MqlTradeResult      result;
         MqlTradeCheckResult check;
         ZeroMemory(request);
         ZeroMemory(result);
         ZeroMemory(check);

         request.symbol = _Symbol;
         request.action = TRADE_ACTION_SLTP;
         request.sl = stopLossPrice;
         request.tp = takeProfitPrice;

         bool isOrderCheck=OrderCheck(request,check);
         string retcode=ResultRetcodeDescription(check.retcode);

         if(!isOrderCheck)
            Print("Error: ","ModifyPosition: ","OrderCheck: ",retcode);

         bool isOrderSend=false;
         if(isOrderCheck)
           {
            isOrderSend=OrderSend(request,result);
            retcode=ResultRetcodeDescription(result.retcode);

            if(!isOrderSend)
               Print("Error: ","ModifyPosition: ","OrderSend: ",retcode);
           }

         bool isExecuted=isOrderCheck && isOrderSend && result.retcode==TRADE_RETCODE_DONE;

         SetAggregatePosition();

         m_LastError=GetLastError();
         if(Write_Log_File)
            m_Logger.WriteLogLine("ModifyPosition: " + _Symbol+
                                  ", StopLoss="   +DoubleToString(stopLossPrice,_Digits)+
                                  ", TakeProfit=" +DoubleToString(takeProfitPrice,_Digits)+
                                  ", RetCode="    +retcode+
                                  ", LastError="  +IntegerToString(m_LastError));

         if(isExecuted)
            return (true);
        }

      Sleep(TRADE_RETRY_WAIT);
     }
   return (false);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::GetTakeProfitPrice(int type,double takeprofit)
  {
   if(takeprofit<Epsilon)
      return (0);

   if(takeprofit<m_StopLevel)
      takeprofit=m_StopLevel;

   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   double takeProfitPrice=(type==OP_BUY)
                          ? tick.bid+takeprofit*_Point
                          : tick.ask-takeprofit*_Point;

   return (NormalizeEntryPrice(takeProfitPrice));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::GetStopLossPrice(int type,double stoploss)
  {
   if(stoploss<Epsilon)
      return (0);

   if(stoploss<m_StopLevel)
      stoploss=m_StopLevel;

   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   double stopLossPrice=(type==OP_BUY)
                        ? tick.bid-stoploss*_Point
                        : tick.ask+stoploss*_Point;

   return (NormalizeEntryPrice(stopLossPrice));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::CorrectTakeProfitPrice(int type,double takeProfitPrice)
  {
   if(takeProfitPrice<Epsilon)
      return (0);

   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   if(type==OP_BUY)
     {
      double minTPPrice=tick.bid+m_StopLevel*_Point;
      if(takeProfitPrice<minTPPrice)
         takeProfitPrice=minTPPrice;
     }
   else // if (type==OP_SELL)
     {
      double maxTPPrice=tick.ask-m_StopLevel*_Point;
      if(takeProfitPrice>maxTPPrice)
         takeProfitPrice=maxTPPrice;
     }

   return (NormalizeEntryPrice(takeProfitPrice));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::CorrectStopLossPrice(int type,double stopLossPrice)
  {
   if(stopLossPrice<Epsilon)
      return (0);

   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   if(type==OP_BUY)
     {
      double minSLPrice=tick.bid-m_StopLevel*_Point;
      if(stopLossPrice>minSLPrice)
         stopLossPrice=minSLPrice;
     }
   else // if(type==OP_SELL)
     {
      double maxSLPrice=tick.ask+m_StopLevel*_Point;
      if(stopLossPrice<maxSLPrice)
         stopLossPrice=maxSLPrice;
     }

   return (NormalizeEntryPrice(stopLossPrice));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::NormalizeEntryPrice(double price)
  {
   double tickSize=SymbolInfoDouble(_Symbol,SYMBOL_TRADE_TICK_SIZE);
   if(tickSize!=0)
      return (NormalizeDouble(MathRound(price/tickSize)*tickSize,_Digits));
   return (NormalizeDouble(price,_Digits));
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::NormalizeEntrySize(double size)
  {
   double minlot  = SymbolInfoDouble(_Symbol,SYMBOL_VOLUME_MIN);
   double maxlot  = SymbolInfoDouble(_Symbol,SYMBOL_VOLUME_MAX);
   double lotstep = SymbolInfoDouble(_Symbol,SYMBOL_VOLUME_STEP);

   if(size<minlot-Epsilon)
      return (0);

   if(MathAbs(size-minlot)<Epsilon)
      return (minlot);

   int steps=(int) MathRound((size-minlot)/lotstep);
   size=minlot+steps*lotstep;

   if(size>=maxlot)
      size=maxlot;

   return (size);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::SetMaxStopLoss()
  {
   if(m_PositionType==OP_FLAT || Protection_Max_StopLoss==0)
      return;

   double stopLossPrice=m_PositionStopLoss;
   int spread=(int)SymbolInfoInteger(_Symbol,SYMBOL_SPREAD);
   int stopLoss=(int)MathRound(MathAbs(m_PositionOpenPrice-m_PositionStopLoss)/_Point);

   if(stopLossPrice<Epsilon || stopLoss>Protection_Max_StopLoss+spread)
     {
      stopLossPrice=(m_PositionType==OP_BUY)
                    ? m_PositionOpenPrice-_Point*(Protection_Max_StopLoss+spread)
                    : m_PositionOpenPrice+_Point*(Protection_Max_StopLoss+spread);

      if(Write_Log_File)
         m_Logger.WriteLogRequest("SetMaxStopLoss",
                                  "StopLossPrice="+DoubleToString(stopLossPrice,_Digits));

      bool result=ModifyPosition(stopLossPrice,m_PositionTakeProfit);

      if(result)
         Print("SetMaxStopLoss(",Protection_Max_StopLoss,") set StopLoss to ",
               DoubleToString(stopLossPrice,_Digits));
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::SetBreakEvenStop()
  {
   if(SetAggregatePosition()<=0)
      return;

   double breakeven=m_StopLevel;
   if(breakeven<m_BreakEven)
      breakeven=m_BreakEven;

   double breakprice=0; // Break Even price including commission.
   double commission=0; // Commission in points.
   if(m_PositionCommission!=0)
      commission=MathAbs(m_PositionCommission)/
                 SymbolInfoDouble(_Symbol,SYMBOL_TRADE_TICK_VALUE);

   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   if(m_PositionType==OP_BUY)
     {
      breakprice=m_PositionOpenPrice+_Point*commission/m_PositionLots;
      if(tick.bid-breakprice>=_Point*breakeven)
        {
         if(m_PositionStopLoss<breakprice)
           {
            if(Write_Log_File)
               m_Logger.WriteLogRequest("SetBreakEvenStop",
                                        "BreakPrice="+DoubleToString(breakprice,_Digits));

            ModifyPosition(breakprice,m_PositionTakeProfit);

            Print("SetBreakEvenStop(",m_BreakEven,") set StopLoss to ",
                  DoubleToString(breakprice,_Digits),", Bid=",tick.bid);
           }
        }
     }
   else
      if(m_PositionType==OP_SELL)
        {
         breakprice=m_PositionOpenPrice-_Point*commission/m_PositionLots;
         if(breakprice-tick.ask>=_Point*breakeven)
           {
            if(m_PositionStopLoss==0||m_PositionStopLoss>breakprice)
              {
               if(Write_Log_File)
                  m_Logger.WriteLogRequest("SetBreakEvenStop",
                                           "BreakPrice="+DoubleToString(breakprice,_Digits));

               ModifyPosition(breakprice,m_PositionTakeProfit);

               Print("SetBreakEvenStop(",m_BreakEven,") set StopLoss to ",
                     DoubleToString(breakprice,_Digits),", Ask=",tick.ask);
              }
           }
        }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::SetTrailingStop(bool isNewBar)
  {
   bool isCheckTS=true;

   if(isNewBar)
     {
      if(m_PositionType==OP_BUY && m_PositionTime>m_BarHighTime)
         isCheckTS=false;

      if(m_PositionType==OP_SELL && m_PositionTime>m_BarLowTime)
         isCheckTS=false;

      m_BarHighTime    = Time(_Symbol, _Period, 0);
      m_BarLowTime     = Time(_Symbol, _Period, 0);
      m_CurrentBarHigh = High(_Symbol, _Period, 0);
      m_CurrentBarLow  = Low(_Symbol, _Period, 0);
     }
   else
     {
      if(High(_Symbol,_Period,0)>m_CurrentBarHigh)
        {
         m_CurrentBarHigh = High(_Symbol, _Period, 0);
         m_BarHighTime    = Time(_Symbol, _Period, 0);
        }
      if(Low(_Symbol,_Period,0)<m_CurrentBarLow)
        {
         m_CurrentBarLow = Low(_Symbol, _Period, 0);
         m_BarLowTime    = Time(_Symbol, _Period, 0);
        }
     }

   if(SetAggregatePosition()<=0)
      return;

   if(m_TrailingMode==TrailingStopMode_Tick)
      SetTrailingStopTickMode();
   else
      if(m_TrailingMode==TrailingStopMode_Bar && isNewBar && isCheckTS)
         SetTrailingStopBarMode();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::SetTrailingStopBarMode()
  {
   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   if(m_PositionType==OP_BUY)
     {
      // Long position
      double stopLossPrice=High(_Symbol,_Period,1)-_Point*m_TrailingStop;
      if(m_PositionStopLoss<stopLossPrice-m_PipsValue)
        {
         if(stopLossPrice<tick.bid)
           {
            if(stopLossPrice>tick.bid-_Point*m_StopLevel)
               stopLossPrice=tick.bid-_Point*m_StopLevel;

            if(Write_Log_File)
               m_Logger.WriteLogRequest("SetTrailingStopBarMode",
                                        "StopLoss="+DoubleToString(stopLossPrice,_Digits));

            ModifyPosition(stopLossPrice,m_PositionTakeProfit);

            Print("Trailing Stop (",m_TrailingStop,") moved to: ",
                  DoubleToString(stopLossPrice,_Digits),", Bid=",tick.bid);
           }
         else
           {
            if(Write_Log_File)
               m_Logger.WriteLogRequest("SetTrailingStopBarMode",
                                        "StopLoss="+DoubleToString(stopLossPrice,_Digits));

            bool isSucceed=CloseCurrentPosition()==0;

            int lastErrorOrdClose=GetLastError();
            lastErrorOrdClose=(lastErrorOrdClose>0) ? lastErrorOrdClose : m_LastError;
           }
        }
     }
   else
      if(m_PositionType==OP_SELL)
        {
         // Short position
         double stopLossPrice=Low(_Symbol,_Period,1)+_Point*m_TrailingStop;
         if(m_PositionStopLoss>stopLossPrice+m_PipsValue)
           {
            if(stopLossPrice>tick.ask)
              {
               if(stopLossPrice<tick.ask+_Point*m_StopLevel)
                  stopLossPrice=tick.ask+_Point*m_StopLevel;

               if(Write_Log_File)
                  m_Logger.WriteLogRequest("SetTrailingStopBarMode",
                                           "StopLoss="+DoubleToString(stopLossPrice,_Digits));

               ModifyPosition(stopLossPrice,m_PositionTakeProfit);

               Print("Trailing Stop (",m_TrailingStop,") moved to: ",
                     DoubleToString(stopLossPrice,_Digits),", Ask=",tick.ask);
              }
            else
              {
               if(Write_Log_File)
                  m_Logger.WriteLogRequest("SetTrailingStopBarMode",
                                           "StopLoss="+DoubleToString(stopLossPrice,_Digits));

               bool isSucceed=CloseCurrentPosition()==0;

               int lastErrorOrdClose=GetLastError();
               lastErrorOrdClose=(lastErrorOrdClose>0) ? lastErrorOrdClose : m_LastError;
              }
           }
        }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::SetTrailingStopTickMode()
  {
   MqlTick tick;
   SymbolInfoTick(_Symbol,tick);

   if(m_PositionType==OP_BUY)
     {
      // Long position
      if(tick.bid>=m_PositionOpenPrice+m_TrailingStop*_Point)
         if(m_PositionStopLoss<tick.bid-(m_TrailingStop+TrailingStop_Moving_Step)*_Point)
           {
            double stopLossPrice=tick.bid-m_TrailingStop*_Point;
            if(Write_Log_File)
               m_Logger.WriteLogRequest("SetTrailingStopTickMode",
                                        "StopLoss="+DoubleToString(stopLossPrice,_Digits));

            ModifyPosition(stopLossPrice,m_PositionTakeProfit);

            Print("Trailing Stop (",m_TrailingStop,") moved to: ",
                  DoubleToString(stopLossPrice,_Digits),", Bid=",tick.bid);
           }
     }
   else
      if(m_PositionType==OP_SELL)
        {
         // Short position
         if(m_PositionOpenPrice-tick.ask>=_Point*m_TrailingStop)
            if(m_PositionStopLoss>tick.ask+_Point *(m_TrailingStop+TrailingStop_Moving_Step))
              {
               double stopLossPrice=tick.ask+m_TrailingStop*_Point;
               if(Write_Log_File)
                  m_Logger.WriteLogRequest("SetTrailingStopTickMode",
                                           "StopLoss="+DoubleToString(stopLossPrice,_Digits));

               ModifyPosition(stopLossPrice,m_PositionTakeProfit);

               Print("Trailing Stop (",m_TrailingStop,") moved to: ",
                     DoubleToString(stopLossPrice,_Digits),", Ask=",tick.ask);
              }
        }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::DetectSLTPActivation()
  {
// Save position values from previous tick.
   double oldStopLoss   = m_PositionStopLoss;
   double oldTakeProfit = m_PositionTakeProfit;
   double oldProfit     = m_PositionProfit;
   int    oldType       = m_PositionType;
   double oldLots       = m_PositionLots;

   SetAggregatePosition();

   if(oldType!=OP_FLAT && m_PositionType==OP_FLAT)
     {
      // Position was closed this tick. It must be due to SL or TP.
      double closePrice=(oldType==OP_BUY) ? m_DataMarket.Bid : m_DataMarket.Ask;
      string stopMessage="Position was closed";
      double activatedStopLoss   = closePrice; // At Stop Loss
      double activatedTakeProfit = closePrice; // or at Take Profit ?

      if(MathAbs(oldStopLoss-closePrice)<2*m_PipsValue)
         stopMessage="Activated StopLoss="+DoubleToString(activatedStopLoss,_Digits);
      else
         if(MathAbs(oldTakeProfit-closePrice)<2*m_PipsValue)
            stopMessage="Activated TakeProfit="+DoubleToString(activatedTakeProfit,_Digits);

      // For Martingale (if used)
      m_ConsecutiveLosses=(oldProfit<0) ? m_ConsecutiveLosses+1 : 0;

      string message=stopMessage+
                     ", ClosePrice="        +DoubleToString(closePrice,_Digits)+
                     ", ClosedLots= "       +DoubleToString(oldLots,2)+
                     ", Profit="            +DoubleToString(oldProfit,2)+
                     ", ConsecutiveLosses=" +IntegerToString(m_ConsecutiveLosses);

      if(Write_Log_File)
         m_Logger.WriteNewLogLine(message);
      Print(message);
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::IsTradeContextFree()
  {
   if(MQL5InfoInteger(MQL5_TRADE_ALLOWED))
      return (true);

   uint startWait=GetTickCount();
   Print("Trade context is busy! Waiting...");

   while(true)
     {
      if(IsStopped())
         return (false);

      uint diff=GetTickCount()-startWait;
      if(diff>30*1000)
        {
         Print("The waiting limit exceeded!");
         return (false);
        }

      if(MQL5InfoInteger(MQL5_TRADE_ALLOWED))
         return (true);
      Sleep(100);
     }

   return (true);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TickType ActionTrade5::GetTickType(DataSet *dataSet,bool isNewBar)
  {
   DataPeriod period      = dataSet.Period;
   int        bars        = dataSet.Bars;
   datetime   serverTime  = dataSet.ServerTime;
   datetime   barOpenTime = dataSet.Time[bars-1];

   TickType type=TickType_Regular;

   if(isNewBar)
     {
      m_BarOpenTimeForLastCloseTick=-1;
      type=TickType_Open;
     }

   bool isClose=((barOpenTime+period*60)-serverTime)<Bar_Close_Advance;
   if(isClose)
     {
      if(m_BarOpenTimeForLastCloseTick==barOpenTime)
        {
         type=TickType_AfterClose;
        }
      else
        {
         m_BarOpenTimeForLastCloseTick=barOpenTime;
         type=isNewBar ? TickType_OpenClose : TickType_Close;
        }
     }

   return (type);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::CalculateTrade(TickType ticktype)
  {
// Exit
   bool closeOk=false;

   if(m_CloseStrPriceType!=StrategyPriceType_CloseAndReverse &&
      m_DataMarket.PositionTicket!=0)
     {
      if(ticktype==TickType_Open && m_CloseStrPriceType==StrategyPriceType_Close &&
         m_OpenStrPriceType!=StrategyPriceType_Close)
        {
         // We have missed close at the previous Bar Close
         TradeDirection direction=AnalyzeExitDirection();
         if(direction==TradeDirection_Both ||
            (direction == TradeDirection_Long  &&
             m_DataMarket.PositionDirection == PosDirection_Short) ||
            (direction == TradeDirection_Short &&
             m_DataMarket.PositionDirection == PosDirection_Long))
           {
            // we have a missed close Order
            if(DoExitTrade())
               UpdateDataMarket(m_DataMarket);
           }
        }
      else
         if(((m_CloseStrPriceType==StrategyPriceType_Open)  &&
             (ticktype==TickType_Open  || ticktype==TickType_OpenClose)) ||
            ((m_CloseStrPriceType==StrategyPriceType_Close) &&
             (ticktype==TickType_Close || ticktype==TickType_OpenClose)))
           {
            // Exit at Bar Open or Bar Close.
            TradeDirection direction=AnalyzeExitDirection();
            if(direction==TradeDirection_Both ||
               (direction == TradeDirection_Long  &&
                m_DataMarket.PositionDirection == PosDirection_Short) ||
               (direction == TradeDirection_Short &&
                m_DataMarket.PositionDirection == PosDirection_Long))
              {
               // Close the current position.
               closeOk=DoExitTrade();
               if(closeOk)
                  UpdateDataMarket(m_DataMarket);
              }
           }
         else
            if(m_CloseStrPriceType==StrategyPriceType_Close &&
               m_OpenStrPriceType!=StrategyPriceType_Close &&
               ticktype==TickType_AfterClose)
              {
               // Exit at after close tick.
               TradeDirection direction=AnalyzeExitDirection();
               if(direction==TradeDirection_Both ||
                  (direction == TradeDirection_Long  &&
                   m_DataMarket.PositionDirection == PosDirection_Short) ||
                  (direction == TradeDirection_Short &&
                   m_DataMarket.PositionDirection == PosDirection_Long))
                  closeOk=DoExitTrade(); // Close the current position.
              }
            else
               if(m_CloseStrPriceType==StrategyPriceType_Indicator)
                 {
                  // Exit at an indicator value.
                  TradeDirection priceReached=AnalyzeExitPrice();
                  if(priceReached==TradeDirection_Long)
                    {
                     TradeDirection direction=AnalyzeExitDirection();
                     if(direction==TradeDirection_Long || direction==TradeDirection_Both)
                        if(m_DataMarket.PositionDirection==PosDirection_Short)
                           closeOk=DoExitTrade(); // Close a short position.
                    }
                  else
                     if(priceReached==TradeDirection_Short)
                       {
                        TradeDirection direction=AnalyzeExitDirection();
                        if(direction==TradeDirection_Short || direction==TradeDirection_Both)
                           if(m_DataMarket.PositionDirection==PosDirection_Long)
                              closeOk=DoExitTrade(); // Close a long position.
                       }
                     else
                        if(priceReached==TradeDirection_Both)
                          {
                           TradeDirection direction=AnalyzeExitDirection();
                           if(direction==TradeDirection_Long || direction==TradeDirection_Short ||
                              direction==TradeDirection_Both)
                              closeOk=DoExitTrade(); // Close the current position.
                          }
                 }
     }

// Checks if we closed a position successfully.
   if(closeOk && !(m_OpenStrPriceType==StrategyPriceType_Close &&
                   ticktype==TickType_Close))
      return;

// This is to prevent new entry after Bar Closing has been executed.
   if(m_CloseStrPriceType==StrategyPriceType_Close && ticktype==TickType_AfterClose)
      return;

   if(((m_OpenStrPriceType==StrategyPriceType_Open)  &&
       (ticktype==TickType_Open  || ticktype==TickType_OpenClose)) ||
      ((m_OpenStrPriceType==StrategyPriceType_Close) &&
       (ticktype==TickType_Close || ticktype==TickType_OpenClose)))
     {
      // Entry at Bar Open or Bar Close.
      TradeDirection direction=AnalyzeEntryDirection();
      if(direction==TradeDirection_Long || direction==TradeDirection_Short)
         DoEntryTrade(direction);
     }
   else
      if(m_OpenStrPriceType==StrategyPriceType_Indicator)
        {
         // Entry at an indicator value.
         TradeDirection priceReached=AnalyzeEntryPrice();
         if(priceReached==TradeDirection_Long)
           {
            TradeDirection direction=AnalyzeEntryDirection();
            if(direction==TradeDirection_Long || direction==TradeDirection_Both)
               DoEntryTrade(TradeDirection_Long);
           }
         else
            if(priceReached==TradeDirection_Short)
              {
               TradeDirection direction=AnalyzeEntryDirection();
               if(direction==TradeDirection_Short || direction==TradeDirection_Both)
                  DoEntryTrade(TradeDirection_Short);
              }
            else
               if(priceReached==TradeDirection_Both)
                 {
                  TradeDirection direction=AnalyzeEntryDirection();
                  if(direction==TradeDirection_Long || direction==TradeDirection_Short)
                     DoEntryTrade(direction);
                 }
        }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
PosDirection ActionTrade5::GetNewPositionDirection(OrderDirection ordDir,
      double ordLots,PosDirection posDir,double posLots)
  {
   if(ordDir!=OrderDirection_Buy && ordDir!=OrderDirection_Sell)
      return (PosDirection_None);

   PosDirection currentDir=posDir;
   double currentLots=posLots;

   switch(currentDir)
     {
      case PosDirection_Long:
         if(ordDir==OrderDirection_Buy)
            return (PosDirection_Long);
         if(currentLots>ordLots+Epsilon)
            return (PosDirection_Long);
         return (currentLots < ordLots - Epsilon
                 ? PosDirection_Short
                 : PosDirection_None);
      case PosDirection_Short:
         if(ordDir==OrderDirection_Sell)
            return (PosDirection_Short);
         if(currentLots>ordLots+Epsilon)
            return (PosDirection_Short);
         return (currentLots < ordLots - Epsilon
                 ? PosDirection_Long
                 : PosDirection_None);
     }


   return (ordDir==OrderDirection_Buy
           ? PosDirection_Long
           : PosDirection_Short);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::InitTrade()
  {
   m_OpenTimeExec=m_Strategy.Slot[0].IndicatorPointer.ExecTime;
   m_OpenStrPriceType=StrategyPriceType_Unknown;
   if(m_OpenTimeExec==ExecutionTime_AtBarOpening)
      m_OpenStrPriceType=StrategyPriceType_Open;
   else
      if(m_OpenTimeExec==ExecutionTime_AtBarClosing)
         m_OpenStrPriceType=StrategyPriceType_Close;
      else
         m_OpenStrPriceType=StrategyPriceType_Indicator;

   m_CloseTimeExec=m_Strategy.Slot[m_Strategy.CloseSlotNumber()].IndicatorPointer.ExecTime;
   m_CloseStrPriceType=StrategyPriceType_Unknown;
   if(m_CloseTimeExec==ExecutionTime_AtBarOpening)
      m_CloseStrPriceType=StrategyPriceType_Open;
   else
      if(m_CloseTimeExec==ExecutionTime_AtBarClosing)
         m_CloseStrPriceType=StrategyPriceType_Close;
      else
         if(m_CloseTimeExec==ExecutionTime_CloseAndReverse)
            m_CloseStrPriceType=StrategyPriceType_CloseAndReverse;
         else
            m_CloseStrPriceType=StrategyPriceType_Indicator;

   m_UseLogicalGroups=m_Strategy.IsUsingLogicalGroups();

   if(m_UseLogicalGroups)
     {
      m_Strategy.Slot[0].LogicalGroup="All";
      m_Strategy.Slot[m_Strategy.CloseSlotNumber()].LogicalGroup="All";

      for(int slot=0; slot<m_Strategy.CloseSlotNumber(); slot++)
        {
         if(!m_GroupsAllowLong.ContainsKey(m_Strategy.Slot[slot].LogicalGroup))
            m_GroupsAllowLong.Add(m_Strategy.Slot[slot].LogicalGroup,false);
         if(!m_GroupsAllowShort.ContainsKey(m_Strategy.Slot[slot].LogicalGroup))
            m_GroupsAllowShort.Add(m_Strategy.Slot[slot].LogicalGroup,false);
        }

      // List of logical groups
      int longCount=m_GroupsAllowLong.Count();
      for(int i=0; i<longCount; i++)
         m_OpeningLogicGroups.Add(m_GroupsAllowLong.Key(i));

      // Logical groups of the closing conditions.
      for(int slot=m_Strategy.CloseSlotNumber()+1; slot<m_Strategy.Slots(); slot++)
        {
         string group=m_Strategy.Slot[slot].LogicalGroup;
         if(!m_ClosingLogicGroups.Contains(group) && group!="all")
            m_ClosingLogicGroups.Add(group); // Adds all groups except "all"
        }

      if(m_ClosingLogicGroups.Count()==0)
         m_ClosingLogicGroups.Add("all");
     }

// Search if N Bars Exit is present as CloseFilter,
// could be any slot after first closing slot.
   m_NBarExit=0;
   for(int slot=m_Strategy.CloseSlotNumber(); slot<m_Strategy.Slots(); slot++)
     {
      if(m_Strategy.Slot[slot].IndicatorName!="N Bars Exit")
         continue;
      m_NBarExit=(int) m_Strategy.Slot[slot].IndicatorPointer.NumParam[0].Value;
      break;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::AnalyzeEntryPrice()
  {
   int bar=m_DataSet[0].Bars-1;

   double buyPrice  = 0;
   double sellPrice = 0;
   for(int i=0; i<m_Strategy.Slot[0].IndicatorPointer.Components(); i++)
     {
      IndicatorComp *component=m_Strategy.Slot[0].IndicatorPointer.Component[i];
      IndComponentType compType=component.DataType;
      if(compType==IndComponentType_OpenLongPrice)
         buyPrice=component.Value[bar];
      else
         if(compType==IndComponentType_OpenShortPrice)
            sellPrice=component.Value[bar];
         else
            if(compType==IndComponentType_OpenPrice || compType==IndComponentType_OpenClosePrice)
               buyPrice=sellPrice=component.Value[bar];
      component=NULL;
     }

   double basePrice= m_DataMarket.Close;
   double oldPrice = m_DataMarket.OldClose;

   bool canOpenLong=(buyPrice>oldPrice+Epsilon  && buyPrice<basePrice+Epsilon) ||
                    (buyPrice>basePrice-Epsilon && buyPrice<oldPrice -Epsilon);

   bool canOpenShort=(sellPrice>oldPrice+Epsilon  && sellPrice<basePrice+Epsilon) ||
                     (sellPrice>basePrice-Epsilon && sellPrice<oldPrice -Epsilon);

   TradeDirection direction=TradeDirection_None;

   if(canOpenLong && canOpenShort)
      direction=TradeDirection_Both;
   else
      if(canOpenLong)
         direction=TradeDirection_Long;
      else
         if(canOpenShort)
            direction=TradeDirection_Short;

   return (direction);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::AnalyzeEntryDirection()
  {
   int bar=m_DataSet[0].Bars-1;

// Do not send entry order when we are not on time
   if(m_OpenTimeExec==ExecutionTime_AtBarOpening)
      for(int i=0; i<m_Strategy.Slot[0].IndicatorPointer.Components(); i++)
        {
         IndicatorComp *component=m_Strategy.Slot[0].IndicatorPointer.Component[i];
         if(component.DataType!=IndComponentType_OpenLongPrice &&
            component.DataType!=IndComponentType_OpenShortPrice &&
            component.DataType!=IndComponentType_OpenPrice)
            continue;
         if(component.Value[bar]<Epsilon)
            return (TradeDirection_None);
         component=NULL;
        }

   for(int i=0; i<m_Strategy.Slots(); i++)
     {
      if(m_Strategy.Slot[i].IndicatorName=="Enter Once")
        {
         if(m_Strategy.Slot[i].IndicatorPointer.ListParam[0].Text=="Enter no more than once a bar")
           {
            if(m_DataSet[0].Time[bar]==m_TimeLastEntryBar)
               return (TradeDirection_None);
           }
         else
            if(m_Strategy.Slot[i].IndicatorPointer.ListParam[0].Text=="Enter no more than once a day")
              {
               if(TimeDayOfYear(m_DataSet[0].Time[bar])==TimeDayOfYear(m_TimeLastEntryBar))
                  return (TradeDirection_None);
              }
            else
               if(m_Strategy.Slot[i].IndicatorPointer.ListParam[0].Text=="Enter no more than once a week")
                 {
                  if(TimeDayOfWeek(m_DataSet[0].Time[bar])>=TimeDayOfWeek(m_TimeLastEntryBar) &&
                     m_DataSet[0].Time[bar]<m_TimeLastEntryBar+7*24*60*60)
                     return (TradeDirection_None);
                 }
               else
                  if(m_Strategy.Slot[i].IndicatorPointer.ListParam[0].Text=="Enter no more than once a month")
                    {
                     if(TimeMonth(m_DataSet[0].Time[bar])==TimeMonth(m_TimeLastEntryBar))
                        return (TradeDirection_None);
                    }
        }
     }

// Determining of the buy/sell entry prices.
   double buyPrice=0;
   double sellPrice=0;
   for(int i=0; i<m_Strategy.Slot[0].IndicatorPointer.Components(); i++)
     {
      IndicatorComp *component=m_Strategy.Slot[0].IndicatorPointer.Component[i];
      IndComponentType compType=component.DataType;
      if(compType==IndComponentType_OpenLongPrice)
         buyPrice=component.Value[bar];
      else
         if(compType==IndComponentType_OpenShortPrice)
            sellPrice=component.Value[bar];
         else
            if(compType==IndComponentType_OpenPrice || compType==IndComponentType_OpenClosePrice)
               buyPrice=sellPrice=component.Value[bar];
      component=NULL;
     }

// Decide whether to open
   bool canOpenLong  = buyPrice  > Epsilon;
   bool canOpenShort = sellPrice > Epsilon;

   if(m_UseLogicalGroups)
     {
      for(int i=0; i<m_OpeningLogicGroups.Count(); i++)
        {
         string group=m_OpeningLogicGroups.Get(i);

         bool groupOpenLong  = canOpenLong;
         bool groupOpenShort = canOpenShort;

         AnalyzeEntryLogicConditions(bar,group,buyPrice,sellPrice,groupOpenLong,groupOpenShort);

         m_GroupsAllowLong.Set(group,groupOpenLong);
         m_GroupsAllowShort.Set(group,groupOpenShort);
        }

      bool groupLongEntry=false;
      for(int i=0; i<m_GroupsAllowLong.Count(); i++)
        {
         string key   = m_GroupsAllowLong.Key(i);
         bool   value = m_GroupsAllowLong.Value(key);
         if((m_GroupsAllowLong.Count()>1 && key!="All") || m_GroupsAllowLong.Count()==1)
            groupLongEntry=groupLongEntry || value;
        }

      bool groupShortEntry=false;
      for(int i=0; i<m_GroupsAllowShort.Count(); i++)
        {
         string key   = m_GroupsAllowShort.Key(i);
         bool   value = m_GroupsAllowShort.Value(key);
         if((m_GroupsAllowShort.Count()>1 && key!="All") || m_GroupsAllowShort.Count()==1)
            groupShortEntry=groupShortEntry || value;
        }

      canOpenLong  = canOpenLong  && groupLongEntry  && m_GroupsAllowLong.Value("All");
      canOpenShort = canOpenShort && groupShortEntry && m_GroupsAllowShort.Value("All");
     }
   else
     {
      AnalyzeEntryLogicConditions(bar,"A",buyPrice,sellPrice,canOpenLong,canOpenShort);
     }

   TradeDirection direction=TradeDirection_None;
   if(canOpenLong && canOpenShort)
      direction=TradeDirection_Both;
   else
      if(canOpenLong)
         direction=TradeDirection_Long;
      else
         if(canOpenShort)
            direction=TradeDirection_Short;

   return (direction);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::AnalyzeEntryLogicConditions(int bar,string group,double buyPrice,
      double sellPrice,bool &canOpenLong,bool &canOpenShort)
  {
   for(int slotIndex=0; slotIndex<=m_Strategy.CloseSlotNumber(); slotIndex++)
     {
      if(m_UseLogicalGroups && m_Strategy.Slot[slotIndex].LogicalGroup!=group &&
         m_Strategy.Slot[slotIndex].LogicalGroup!="All")
         continue;

      for(int i=0; i<m_Strategy.Slot[slotIndex].IndicatorPointer.Components(); i++)
        {
         IndicatorComp *component=m_Strategy.Slot[slotIndex].IndicatorPointer.Component[i];
         if(component.DataType==IndComponentType_AllowOpenLong && component.Value[bar]<0.5)
            canOpenLong=false;

         if(component.DataType==IndComponentType_AllowOpenShort && component.Value[bar]<0.5)
            canOpenShort=false;

         if(component.PosPriceDependence!=PositionPriceDependence_None)
           {
            double indicatorValue=component.Value[bar-component.UsePreviousBar];
            switch(component.PosPriceDependence)
              {
               case PositionPriceDependence_PriceBuyHigher:
                  canOpenLong=canOpenLong && buyPrice>indicatorValue+Epsilon;
                  break;
               case PositionPriceDependence_PriceBuyLower:
                  canOpenLong=canOpenLong && buyPrice<indicatorValue-Epsilon;
                  break;
               case PositionPriceDependence_PriceSellHigher:
                  canOpenShort=canOpenShort && sellPrice>indicatorValue+Epsilon;
                  break;
               case PositionPriceDependence_PriceSellLower:
                  canOpenShort=canOpenShort && sellPrice<indicatorValue-Epsilon;
                  break;
               case PositionPriceDependence_BuyHigherSellLower:
                  canOpenLong=canOpenLong && buyPrice>indicatorValue+Epsilon;
                  canOpenShort=canOpenShort && sellPrice<indicatorValue-Epsilon;
                  break;
               case PositionPriceDependence_BuyLowerSelHigher:
                  canOpenLong=canOpenLong && buyPrice<indicatorValue-Epsilon;
                  canOpenShort=canOpenShort && sellPrice>indicatorValue+Epsilon;
                  break;
              }
           }
         component=NULL;
        }
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::AnalyzeEntrySize(OrderDirection ordDir,PosDirection &newPosDir)
  {
   double size=0;
   PosDirection posDir=m_DataMarket.PositionDirection;
// Orders modification on a fly
// Checks whether we are on the market
   if(posDir==PosDirection_Long || posDir==PosDirection_Short)
     {
      // We are on the market and have Same Dir Signal
      if((ordDir==OrderDirection_Buy && posDir==PosDirection_Long) ||
         (ordDir==OrderDirection_Sell && posDir==PosDirection_Short))
        {
         size=0;
         newPosDir=posDir;
         if(m_DataMarket.PositionLots+TradingSize(m_Strategy.AddingLots) <
            m_Strategy.MaxOpenLots+Epsilon)
           {
            switch(m_Strategy.SameSignalAction)
              {
               case SameDirSignalAction_Add:
                  size=TradingSize(m_Strategy.AddingLots);
                  break;
               case SameDirSignalAction_Winner:
                  if(m_DataMarket.PositionProfit>Epsilon)
                     size=TradingSize(m_Strategy.AddingLots);
                  break;
               case SameDirSignalAction_Loser:
                  if(m_DataMarket.PositionProfit<-Epsilon)
                     size=TradingSize(m_Strategy.AddingLots);
                  break;
              }
           }
        }
      else
         if((ordDir==OrderDirection_Buy  && posDir==PosDirection_Short) ||
            (ordDir==OrderDirection_Sell && posDir==PosDirection_Long))
           {
            // In case of an Opposite Dir Signal
            switch(m_Strategy.OppSignalAction)
              {
               case OppositeDirSignalAction_Reduce:
                  if(m_DataMarket.PositionLots>TradingSize(m_Strategy.ReducingLots))
                    {
                     // Reducing
                     size=TradingSize(m_Strategy.ReducingLots);
                     newPosDir=posDir;
                    }
                  else
                    {
                     // Closing
                     size=m_DataMarket.PositionLots;
                     newPosDir=PosDirection_Closed;
                    }
                  break;
               case OppositeDirSignalAction_Close:
                  size=m_DataMarket.PositionLots;
                  newPosDir=PosDirection_Closed;
                  break;
               case OppositeDirSignalAction_Reverse:
                  size=m_DataMarket.PositionLots+TradingSize(m_Strategy.EntryLots);
                  newPosDir=(posDir==PosDirection_Long) ? PosDirection_Short : PosDirection_Long;
                  break;
               case OppositeDirSignalAction_Nothing:
                  size=0;
                  newPosDir=posDir;
                  break;
              }
           }
     }
   else
     {
      // We are flat
      size=TradingSize(m_Strategy.EntryLots);
      if(m_Strategy.UseMartingale && m_DataMarket.ConsecutiveLosses>0)
        {
         double correctedAmount = size*MathPow(m_Strategy.MartingaleMultiplier, m_DataMarket.ConsecutiveLosses);
         double normalizedAmount = NormalizeEntrySize(correctedAmount);
         size = MathMax(normalizedAmount, SymbolInfoDouble(_Symbol,SYMBOL_VOLUME_MIN));
        }
      size=MathMin(size,m_Strategy.MaxOpenLots);
      newPosDir=ordDir==OrderDirection_Buy ? PosDirection_Long : PosDirection_Short;
     }

   return (size);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::AnalyzeExitPrice()
  {
   IndicatorSlot *slot=m_Strategy.Slot[m_Strategy.CloseSlotNumber()];
   int bar=m_DataSet[0].Bars-1;

// Searching the exit price in the exit indicator slot.
   double buyPrice  = 0;
   double sellPrice = 0;
   for(int i=0; i<slot.IndicatorPointer.Components(); i++)
     {
      IndicatorComp *comp=slot.IndicatorPointer.Component[i];
      IndComponentType compType=comp.DataType;

      if(compType==IndComponentType_CloseLongPrice)
         sellPrice=comp.Value[bar];
      else
         if(compType==IndComponentType_CloseShortPrice)
            buyPrice=comp.Value[bar];
         else
            if(compType==IndComponentType_ClosePrice || compType==IndComponentType_OpenClosePrice)
               buyPrice=sellPrice=comp.Value[bar];

      comp=NULL;
     }

// We can close if the closing price is higher than zero.
   bool canCloseLong =sellPrice>Epsilon;
   bool canCloseShort=buyPrice>Epsilon;

// Check if the closing price was reached.
   if(canCloseLong)
      canCloseLong=(sellPrice>m_DataMarket.OldBid+Epsilon && sellPrice<m_DataMarket.Bid+Epsilon) ||
                   (sellPrice>m_DataMarket.Bid-Epsilon    && sellPrice<m_DataMarket.OldBid-Epsilon);
   if(canCloseShort)
      canCloseShort=(buyPrice>m_DataMarket.OldBid+Epsilon && buyPrice<m_DataMarket.Bid+Epsilon) ||
                    (buyPrice>m_DataMarket.Bid-Epsilon    && buyPrice<m_DataMarket.OldBid-Epsilon);

// Determine the trading direction.
   TradeDirection direction=TradeDirection_None;

   if(canCloseLong && canCloseShort)
      direction=TradeDirection_Both;
   else
      if(canCloseLong)
         direction=TradeDirection_Short;
      else
         if(canCloseShort)
            direction=TradeDirection_Long;

   slot=NULL;
   return (direction);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::AnalyzeExitDirection()
  {
   int bar=m_DataSet[0].Bars-1;
   int closeSlot=m_Strategy.CloseSlotNumber();

   if(m_CloseTimeExec==ExecutionTime_AtBarClosing)
      for(int i=0; i<m_Strategy.Slot[closeSlot].IndicatorPointer.Components(); i++)
        {
         if(m_Strategy.Slot[closeSlot].IndicatorPointer.Component[i].DataType != IndComponentType_CloseLongPrice &&
            m_Strategy.Slot[closeSlot].IndicatorPointer.Component[i].DataType != IndComponentType_CloseShortPrice &&
            m_Strategy.Slot[closeSlot].IndicatorPointer.Component[i].DataType != IndComponentType_ClosePrice)
            continue;
         if(m_Strategy.Slot[closeSlot].IndicatorPointer.Component[i].Value[bar]<Epsilon)
            return (TradeDirection_None);
        }

   if(m_Strategy.CloseSlots()==0)
      return (TradeDirection_Both);

   if(m_NBarExit>0 && (m_DataMarket.PositionOpenTime+(m_NBarExit*((int) m_DataSet[0].Period*60))<m_DataSet[0].ServerTime))
      return (TradeDirection_Both);

   TradeDirection direction=TradeDirection_None;

   if(m_UseLogicalGroups)
     {
      for(int i=0; i<m_ClosingLogicGroups.Count(); i++)
        {
         string group=m_ClosingLogicGroups.Get(i);
         TradeDirection groupDirection=TradeDirection_Both;

         // Determining of the slot direction
         for(int slot=m_Strategy.CloseSlotNumber()+1; slot<m_Strategy.Slots(); slot++)
           {
            TradeDirection slotDirection=TradeDirection_None;
            if(m_Strategy.Slot[slot].LogicalGroup==group || m_Strategy.Slot[slot].LogicalGroup=="all")
              {
               for(int c=0; c<m_Strategy.Slot[slot].IndicatorPointer.Components(); c++)
                  if(m_Strategy.Slot[slot].IndicatorPointer.Component[c].Value[bar]>0)
                     slotDirection=GetClosingDirection(slotDirection,m_Strategy.Slot[slot].IndicatorPointer.Component[c].DataType);

               groupDirection=ReduceDirectionStatus(groupDirection,slotDirection);
              }
           }

         direction=IncreaseDirectionStatus(direction,groupDirection);
        }
     }
   else
     {
      // Search close filters for a closing signal.
      for(int slot= m_Strategy.CloseSlotNumber()+1; slot<m_Strategy.Slots(); slot++)
         for(int c=0; c<m_Strategy.Slot[slot].IndicatorPointer.Components(); c++)
            if(m_Strategy.Slot[slot].IndicatorPointer.Component[c].Value[bar]>Epsilon)
               direction=GetClosingDirection(direction,m_Strategy.Slot[slot].IndicatorPointer.Component[c].DataType);
     }

   return (direction);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::ReduceDirectionStatus(TradeDirection baseDirection,TradeDirection direction)
  {
   if(baseDirection==direction || direction==TradeDirection_Both)
      return (baseDirection);

   if(baseDirection==TradeDirection_Both)
      return (direction);

   return (TradeDirection_None);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::IncreaseDirectionStatus(TradeDirection baseDirection,TradeDirection direction)
  {
   if(baseDirection==direction || direction==TradeDirection_None)
      return (baseDirection);

   if(baseDirection==TradeDirection_None)
      return (direction);

   return (TradeDirection_Both);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
TradeDirection ActionTrade5::GetClosingDirection(TradeDirection baseDirection,IndComponentType compDataType)
  {
   TradeDirection newDirection=baseDirection;

   if(compDataType==IndComponentType_ForceClose)
     {
      newDirection=TradeDirection_Both;
     }
   else
      if(compDataType==IndComponentType_ForceCloseShort)
        {
         if(baseDirection==TradeDirection_None)
            newDirection=TradeDirection_Long;
         else
            if(baseDirection==TradeDirection_Short)
               newDirection=TradeDirection_Both;
        }
      else
         if(compDataType==IndComponentType_ForceCloseLong)
           {
            if(baseDirection==TradeDirection_None)
               newDirection=TradeDirection_Short;
            else
               if(baseDirection==TradeDirection_Long)
                  newDirection=TradeDirection_Both;
           }

   return (newDirection);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::GetStopLossPoints(double lots)
  {
   double indStop = DBL_MAX;
   bool isIndStop = true;
   int closeSlot=m_Strategy.CloseSlotNumber();
   string name=m_Strategy.Slot[closeSlot].IndicatorName;

   if(name=="Account Percent Stop")
      indStop=AccountPercentStopPoints(m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[0].Value,lots);
   else
      if(name=="ATR Stop")
         indStop=m_Strategy.Slot[closeSlot].IndicatorPointer.Component[0].Value[m_DataSet[0].Bars-1]/m_DataSet[0].Point;
      else
         if(name=="Stop Loss" || name=="Stop Limit")
            indStop=m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[0].Value;
         else
            if(name=="Trailing Stop" || name=="Trailing Stop Limit")
               indStop=m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[0].Value;
            else
               isIndStop=false;

   double permStop = m_Strategy.UsePermanentSL ? m_Strategy.PermanentSL : DBL_MAX;
   double stopLoss = 0;

   if(isIndStop || m_Strategy.UsePermanentSL)
     {
      stopLoss=MathMin(indStop,permStop);
      if(stopLoss<m_DataSet[0].StopLevel)
         stopLoss=m_DataSet[0].StopLevel;
      stopLoss=MathRound(stopLoss);
     }

   return (stopLoss);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::GetTakeProfitPoints()
  {
   double takeprofit = 0;
   double permLimit  = m_Strategy.UsePermanentTP ? m_Strategy.PermanentTP : DBL_MAX;
   double indLimit   = DBL_MAX;
   bool   isIndLimit = true;
   int    closeSlot  = m_Strategy.CloseSlotNumber();
   string name       = m_Strategy.Slot[closeSlot].IndicatorName;

   if(name=="Take Profit")
      indLimit=m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[0].Value;
   else
      if(name=="Stop Limit" || name=="Trailing Stop Limit")
         indLimit=m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[1].Value;
      else
         isIndLimit=false;

   if(isIndLimit || m_Strategy.UsePermanentTP)
     {
      takeprofit=MathMin(indLimit,permLimit);
      if(takeprofit<m_DataSet[0].StopLevel)
         takeprofit=m_DataSet[0].StopLevel;
      takeprofit=MathRound(takeprofit);
     }

   return (takeprofit);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::DoEntryTrade(TradeDirection tradeDir)
  {
   double price;
   OrderDirection  ordDir;
   OperationType   opType;
   TraderOrderType type;

   if(m_TimeLastEntryBar!=m_BarTime)
      m_IsEnteredLong=m_IsEnteredShort=false;

   switch(tradeDir)
     {
      case TradeDirection_Long: // Buy
         if(m_IsEnteredLong)
            return;
         price  = m_DataMarket.Ask;
         ordDir = OrderDirection_Buy;
         opType = OperationType_Buy;
         type   = TraderOrderType_Buy;
         break;
      case TradeDirection_Short: // Sell
         if(m_IsEnteredShort)
            return;
         price  = m_DataMarket.Bid;
         ordDir = OrderDirection_Sell;
         opType = OperationType_Sell;
         type   = TraderOrderType_Sell;
         break;
      default: // Wrong direction of trade.
         return;
     }

   PosDirection newPosDir=PosDirection_None;
   double size=AnalyzeEntrySize(ordDir,newPosDir);

   if(size<m_DataSet[0].MinLot-Epsilon)
      return;  // The entry trade is cancelled.

   string symbol     = m_DataSet[0].Symbol;
   double lots       = size;
   double stoploss   = GetStopLossPoints(size);
   double takeprofit = GetTakeProfitPoints();
   double point      = m_DataSet[0].Point;

   int trlStop=0;
   TrailingStopMode trlMode=TrailingStopMode_Bar;
   int closeSlot=m_Strategy.CloseSlotNumber();
   string name=m_Strategy.Slot[closeSlot].IndicatorName;

   if(name=="Trailing Stop" || name=="Trailing Stop Limit")
     {
      trlStop=(int) m_Strategy.Slot[closeSlot].IndicatorPointer.NumParam[0].Value;
      string mode=m_Strategy.Slot[closeSlot].IndicatorPointer.ListParam[1].Text;
      if(mode!="New bar")
         trlMode=TrailingStopMode_Tick;
     }

   int breakEven=m_Strategy.UseBreakEven ? m_Strategy.BreakEven : 0;

   m_TrailingMode = trlMode;
   m_TrailingStop = trlStop;
   m_BreakEven    = breakEven;

   bool response=ManageOrderSend(type,lots,stoploss,takeprofit);

   if(response)
     {
      // The order was executed successfully.
      m_TimeLastEntryBar=m_BarTime;
      if(type==TraderOrderType_Buy)
         m_IsEnteredLong=true;
      else
         m_IsEnteredShort=true;

      m_DataMarket.WrongStopLoss   = 0;
      m_DataMarket.WrongTakeProf   = 0;
      m_DataMarket.WrongStopsRetry = 0;
     }
   else
     {
      // Error in operation execution.
      m_DataMarket.WrongStopLoss = (int) stoploss;
      m_DataMarket.WrongTakeProf = (int) takeprofit;
     }
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::DoExitTrade()
  {
   if(!m_DataMarket.IsFailedCloseOrder)
      m_DataMarket.IsSentCloseOrder=true;
   m_DataMarket.CloseOrderTickCounter=0;

   bool orderResponse=CloseCurrentPosition();

   m_DataMarket.WrongStopLoss   = 0;
   m_DataMarket.WrongTakeProf   = 0;
   m_DataMarket.WrongStopsRetry = 0;

   return (orderResponse);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
bool ActionTrade5::IsWrongStopsExecution()
  {
   const int maxRetry=4;

   if(m_DataMarket.PositionDirection==PosDirection_Closed ||
      m_DataMarket.PositionLots<Epsilon ||
      m_DataMarket.WrongStopsRetry>=maxRetry)
     {
      m_DataMarket.WrongStopLoss  =0;
      m_DataMarket.WrongTakeProf  =0;
      m_DataMarket.WrongStopsRetry=0;
      return (false);
     }

   bool isWrongStop=(m_DataMarket.WrongStopLoss>0 && m_DataMarket.PositionStopLoss<Epsilon) ||
                    (m_DataMarket.WrongTakeProf>0 && m_DataMarket.PositionTakeProfit<Epsilon);
   return (isWrongStop);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void ActionTrade5::ResendWrongStops()
  {
   double lots       = m_DataMarket.PositionLots;
   double price      = m_DataMarket.PositionDirection == PosDirection_Long ? m_DataMarket.Bid : m_DataMarket.Ask;
   int    ticket     = m_DataMarket.PositionTicket;
   double stoploss   = m_DataMarket.WrongStopLoss;
   double takeprofit = m_DataMarket.WrongTakeProf;

   if(stoploss>0)
     {
      double stopLossPrice=0;
      if(m_DataMarket.PositionDirection==PosDirection_Long)
         stopLossPrice=m_DataMarket.Bid-stoploss*m_DataSet[0].Point;
      else
         if(m_DataMarket.PositionDirection==PosDirection_Short)
            stopLossPrice=m_DataMarket.Ask+stoploss*m_DataSet[0].Point;
     }
   if(takeprofit>0)
     {
      double takeProfitPrice=0;
      if(m_DataMarket.PositionDirection==PosDirection_Long)
         takeProfitPrice=m_DataMarket.Bid+takeprofit*m_DataSet[0].Point;
      else
         if(m_DataMarket.PositionDirection==PosDirection_Short)
            takeProfitPrice=m_DataMarket.Ask-takeprofit*m_DataSet[0].Point;
     }

   bool responseOk=ModifyPosition(stoploss,takeprofit);
   if(responseOk)
     {
      m_DataMarket.WrongStopLoss   = 0;
      m_DataMarket.WrongTakeProf   = 0;
      m_DataMarket.WrongStopsRetry = 0;
     }
   else
      m_DataMarket.WrongStopsRetry++;
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::TradingSize(double size)
  {
   if(m_Strategy.UseAccountPercentEntry)
      size=(size/100)*m_DataMarket.AccountEquity/m_DataSet[0].MarginRequired;
   return NormalizeEntrySize(size);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
double ActionTrade5::AccountPercentStopPoints(double percent,double lots)
  {
   double balance   = m_DataMarket.AccountBalance;
   double moneyrisk = balance*percent/100;
   double spread    = m_DataSet[0].Spread;
   double tickvalue = m_DataSet[0].TickValue;
   double stoploss  = moneyrisk/(lots*tickvalue) - spread;
   return (stoploss);
  }


// The Forex Strategy Builder Expert
ActionTrade5* actionsTrade;

//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
int OnInit()
  {
   actionsTrade = new ActionTrade5();

   actionsTrade.Entry_Amount    = (Entry_Amount   >77700)?0.1:Entry_Amount;
   actionsTrade.Maximum_Amount  = (Maximum_Amount >77700)?0.1:Maximum_Amount;
   actionsTrade.Adding_Amount   = (Adding_Amount  >77700)?0.1:Adding_Amount;
   actionsTrade.Reducing_Amount = (Reducing_Amount>77700)?0.1:Reducing_Amount;

   actionsTrade.Strategy_File_Name       = Strategy_File_Name;
   actionsTrade.Strategy_XML             = Strategy_XML;
   actionsTrade.Max_Data_Bars            = Max_Data_Bars;
   actionsTrade.Protection_Min_Account   = Protection_Min_Account;
   actionsTrade.Protection_Max_StopLoss  = Protection_Max_StopLoss;
   actionsTrade.Separate_SL_TP           = Separate_SL_TP;
   actionsTrade.Write_Log_File           = Write_Log_File;
   actionsTrade.TrailingStop_Moving_Step = TrailingStop_Moving_Step;
   actionsTrade.FIFO_order               = FIFO_order;
   actionsTrade.MaxLogLinesInFile        = MaxLogLinesInFile;
   actionsTrade.Bar_Close_Advance        = Bar_Close_Advance;

   int result=actionsTrade.OnInit();

   if(result==INIT_SUCCEEDED)
      actionsTrade.OnTick();

   return (result);
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnTick()
  {
   if(__symbol!=_Symbol || __period!=_Period)
     {
      if(__period > 0)
        {
         actionsTrade.OnDeinit(-1);
         actionsTrade.OnInit();
        }
      __symbol=_Symbol;
      __period=_Period;
     }

   actionsTrade.OnTick();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnTrade()
  {
   actionsTrade.OnTrade();
  }
//+------------------------------------------------------------------+
//|                                                                  |
//+------------------------------------------------------------------+
void OnDeinit(const int reason)
  {
   actionsTrade.OnDeinit(reason);

   if(CheckPointer(actionsTrade)==POINTER_DYNAMIC)
      delete actionsTrade;
  }
//+------------------------------------------------------------------+

//+------------------------------------------------------------------+
